IMPLEMENTATION MODULE Buffer;

IMPORT SYSTEM, UNSAFE;
FROM Transition IMPORT OCTET;
FROM Transition IMPORT CAPACITYo, IncAdr;

FROM Storage IMPORT ALLOCATE, DEALLOCATE;

TYPE
  Buffer = POINTER TO BufferInfo;
  BufferInfo = RECORD
    startAdr : UNSAFE.ADDRESS;
    currPos, maxPos, lastValidPos : Pos;
    offset : Pos; (* to delineate, say, a file window for the buffer *)
    dataHasBeenEntered : BOOLEAN;  (* TRUE if data has been read into it since last reset *)
    isDirty : BOOLEAN   (* TRUE if data has been written into it since last reset *)
  END;

CONST
  InvalidBuff = NIL;


PROCEDURE InvalidBuffer () : Buffer;
(* returns the invalid Buffer *)

BEGIN
  RETURN InvalidBuff
END InvalidBuffer;


PROCEDURE New ( (*NEW*) VAR buffer : Buffer; size : Size);

BEGIN
  NEW (buffer);
  ALLOCATE (buffer^.startAdr, size);
  buffer^.maxPos := size - 1;
  Reset (buffer);
  buffer^.offset := 0
END New;


PROCEDURE Release (VAR buffer : Buffer);

BEGIN
  DEALLOCATE (buffer^.startAdr, buffer^.maxPos+1);
  DISPOSE (buffer)
END Release;


(* info section *)
PROCEDURE BufferAdr ( buffer : Buffer ) : SYSTEM.ADDRESS;
(* returns the address of the first octet in the buffer for use by file handlers and the like to
enter data directly. *)

BEGIN
  RETURN buffer^.startAdr
END BufferAdr;


PROCEDURE isDirty ( buffer : Buffer) : BOOLEAN;

BEGIN
  RETURN buffer^.isDirty
END isDirty;


PROCEDURE isFull ( buffer : Buffer) : BOOLEAN;

BEGIN
  RETURN buffer^.lastValidPos = buffer^.maxPos + 1
END isFull;


PROCEDURE hasUnreadData ( buffer : Buffer) : BOOLEAN;
(* returns TRUE if the current position is leas than or equal to the last readable position *)

BEGIN
  RETURN buffer^.dataHasBeenEntered AND ( buffer^.currPos <= buffer^.lastValidPos )
END hasUnreadData;


PROCEDURE count ( buffer : Buffer) : Pos;

BEGIN
  IF hasUnreadData (buffer) THEN
    RETURN buffer^.lastValidPos - buffer^.currPos
  ELSE
    RETURN 0
  END; (* IF buffer *)
END count;


PROCEDURE eob ( buffer : Buffer ) : BOOLEAN;
(* Returns TRUE if currentPos has reached the end of the buffered material, otherwise FALSE. *)

BEGIN
  RETURN buffer^.currPos = buffer^.lastValidPos
END eob;


PROCEDURE currentPos ( buffer : Buffer) : Pos;

BEGIN
  RETURN buffer^.currPos
END currentPos;

PROCEDURE lastValidPos ( buffer : Buffer ) : Pos;
(* returns the highest valid position for reading from the buffer *)

BEGIN
  RETURN buffer^.lastValidPos
END lastValidPos;


PROCEDURE isValidReadPos ( buffer : Buffer; pos : Pos ) : BOOLEAN;
(* Returns TRUE if <pos> is a valid position to set for reading from <buffer>, otherwise FALSE.
  the value must be in the range 0..buffer^.lastValidPos, that is, the range of
  positions that are active in the buffer *)

BEGIN
  RETURN pos <=  buffer^.lastValidPos
END isValidReadPos;


PROCEDURE isValidWritePos ( buffer : Buffer; pos : Pos ) : BOOLEAN;
(* Returns TRUE if <pos> is a valid position to set for writing to <buffer>, otherwise FALSE.
  the value must be in the range 0..MIN ( CAPACITY (buffer), size (buffer-1 ) that is,
  the range of positions for writing into the buffer *)

BEGIN
  RETURN pos <= buffer^.lastValidPos + 1
END isValidWritePos;


PROCEDURE SetOffset ( buffer : Buffer; offset : Pos );
(*  Sets the offset position of <buffer> to <pos>. 
This parameter is needed to define, say, the start of the window into a file
  that is represented by the buffer.*)

BEGIN
  buffer^.offset := offset
END SetOffset;

PROCEDURE offset ( buffer : Buffer ) : Pos ;
(*  Returns the value of the offset position of <buffer> to <pos>. *)

BEGIN
  RETURN buffer^.offset
END offset;

PROCEDURE Reset ( buffer : Buffer);
(* sets the position back to zero *)

BEGIN
  buffer^.currPos := 0;
  buffer^.lastValidPos := 0;
  buffer^.dataHasBeenEntered := FALSE;
  buffer^.isDirty := FALSE
END Reset;


PROCEDURE SetPos ( buffer : Buffer; pos : Pos );
(* Pre: Pos is valid for what the user intends to do
   Sets the current position of <buffer> to <pos>. 
   IF pos <> lastValidWrite Pos, does nothing *)
BEGIN
  IF pos <= buffer^.lastValidPos + 1 THEN
    buffer^.currPos := pos;
  END
END SetPos;


PROCEDURE Advance ( buffer : Buffer; offset : Pos );
(* Pre: the current pos plus the offset must be a valid position
   Advances the current position of <buffer> by <offset>. *)
VAR
  newPos : Pos;
BEGIN
  newPos := buffer^.currPos + offset;
  SetPos ( buffer, newPos )
END Advance;


PROCEDURE Rewind ( buffer : Buffer );
(* Sets the read/write position for buffer accessor <buffer> to the  beginning
   of the buffer. Does not reset isDirty. *)

BEGIN
  buffer^.currPos := 0;
END Rewind;


(* operations *)


PROCEDURE ReadOctet ( buffer : Buffer; VAR data : OCTET );
(* Pre: the read position is valid
  Reads one octet of data at the current position of <buffer>, passes it back
  in <data> and advances the read/write position of <buffer> by one. *)
VAR
  bufAdr : POINTER TO OCTET;

BEGIN
    bufAdr := SYSTEM.ADDADR (buffer^.startAdr, buffer^.currPos);
    data := bufAdr^;
    buffer^.currPos := buffer^.currPos + 1
END ReadOctet;


PROCEDURE ReadBlock ( buffer : Buffer; VAR data : ARRAY OF OCTET; VAR octetsRead : Size );
(*  Pre: the read position is valid
   Reads a block of data up to the capacity of <data>, starting at the current
   position  of <buffer>  and passes it back in <data>.  The read/write position
   of <buffer> is advanced accordingly. *)

VAR
  destEnd, destCounter, bufCounter : Pos;
  bufAdr : POINTER TO OCTET;

BEGIN
  destCounter := 0;
  bufAdr := SYSTEM.ADDADR (buffer^.startAdr, buffer^.currPos);
  destEnd := CAPACITYo (data) - 1;
  WHILE ( destCounter <= destEnd ) AND ( buffer^.currPos <= buffer^.lastValidPos ) DO
    data [destCounter] := bufAdr^;
    destCounter := destCounter + 1;
    buffer^.currPos := buffer^.currPos + 1;
    IncAdr (bufAdr)
  END;
  octetsRead := destCounter - 1;
END ReadBlock;


PROCEDURE ReadBlockToOffset ( buffer : Buffer; VAR data : ARRAY OF OCTET; offset : Pos; VAR octetsRead : Size );
(*  Pre: the read position is valid
   Reads a block of data up to the capacity of <data>, starting at the current
   position  of <buffer>  and passes it back in <data> starting at the position <offset> .
   The read/write position  of <buffer> is advanced accordingly.
   octetsRead is set to <offset> + the number of octets read on the current call. *)

VAR
  destEnd, destCounter, bufCounter : Pos;
  bufAdr : POINTER TO OCTET;

BEGIN
  destCounter := offset;
  bufAdr := SYSTEM.ADDADR (buffer^.startAdr, buffer^.currPos);
  destEnd := CAPACITYo (data) - 1;
  WHILE ( destCounter <= destEnd ) AND ( buffer^.currPos <= buffer^.lastValidPos ) DO
    data [destCounter] := bufAdr^;
    destCounter := destCounter + 1;
    buffer^.currPos := buffer^.currPos + 1;
    IncAdr (bufAdr)
  END;
  octetsRead := destCounter - offset;
END ReadBlockToOffset;

PROCEDURE WriteOctet ( buffer : Buffer; data : OCTET );
(* Pre: Buffer is not full
   Writes the octet passed in <data> to <buffer> and advances the current position by 1. *)
VAR
  bufAdr : POINTER TO OCTET;

BEGIN
  bufAdr := SYSTEM.ADDADR (buffer^.startAdr, buffer^.currPos);
  bufAdr^ := data;
  buffer^.currPos := buffer^.currPos + 1;
  IF buffer^.currPos >  buffer^.lastValidPos THEN
     buffer^.lastValidPos := buffer^.currPos
  END;
  buffer^.isDirty := TRUE;
  buffer^.dataHasBeenEntered := TRUE;
END WriteOctet;


PROCEDURE WriteBlock (buffer : Buffer; source : (*CAST*) ARRAY OF OCTET; VAR octetsWritten : Size );

VAR
  sourceEnd, sourceCounter : Pos;
  num : Size;
  bufAdr : POINTER TO OCTET;

BEGIN
  bufAdr := SYSTEM.ADDADR (buffer^.startAdr, buffer^.currPos);
  sourceCounter := 0;
  sourceEnd := CAPACITYo (source) - 1;
  IF (sourceCounter <= sourceEnd) THEN
  REPEAT
    bufAdr^ := source [sourceCounter];
    sourceCounter := sourceCounter + 1;
    buffer^.currPos := buffer^.currPos + 1;
    IncAdr (bufAdr)
  UNTIL (sourceCounter = sourceEnd) OR (buffer^.currPos = buffer^.maxPos)
  END;

  buffer^.isDirty := buffer^.isDirty OR ( sourceCounter > 0 );
  buffer^.dataHasBeenEntered := buffer^.dataHasBeenEntered OR ( sourceCounter > 0 );
  IF buffer^.currPos >  buffer^.lastValidPos THEN
     buffer^.lastValidPos := buffer^.currPos
  END;
   octetsWritten := sourceCounter
END WriteBlock;


PROCEDURE WriteBlockFromOffset ( buffer : Buffer; data : ARRAY OF OCTET; offset : Pos; VAR octetsWritten : Size );
(* Writes a block of data starting from the <offset> in the data array up to the capacity of <data>, 
  and starting to enter at the current position  of <buffer>  .
  The read/write position  of <buffer> is advanced accordingly.
  octetsWritten is set to <offset> + the number of octets written on the current call. *)

VAR
  sourceEnd, sourceCounter : Pos;
  num : Size;
  bufAdr : POINTER TO OCTET;

BEGIN
  bufAdr := SYSTEM.ADDADR (buffer^.startAdr, buffer^.currPos);
  sourceCounter := offset;
  sourceEnd := CAPACITYo (data) - 1;
  WHILE (sourceCounter <= sourceEnd) AND (buffer^.currPos <= buffer^.maxPos) DO
    bufAdr^ := data [sourceCounter];
    sourceCounter := sourceCounter + 1;
    buffer^.currPos := buffer^.currPos + 1;
    IncAdr (bufAdr)
  END;
  buffer^.isDirty := buffer^.isDirty OR ( sourceCounter > 0 );
  buffer^.dataHasBeenEntered := buffer^.dataHasBeenEntered OR ( sourceCounter > 0 );
   IF buffer^.currPos >=  buffer^.lastValidPos THEN
     buffer^.lastValidPos := buffer^.currPos
  END;
   octetsWritten := sourceCounter
END WriteBlockFromOffset;


PROCEDURE RecordDirectEntry  ( buffer : Buffer; start : Pos; numOctetsEntered : Size );
(* Inform the buffer about an entry made directly to one of its addresses to allow it to update its counters. This must be
called, say, after reading via a file handler into the buffer using its address. Does not set isDirty *)

BEGIN
  buffer^.currPos := start;
  buffer^.dataHasBeenEntered := TRUE;
  IF start + numOctetsEntered > buffer^.lastValidPos THEN 
    buffer^.lastValidPos := start + numOctetsEntered -1
  END (* IF start *)
END RecordDirectEntry;

END Buffer. 