IMPLEMENTATION MODULE CharArray;

(* (C) 2009-2017 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

FROM Transition IMPORT CAPACITYch;
IMPORT ISO646, Perror;
IMPORT Buffer; FROM Buffer IMPORT BufferT;
IMPORT IOSupport; FROM IOSupport IMPORT*;

VAR
  clipboard : BufferT;
  clipboardSize : CARDINAL;
  
PROCEDURE Init ();
BEGIN
  Buffer.New ( clipboard, DefaulClipboardSize );
  clipboardSize := DefaulClipboardSize; 
END Init;


(* Additional support for ARRAY OF CHAR not in the module by that name, which is runtime support. Any utilities necessary
for handling this type should be located here. *)


(* ****** The Clipboard ******* *)

(* The following implement cut copy and paste with an internal clipboard.*)

PROCEDURE ResetClipboard ( newSize : CARDINAL );
(* The current clipboard is emptied. If newSize = 0 nothing else happens.
  Else if 0 < newSize <=  MinClipboardSize it is sized to MinClipboardSize.
  Otherwise it is set to newSize *)

BEGIN
  Buffer.Reset ( clipboard ); (* empty it *)
  IF ( newSize = clipboardSize ) OR ( newSize = 0 ) THEN RETURN END; (* no size change *)
  IF newSize < MinClipboardSize THEN newSize := MinClipboardSize END;
 
 (* if we get here, we are replacing the lipboard with one of a different size *)
  Buffer.Release ( clipboard ); 
  Buffer.New ( clipboard, newSize);
  clipboardSize := newSize
END ResetClipboard;

PROCEDURE Cut
  ( VAR target : ARRAY OF CHAR; startIndex : CARDINAL; numToCut : CARDINAL );
(* Deletes numToCut chars from target starting at startIndex. If numToCut is greater
than or equal to the length of the string, the target becomes an empty string. Too large 
a number to cut does not therefore yield an error.
Otherwise the characters after the ones deleted are moved over and the string re-terminated.
The characters cut are copied to the clipboard.*)

BEGIN
  Copy (target, startIndex, numToCut);
  Delete (target, startIndex, numToCut)
END Cut;


PROCEDURE Copy
  ( source : ARRAY OF CHAR; startIndex : CARDINAL; numToCopy : CARDINAL );
(* Copies numToCopy chars from source starting at startIndex to the clipboard. 
    If the parameters provided would mean copying past the end of the string, 
    copying will be done as far as the end of the string and there is no error. *)

VAR
  index, end, lastSrc : CARDINAL;
  
BEGIN  
  Buffer.Reset (clipboard);
  lastSrc := LENGTH (source) - 1;
  index := startIndex;
  end := startIndex + numToCopy - 1;
  IF end > lastSrc THEN end := lastSrc END; (* only copy what's available *)
  WHILE index <= end DO
    Buffer.WriteOctet (clipboard, source [index] );
    index := index + 1;
  END; (* while *)
  Buffer.Rewind (clipboard) (* set back to start for potential paste *)
END Copy;


PROCEDURE Paste
  ( VAR target : ARRAY OF CHAR; startIndex, numToReplace : CARDINAL );
(* Replaces numToReplace chars from target starting at startIndex by the contents of
  the clipboard. If numToReplace exceeds the number of characters with higher indices than
  startIndex, all are replaced and there is no error. Thus if both numToReplace and startIndex
  are zero, target simply is assigned the contents of the clipboard and is terminated.
  Otherwise, if the number of characters in the clipboard exceeds numToReplace, any subsequent 
  characters in target are moved to higher positions to make room. If the number of characters 
  in the clipboard is fewer than numToReplace, subsequent characters in target are moved
  to lower positions to close up the gap. If the resulting string length would exceed the capacity of target, runtime fault
  StringCapacityExceeded is raised.*)

VAR
  uCount, lCount, maxChars, lenTgt, tailSize, endIndex, numToPaste, distToMove : CARDINAL;

BEGIN
  maxChars := CAPACITYch (target);
  lenTgt := LENGTH (target);
  
  tailSize := lenTgt -startIndex;  
  IF numToReplace > tailSize THEN numToReplace := tailSize END;
  numToPaste := Buffer.count (clipboard);

  IF lenTgt - numToReplace + numToPaste > maxChars THEN
    Perror.prerr ( "StringCapacityExceeded error in paste. ");
    RETURN
  ELSIF startIndex > lenTgt  THEN
    Perror.prerr ( "String Index Out Of Bounds error in paste. ");
    RETURN
  END; (* IF lenTgt *)
  
  IF numToPaste > numToReplace THEN (* move right to make room *)
    distToMove := numToPaste - numToReplace;
   uCount := lenTgt + distToMove;
   target [uCount] := ISO646.NUL;
   lCount := lenTgt;
    endIndex := startIndex + numToReplace;
    WHILE lCount > endIndex DO
      lCount := lCount - 1;
      uCount := uCount - 1;
      target [uCount] := target [lCount]
    END (* WHILE *)
  ELSIF numToPaste < numToReplace THEN (* move left *)
    distToMove := numToReplace - numToPaste;
    lCount := startIndex + numToPaste;
    uCount := lCount + distToMove;
    endIndex := lenTgt;
    WHILE uCount <= endIndex DO
      target [lCount] := target [uCount];
      lCount := lCount + 1;
      uCount := uCount + 1
    END (* WHILE *)
  ELSE (* no need to move *)
    
  END; (* IF numToPaste *)
  
  (* made room so do the paste *)
  lCount := startIndex;
  endIndex := lCount + numToPaste -  1;
  WHILE lCount <= endIndex DO
    Buffer.ReadOctet ( clipboard, target [lCount] );
    lCount := lCount + 1
  END; (* while *) 
  Buffer.Rewind (clipboard) (* set back to start for potential copy/paste *)
END Paste;


(* ****** END Clipboard ******* *)

PROCEDURE Delete 
  ( VAR target : ARRAY OF CHAR; startIndex : CARDINAL; numToDelete : CARDINAL);
(* removes numToDelete characters from target starting at startIndex, moving subsequent chars 
  back to fill up the gap and re-terminating the string. This does not put the removed characters
  into the clipboard. If the sum of startIndex and numToDelete incorporates more
  characters than there are following startIndex, all are deleted and there is no error. *)

VAR
  uCount, lCount, last : CARDINAL;
  
BEGIN
  last := LENGTH (target) - 1;
  lCount := startIndex;
  uCount := startIndex + numToDelete;
  IF uCount > last THEN
    uCount := last + 1;
    lCount := lCount + 1
  END; (* and then the following loop is skipped *)
  WHILE (uCount <= last) DO
    target [lCount] := target [uCount];
    lCount := lCount + 1;
    uCount := uCount + 1
  END; (* while *)
  target [lCount] := ISO646.NUL
END Delete;

(*   end of Clipboard section      *)


PROCEDURE ZeroOutCharArray ( VAR str : ARRAY OF CHAR );
  (* replaces everything in the string with CHR (0) *)
VAR
  count : CARDINAL; (* Counting Variable *)
  hiStr : CARDINAL; (* hold high of string for comparisons *)

BEGIN
  hiStr  := CAPACITYch (str) - 1;
  count := 0;
  WHILE (count <= hiStr)  DO
    str[count] := CHR(0);
  count := count + 1
  END (* while *)
END ZeroOutCharArray;


PROCEDURE ToUpper ( VAR str : ARRAY OF CHAR );
  (* replaces all lowercase letters in the string with the corresponding
     uppercase letter leaving everything else untouched. Skips \n and \t *)

VAR
  count, len : CARDINAL;
  ch : CHAR;

BEGIN
  count := 0;
  len := LENGTH (str);
  WHILE count < len DO
    IF (str[count] = ISO646.REVERSE_SOLIDUS) THEN (* skip \n and \t *)
      count := count + 1;
      IF ( str[count] = "n" ) OR  ( str[count] = "t") THEN
        count := count + 1;
        IF count = len THEN RETURN END (* catch case of \n or \t at the end *)
      END (* IF ( str[count] = "n" ) *)
    END; (* IF (str[count] = ISO646.REVERSE_SOLIDUS) *)
    str[count] := ISO646.toUpper ( str[count] );
    count := count + 1
  END (* while *)
END ToUpper;

PROCEDURE ToLower ( VAR str : ARRAY OF CHAR );
  (* replaces all uppercase letters in the string with the corresponding
     lowercase letter leaving everything else untouched. *)

VAR
  count, len : CARDINAL;

BEGIN
  count := 0;
  len := LENGTH (str);
  WHILE count < len DO
    str[count] := ISO646.toLower ( str[count] );
    count := count + 1
  END (* while *)
END ToLower;


PROCEDURE Capitalize ( VAR str : ARRAY OF CHAR );
  (* uppercases the first letter in the string, and all subsequent letters that follow
     whitespace, tab, \n, or \t in the string that with the corresponding
     uppercase letter leaving everything else untouched. *)

VAR
  count, len : CARDINAL;
  trigger : BOOLEAN;
  
BEGIN
  count := 0;
  len := LENGTH (str);
  IF len > 0 THEN str[0] := ISO646.toUpper ( str[0] ) END;
  WHILE count < len DO
    trigger := FALSE;
    IF (str[count] = ISO646.REVERSE_SOLIDUS) THEN (* skip \n and \t *)
      count := count + 1;
      IF ( str[count] = "n" ) OR  ( str[count] = "t") THEN
        trigger := TRUE;
        count := count + 1;
        IF count = len THEN RETURN END (* catch case of \n or \t at the end *)
      END (* IF ( str[count] = "n" ) *)
    ELSE
      trigger := ISO646.isWhitespaceOrTab (str[count] );
      WHILE ISO646.isWhitespaceOrTab (str[count] ) DO
        count := count + 1;
        IF count = len THEN RETURN END (* catch case of whitespace at the end *)
      END (* WHILE ISO646.isWhitespaceOrTab *)
    END; (* IF (str[count] = ISO646.REVERSE_SOLIDUS) *)
    IF trigger THEN
      str[count] := ISO646.toUpper ( str[count] );
    END; (* IF trigger *)
    count := count + 1
  END (* while *)
END Capitalize;


PROCEDURE CharToAOC ( VAR dest : ARRAY OF CHAR; ch : CHAR );
BEGIN
  dest[0] := ch;
  IF (ch # CHR(0)) AND (HIGH(dest) > 0) THEN
    dest[1] := CHR(0)
  END (* IF *) 
END CharToAOC;


PROCEDURE CopyTokenFromIndex
  ( VAR target : ARRAY OF CHAR; source : ARRAY OF CHAR; n : CARDINAL );
(* Copies only the alphanumeric characters in source starting at the n-th character and ending
when encountering the first non-alphanumeric character to target array. If the character at position
 is non alphanumeric, returns the empty string.*)
VAR
  sCount, tCount : CARDINAL;

BEGIN
  tCount := 0;
  sCount := n;
  IF n >= LENGTH (source ) THEN 
    target [tCount] := ISO646.NUL;
    RETURN
  END; (* if *) 
  WHILE ISO646.isAlphaNumeric (source[sCount] ) DO
    target [tCount] := source [sCount];
    sCount := sCount + 1;
    tCount := tCount + 1;
  END; (* while *)
  target [tCount] := ISO646.NUL
END CopyTokenFromIndex;

PROCEDURE nextIndexOfChar ( (*CONST*) source : ARRAY OF CHAR; startIndex : CARDINAL; ch : CHAR ) : CARDINAL;
  (* Beginning at startIndex, find the next instance of ch and return its index position in the array. 
    If ch is not found in that span, returns LENGTH ( source) *)
VAR
  count : CARDINAL;
BEGIN
  count := startIndex;
  WHILE (source [count] <> ISO646.NUL) AND (source [count] <> ch) DO
    count := count + 1;
  END; (* while *)
  RETURN count
END nextIndexOfChar;

PROCEDURE collatesBefore
  ( array1, array2 : ARRAY OF CHAR ) : BOOLEAN;
(* Returns TRUE if array1 comes before array2 in ASCII collation order. *)
VAR
  count : CARDINAL; (* Counting Variable *)
  len1, len2 : CARDINAL; (* to hold the lengths of the two strings for comparisons *)
  hi, hi1, hi2 : CARDINAL; (* holds highest index to compare *)

BEGIN
  len1 := LENGTH (array1);
  len2 := LENGTH (array2);
  hi1 := len1 - 1;
  hi2 := len2 - 1;
  IF hi1 < hi2 (* use least of two counts *)  THEN
    hi := hi1
  ELSE
    hi := hi2
  END; (* if *)

  count := 0;
  (* check for leading chars being equal first, land on first one not equal *)
  WHILE (count <= hi) AND
        (array1[count] = array2 [count])  AND
        (array1[count] # CHR (0)) (* if only #2 hits CHR (0) they won't be equal *) DO
    count := count + 1
  END; (* WHILE *)

  IF count <= hi (* didn't go off the end of shortest array *)  THEN
      (* compare last item *)
    IF array1[count] < array2 [count] THEN
       RETURN TRUE;
     ELSE
       RETURN FALSE
     END (* IF stringVal1 *)
  END; (* IF count <= hi *)

  (* if got here, then they are equal as far as the end of the shortest *)
  IF len1 < len2 THEN
    RETURN TRUE;
  ELSE
    RETURN FALSE;
  END (* IF len1 < len2 *)
END collatesBefore;

PROCEDURE matches
  ( array1, array2 : ARRAY OF CHAR ) : BOOLEAN;
(* Returns TRUE if the content of array1 matches that of array2. *)
VAR
  count : CARDINAL; (* Counting Variable *)
  len : CARDINAL; (* to hold the lengths of the two strings for comparisons *)
  hi : CARDINAL; (* holds highest index to compare *)

BEGIN
  len := LENGTH (array1);
  IF len <> LENGTH (array2) THEN RETURN FALSE END;
  hi := len - 1;

  count := 0;
  (* check for leading chars being equal first, land on first one not equal *)
  WHILE (count <= hi) AND
        (array1[count] = array2[count])  AND
        (array1[count] # CHR (0)) (* if only #2 hits CHR (0) they won't be equal *) DO
    count := count + 1
  END; (* WHILE *)

  IF count < hi (* didn't go off the end of the arrays *)  THEN
    RETURN FALSE
  ELSE
    RETURN TRUE
  END; (* IF count <= hi *)

END matches;

PROCEDURE collatesBeforeOrMatches
  ( array1, array2 : ARRAY OF CHAR ) : BOOLEAN;
(* Returns TRUE if array1 comes before array2 in ASCII collation order. *)
VAR
  count : CARDINAL; (* Counting Variable *)
  len1, len2 : CARDINAL; (* to hold the lengths of the two strings for comparisons *)
  hi, hi1, hi2 : CARDINAL; (* holds highest index to compare *)

BEGIN
  len1 := LENGTH (array1);
  len2 := LENGTH (array2);
  hi1 := len1 - 1;
  hi2 := len2 - 1;
  IF hi1 < hi2 (* use least of two counts *)  THEN
    hi := hi1
  ELSE
    hi := hi2
  END; (* if *)

  count := 0;
  (* check for leading chars being equal first, land on first one not equal *)
  WHILE (count <= hi) AND
        (array1[count] = array2[count])  AND
        (array1[count] # CHR (0)) (* if only #2 hits CHR (0) they won't be equal *) DO
    count := count + 1
  END; (* WHILE *)

  IF count <= hi (* didn't go off the end of shortest array *)  THEN
      (* compare last item *)
    IF array1[count] <= array2[count] THEN
       RETURN TRUE;
     ELSE
       RETURN FALSE
     END (* IF stringVal1 *)
  END; (* IF count <= hi *)

  (* if got here, then they are equal as far as the end of the shortest *)
  IF len1 <= len2 THEN
    RETURN TRUE;
  ELSE
    RETURN FALSE
  END (* IF len1 < len2 *)
END collatesBeforeOrMatches;


PROCEDURE collatesBeforeLex
  ( array1, array2 : ARRAY OF CHAR ) : BOOLEAN;
(* Returns TRUE if array1 comes before array2 in ASCII collation order. *)
VAR
  count : CARDINAL; (* Counting Variable *)
  len1, len2 : CARDINAL; (* to hold the lengths of the two strings for comparisons *)
  hi, hi1, hi2 : CARDINAL; (* holds highest index to compare *)

BEGIN
  len1 := LENGTH (array1);
  len2 := LENGTH (array2);
  hi1 := len1 - 1;
  hi2 := len2 - 1;
  IF hi1 < hi2 (* use least of two counts *)  THEN
    hi := hi1
  ELSE
    hi := hi2
  END; (* if *)

  count := 0;
  (* check for leading chars being equal first, land on first one not equal *)
  WHILE (count <= hi) AND
        ( ISO646.LexORD (array1[count] )  = ISO646.LexORD (array2 [count]) ) AND
        (array1[count]  # CHR (0)) (* if only #2 hits CHR (0) they won't be equal *) DO
    count := count + 1
  END; (* WHILE *)

  IF count <= hi (* didn't go off the end of shortest array *)  THEN
      (* compare last item *)
    IF ISO646.LexORD (array1[count] ) < ISO646.LexORD (array2 [count] ) THEN
       RETURN TRUE;
     ELSE
       RETURN FALSE
     END (* IF stringVal1 *)
  END; (* IF count <= hi *)

  (* if got here, then they are equal as far as the end of the shortest *)
  IF len1 < len2 THEN
    RETURN TRUE;
  ELSE
    RETURN FALSE;
  END (* IF len1 < len2 *)
END collatesBeforeLex;


PROCEDURE matchesLex
  ( array1, array2 : ARRAY OF CHAR ) : BOOLEAN;
(* Returns TRUE if the content of array1 lexically matches that of array2. *)
(* Returns TRUE if the content of array1 matches that of array2. *)
VAR
  count : CARDINAL; (* Counting Variable *)
  len : CARDINAL; (* to hold the lengths of the two strings for comparisons *)
  hi : CARDINAL; (* holds highest index to compare *)

BEGIN
  len := LENGTH (array1);
  IF len <> LENGTH (array2) THEN RETURN FALSE END;
  hi := len - 1;

  count := 0;
  (* check for leading chars being equal first, land on first one not equal *)
  WHILE (count < hi) AND
        ( ISO646.LexORD(array1[count]) = ISO646.LexORD(array2[count]) ) AND
        (array1[count] # CHR (0)) (* if only #2 hits CHR (0) they won't be equal *) DO
    count := count + 1
  END; (* WHILE *)

  IF count < hi (* didn't go off the end of the arrays *)  THEN
    RETURN FALSE
  ELSE
    RETURN TRUE
  END; (* IF count <= hi *)

END matchesLex;

PROCEDURE collatesBeforeOrMatchesLex
  ( array1, array2 : ARRAY OF CHAR ) : BOOLEAN;
(* Returns TRUE if array1 comes before or matches array2 in lexicographical collation order. *)

VAR
  count : CARDINAL; (* Counting Variable *)
  len1, len2 : CARDINAL; (* to hold the lengths of the two strings for comparisons *)
  hi, hi1, hi2 : CARDINAL; (* holds highest index to compare *)

BEGIN
  len1 := LENGTH (array1);
  len2 := LENGTH (array2);
  hi1 := len1 - 1;
  hi2 := len2 - 1;
  IF hi1 < hi2 (* use least of two counts *)  THEN
    hi := hi1
  ELSE
    hi := hi2
  END; (* if *)

  count := 0;
  (* check for leading chars being equal first, land on first one not equal *)
  WHILE (count <= hi) AND
        ( ISO646.LexORD(array1[count]) = ISO646.LexORD(array2[count]) ) AND
        (array1[count] # CHR (0)) (* if only #2 hits CHR (0) they won't be equal *) DO
    count := count + 1
  END; (* WHILE *)

  IF count <= hi (* didn't go off the end of shortest array *)  THEN
      (* compare last item *)
    IF ISO646.LexORD(array1[count]) <= ISO646.LexORD(array2[count]) THEN
       RETURN TRUE;
     ELSE
       RETURN FALSE
     END (* IF stringVal1 *)
  END; (* IF count <= hi *)

  (* if got here, then they are equal as far as the end of the shortest *)
  IF len1 <= len2 THEN
    RETURN TRUE;
  ELSE
    RETURN FALSE
  END (* IF len1 < len2 *)
END collatesBeforeOrMatchesLex;


BEGIN (* main *)
  Init()
END CharArray.