(* This module implements a parser for the format strings used in WRITEF statements
and a formatter that builds out the string after the caller has converted the
entity and supplied it with the basic string. *)

IMPLEMENTATION MODULE FormatIO;
IMPORT SYSTEM, UNSAFE;

(* R10 Stuff *)
IMPORT ISO646, ARRAYOFCHAR, CharArray;
IMPORT Perror;
IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;

(* My utilities *)
IMPORT WholeStrs;
FROM IOSupport IMPORT*;

(* diagnostic 
IMPORT CharIO, CardinalIO, StdIO; *)

 (* PART ONE: types and global variables *)
TYPE
    FmtRes = (StrOK, NoFormatChar, InvalidChar, FLenOverflow, NumRequiredIntegralDigitsOverflow,
      NumIntDigitsForFloatOverflow, NumColsOverflow, NoClosingStrQuote, ArrayEnclosureMismatch,
      extraChars, missingEscChar, invalidEscChar );

VAR
  strNulls : Str;
 
CONST
  Concat = ARRAYOFCHAR.Concat;
  Assign = ARRAYOFCHAR.Assign;
  NullStr = "";
  SingleQuote = ISO646.SINGLE_QUOTE;
  DoubleQuote = ISO646.DOUBLE_QUOTE;
  CurrencyChar = ISO646.CURRENCY_SIGN;
  DefPadChar = ISO646.SPACE;
  NewlineEscSequence = "\n";
  CommaSeparator = ", "; (* default array item separater string is comma space *)
  SemicolonSeparator = ", ";  (* alternate array item separater string is semicolon space *)
  SpcSeparator = " "; (* likewise for strings, but just a single space; used for the default in grouping digit strings *)
  ArrayOptionStart =ISO646.VERTICAL_LINE;
  MaxFieldWidth = 999;
  MaxCols = 99;
  MaxDecimals = 999;
  MaxFills = 99;
  MaxInGroup = 99;
  MaxDigitsForFloat = 999;

  DefaultFormatScanRec = FormatScanRec
  {
  (* len *) 0,
  (* just*) Right,
  (* normFlag *) FALSE,
  (* complexFlag *) FALSE,
  (* isCurrency *) FALSE,
  (* reqSign *) FALSE,
  (* noSign *) FALSE,
  (* yes4Bool *) FALSE,
  (* isConcat *) FALSE,
  (* numeralFormat *) Dec,
  (* complexFormat *) i,
  (* realFormat *) Scientific,
  (* bitsetFormat *) Literal,
  (* quoteChar *) DoubleQuote,
  (* decChar *) ISO646.PERIOD,
  (* fillChar, groupSepChar *) ISO646.NUL, ISO646.NUL,
  (* numDecPlaces, numRequiredIntegralDigits, numIntDigitsForFloat, fieldLen *) 0, 0, 1, 0,
  (* numInGroup *) 3,
  (* prefixStr, suffixStr, customFormatStr *) NullStr, NullStr, NullStr,
  (* isColumnFormat *) FALSE,
  (* numCols *) 1,
  (* isArrayFormat *) FALSE,
  (* arrayOpenStr *) "{ ",
  (* arrayCloseStr *) " }",
  (* arrayItemSeperator *) CommaSeparator,
  (* terminalEscStr *) ""
  };

  RemoveChars = CharSet{"$", "*","-", " ", SingleQuote, DoubleQuote, "," , "x" , "b" , "u" };  (* chars to remove from the front of strings before formatting *)

   (* PART TWO The Format  string scanner material *)

(* Procedures for Custom scanning *)
VAR
  customScanSet : ScanSet;
  validCustomFmtChars : CharSet;
  
PROCEDURE SetCustomScanSet (custom : ScanSet);
(* a typeIO library can call this to inform a subsequent WRITEF what steps to take in the main parse
One would normally pass DefaultCustomScanSet [+ ScanSet {steps to add}) ] [- ScanSet {steps to remove}) ] *)

BEGIN
  customScanSet := custom
END SetCustomScanSet;


PROCEDURE GetCustomScanSet () : ScanSet;
(* a typeIO library can call this to retrieve the scan set it has entered *)

BEGIN
  RETURN customScanSet;
END GetCustomScanSet;


PROCEDURE SetValidCustomFmtChars (chars : CharSet);
(* a typeIO library can call this to inform a subsequent WRITEF what Format chars are allowed in a custom format string
Once would normally pass ValidCustomFmtChars [+ CharSet {chars to add}) ] [- CharSet {chars to remove}) ] *)

BEGIN
  validCustomFmtChars := chars
END SetValidCustomFmtChars;


PROCEDURE GetValidCustomFmtChars () : CharSet;
(* a typeIO library can call this to retrieve the Custom Format Chars it has set. *)

BEGIN

  RETURN validCustomFmtChars;
END GetValidCustomFmtChars;
(* Utility Procs First two are exported *)

PROCEDURE GetPadChar () : CHAR;
(* returns the current default space character used in xxIO modules *)

BEGIN
  RETURN padChar
END GetPadChar;


PROCEDURE SetPadChar (ch :  CHAR);
(* allows the padding char to be changed for debugging *)

BEGIN
  padChar := ch
END SetPadChar;


PROCEDURE GetDefaultFormatScanRec () : FormatScanRec;
(* returns a format record with all the defaults set *)

BEGIN
  RETURN DefaultFormatScanRec
END GetDefaultFormatScanRec;


PROCEDURE GetPadStr (len :  StrRange) : Str;
(* returns a string with len default padding characters *)

BEGIN
  RETURN GetPadStrOfChar (len, padChar)
END GetPadStr;


PROCEDURE GetPadStrOfChar (len :  StrRange; ch : CHAR) : Str;
 (* returns a string with len supplied padding characters *)
VAR
  temp : Str;
  count : CARDINAL;

BEGIN
  count := 0;
  IF len > 0 THEN
   REPEAT
    temp [count] := ch;
    count := count + 1;
   UNTIL count = len
  END (* if *);
    temp [count] := CHR (0);
  RETURN temp
END GetPadStrOfChar;


PROCEDURE fetchNumFromFmtStr (str : ARRAY OF CHAR; VAR index : StrRange) : CARDINAL;
(* pulls a string of digits from the input string and converts to a cardinal
   if no digits at location, returns a zero *)
VAR
  res : IOStatusT;
  fCount, sLen : StrRange;
  theCard : CARDINAL;
  numStr : Str;

BEGIN (* fetchNumFromFmtStr ( *)
  sLen := LENGTH (str);
  fCount := 0;
  theCard := 0;
  IF sLen = 0 THEN RETURN 0  END; (*IF*)
   WHILE  (index <=  sLen-1) AND (ISO646.isDigit (str [index]) ) DO (* copy the numeric chars *)
     numStr [fCount ] := str [index];
     fCount := fCount + 1;
     index := index + 1
   END; (* while *)
   index := index - 1; (* set back to point at the last digit if we parsed anything *)
   numStr [fCount] := CHR (0);
   WholeStrs.StrToCard (numStr, theCard, res); 
   RETURN theCard
END fetchNumFromFmtStr;


PROCEDURE fetchStrFromFmtStr (str : ARRAY OF CHAR; VAR index : StrRange) : Str;
(* pulls a single quote delimited string of chars from the input string and returns
   if no alphanumeric chars at location, returns a string of length zero
   Pre: the index is that of the first char after the opening quote *)
VAR
  fCount, sLen : StrRange;
  theStr : Str;

BEGIN (* fetchStrFromFmtStr *)
  fCount := 0;
  sLen := LENGTH (str);
  WHILE  (index <  sLen) AND ISO646.isNotQuote (str [index]) DO (* copy the chars *)
    theStr [fCount] := str [index];
    fCount := fCount + 1;
    index := index + 1
  END (* while *);
  IF fCount # 0 THEN
    theStr [fCount] := CHR (0);
  END; (* if *)
  RETURN theStr
END fetchStrFromFmtStr;


PROCEDURE ScanStrForFormat (fmtStr : ARRAY OF CHAR; steps : ScanSet; validFmt : CharSet; VAR fsr : FormatScanRec);
(* The client procedure need only supply the valid format characters in the string
  and the valid scan flags for the type in question as all the other elements default
  Any non-applicable sections will simply be skipped.
  The procedure returns a format scan record with the defaults changed as specified by the fntStr  *)

 VAR
  fCount, fStrLen : StrRange;
  thisChar, openArrayChar : CHAR;
  adjustedGroupChars : CharSet;
  temp : CARDINAL;
  tempEscStr : Str3;
  thisIsDigit : BOOLEAN;

BEGIN (* ScanStrForFormat *)
  fsr := GetDefaultFormatScanRec(); (* get a clean format scan record with the defaults *)

  fStrLen := LENGTH (fmtStr);
  IF (fStrLen = 0) THEN RETURN END;
  fsr.len := fStrLen;

                  (* step 0 InitDefaults *)
  fCount := 0;
  adjustedGroupChars := ValidGroupChars; (* need a variable; may change *)
  IF (validFmt = ValidByteFmtChars) OR  (validFmt = ValidWordFmtChars) THEN
    fsr.numeralFormat := HexAll;
    fsr.numRequiredIntegralDigits := 2
  END; (* IF (validFmt *)
 IF steps = WordScanSet THEN
   fsr.numeralFormat := HexAll;
   fsr.groupSepChar := SingleQuote;
   fsr.numInGroup := 8;
   fsr.numRequiredIntegralDigits := 8*SYSTEM.TSIZE (UNSAFE.WORD)
 END; (* IF steps *)
 (* note in what follows that the variable thisChar always holds the current char
 being processed from the format string *)

      (* section I -- general formatting options *)
             (* step 1 check for general type -- always a valid  option *)
  thisChar := fmtStr [fCount];
  IF thisChar IN validFmt THEN 

    fsr.normFlag := (thisChar = "N");
    CASE thisChar OF
       "N": (* do nothing, just accept the defaults *)
            (* assumed if omitted *)
    |  "S":
          IF RealScanSet <= steps  THEN (* works for complex too *)
            fsr.numeralFormat := Dec;
            fsr.realFormat := Scientific;
          END;
    |  "F":
          IF RealScanSet <= steps THEN
            fsr.numeralFormat := Dec;
            fsr.realFormat := Fixed
          END;
    |  "E" :
         IF RealScanSet <= steps THEN
            fsr.numeralFormat := Dec;
            fsr.realFormat := Eng
        END;
    |  "L" :
         IF RealScanSet <= steps THEN
            fsr.numeralFormat := Dec;
            fsr.realFormat := Float;
            fCount := fCount + 1;
            IF (fCount = fStrLen) THEN RETURN END;
            temp := fetchNumFromFmtStr (fmtStr, fCount);
            IF temp > MaxDigitsForFloat THEN
              Error (NumIntDigitsForFloatOverflow, fCount); RETURN
            END; (*IF*)
            IF temp # 0 THEN fsr.numIntDigitsForFloat := temp END
        END;
    |  "$" :
        IF (steps = RealScanSet) THEN (* not for complex *)
           fsr.prefixStr := CurrencyChar;
           fsr.numeralFormat := Dec;
           fsr.realFormat := Currency;
           fsr.numDecPlaces := 2
        END;
    |  "b" :
          fsr.numeralFormat := Bin;
          fsr.prefixStr := "0b'";
          fsr.numRequiredIntegralDigits := 1

    |  "x" :
          fsr.numeralFormat := Hex;
          fsr.prefixStr := "0x'";
    |  "u" :
          fsr.numeralFormat := Hex;
          fsr.prefixStr := "0u";
    |  "Q" :
          fsr.prefixStr := SingleQuote;
          fsr.suffixStr := SingleQuote;
    |  "D" :
          fsr.prefixStr := DoubleQuote;
          fsr.suffixStr := DoubleQuote;
    |  SingleQuote, DoubleQuote :
          fsr.quoteChar := thisChar;
    |  "Y" :
          fsr.yes4Bool := TRUE;
    |  "T" :
          fsr.yes4Bool := FALSE;      (* default anyway *)
    |  "d" :
          fsr.numeralFormat := Dec
    |  "C" : (* custom format, needs a custom string *)
          fCount := fCount + 1;
          IF (fCount = fStrLen) THEN RETURN END;
          thisChar := fmtStr [fCount];
          IF ISO646.isQuote (thisChar) THEN
            fCount := fCount + 1;
            IF (fCount = fStrLen) THEN
              Error (NoClosingStrQuote, fCount);
              RETURN
            END;
            Assign (fsr.customFormatStr, fetchStrFromFmtStr (fmtStr, fCount));
            thisChar := fmtStr [fCount];
            IF NOT ISO646.isQuote (thisChar) THEN
              Error (NoClosingStrQuote, fCount);
              RETURN
            END; (* IF NOT ISO646.isQuote *)
            fCount := fCount + 1; (* skip over quote *)
            IF (fCount = fStrLen) THEN RETURN END;
          END; (* IF ISO646.isQuote *)
          thisChar := fmtStr [fCount];
          fCount := fCount - 1 
    END (*. CASE .*);
    fCount := fCount + 1;
    IF (fCount = fStrLen) THEN RETURN END;
    thisChar := fmtStr [fCount]; (* set up for next char *)
  END(*IF thisChar*);  

      (* step 2 we parse for complex chars *)
  IF Complex IN steps THEN
    IF thisChar IN ValidCmplxIMPartChars THEN
      fsr.complexFlag := TRUE;
      CASE thisChar OF
        "i" :
          fsr.complexFormat := i;
      | "j" :
          fsr.complexFormat := j;
      | "r" :
          fsr.complexFormat := Raw;
      ELSE
          fsr.complexFlag := FALSE
      END(*CASE*);
    fCount := fCount + 1;
    IF (fCount = fStrLen) THEN RETURN END;
    thisChar := fmtStr [fCount]
    END(*IF thisChar*)
  END; (* IF Complex *)

    (* finished section I *)
    (* start section II *)
            (* step 3 check for prefix string *)
  IF Prefix IN steps THEN
    IF ISO646.isQuote (thisChar) THEN
      fCount := fCount + 1;
      IF (fCount = fStrLen) THEN
        Error (NoClosingStrQuote, fCount);
        RETURN
      END;
      Concat (fsr.prefixStr,fetchStrFromFmtStr (fmtStr, fCount), fsr.prefixStr);
      thisChar := fmtStr [fCount];
      IF NOT ISO646.isQuote (thisChar) THEN
        Error (NoClosingStrQuote, fCount);
        RETURN
      END; (* IF NOT ISO646.isQuote *)
      fCount := fCount + 1; (* skip over quote *)
      IF (fCount = fStrLen) THEN RETURN END;
    END (* IF ISO646.isQuote *)
  END; (* IF prefix *)

    (* finished section II *)
    (* start section III -- specific formatting options *)
    
  WHILE thisChar IN ValidOptionTriggerChars DO
    (* the options in this section can be entered in any order *)
  
                (* step 4 check for sign *)
    IF Sign IN steps THEN
      IF thisChar IN ValidSignChars THEN
        fsr.reqSign := (thisChar = "+");
        fsr.noSign := (thisChar = '-');
        fCount := fCount + 1;
        IF (fCount = fStrLen) THEN RETURN END;
        thisChar := fmtStr [fCount]
      END(*IF thisChar*)
    END; (*IF Prefix*)

               (* step 5 check for fill chars *)
    IF Fill IN steps THEN
      thisIsDigit := ISO646.isDigit (thisChar);
      IF (thisChar IN ValidFillChars) OR thisIsDigit THEN
        IF thisChar = "*" THEN
          fsr.fillChar :=  "*";
        ELSIF thisChar = "z" THEN
          fsr.fillChar :="0";
        ELSE (* s *) fsr.fillChar := ISO646.SPACE
        END; (* IF thisChar *)
        IF fsr.numDecPlaces > MaxFills THEN
          Error (NumRequiredIntegralDigitsOverflow, fCount); RETURN
        END; (*IF*)
        IF NOT thisIsDigit THEN
          fCount := fCount + 1;
          IF (fCount = fStrLen) THEN RETURN END
        END; (* IF NOT thisIsDigit *)
        fsr.numRequiredIntegralDigits := fetchNumFromFmtStr (fmtStr, fCount);
        fCount := fCount + 1;
        IF (fCount = fStrLen) THEN RETURN END;
        thisChar := fmtStr [fCount];
      END(*IF thisChar*)
    END(* IF fill *);

            (* step 6 check for decimal *)
    IF Decimal IN steps THEN
      IF thisChar IN ValidDecimalChars THEN
        fsr.decChar := thisChar;
        fCount := fCount + 1;
        IF (fCount = fStrLen) THEN RETURN END;
        fsr.numDecPlaces := fetchNumFromFmtStr (fmtStr, fCount);
        IF fsr.numDecPlaces > MaxDecimals THEN
          Error (NumRequiredIntegralDigitsOverflow, fCount); RETURN
        END; (*IF fsr.numDecPlaces*)
        fCount := fCount + 1;
        IF (fCount = fStrLen) THEN RETURN END;
        thisChar := fmtStr [fCount]
      END(*IF thisChar*)
    END(*IF decimal*);

              (* step 7 check for groups *)
    IF (Groups IN steps) THEN
         (* first adjust the permissable chars *)
      IF steps = WordScanSet THEN (* UNSAFE.WORD/UNSAFE.ADDRESS *)
        adjustedGroupChars := adjustedGroupChars - CharSet{ ISO646.SEMICOLON }
      ELSIF steps <=RealScanSet  THEN (* REAL or COMPLEX *)
        adjustedGroupChars := ValidGroupChars - CharSet { ISO646.COMMA, ISO646.COLON }
      ELSIF "S" IN validFmt  THEN (* not real or complex so BITSET *)
         adjustedGroupChars := adjustedGroupChars - CharSet { ISO646.SEMICOLON, ISO646.COLON };
      ELSE (* Whole Type *) adjustedGroupChars := adjustedGroupChars - CharSet { ISO646.PERIOD }
      END(*IF steps = *);

      IF thisChar IN adjustedGroupChars THEN
        IF steps = BitsetScanSet THEN
          fsr.bitsetFormat := Binary;
          fsr.numInGroup := 512; (* big default so nothing provided means the char is a trigger to binary only *)
        END;
        IF thisChar = ISO646.UNDERSCORE THEN
           fsr.groupSepChar :=  ISO646.SPACE;
        ELSIF (Decimal IN steps) AND (thisChar = ISO646.SEMICOLON) THEN (* for reals *)
          IF fsr.decChar = ISO646.COMMA THEN
            fsr.groupSepChar := ISO646.PERIOD
          ELSE
            fsr.groupSepChar := ISO646.COMMA
          END; (*IF fsr.decChar*)
        ELSIF steps = WholeScanSet THEN
          IF thisChar = ISO646.COLON THEN
            fsr.groupSepChar := ISO646.PERIOD
          ELSE (* its a semicolon *)
            fsr.groupSepChar := ISO646.COMMA
          END; (*IF IF thisChar = ISO646.COLON*)
        ELSE fsr.groupSepChar := thisChar;
        END; (* thisChar = ISO646.UNDERSCORE *)
        fCount := fCount + 1;
        IF (fCount = fStrLen) THEN RETURN END;
        thisChar :=  fmtStr [fCount];

              (* step 7a for raw output check for group size following the group trigger *)
        IF (validFmt = ValidBitsetFmtChars) OR (validFmt = ValidWordFmtChars) OR (validFmt = ValidAddressFmtChars)
          OR (fsr.numeralFormat = Bin) OR (fsr.numeralFormat = Hex) THEN
        (* otherwise the default is 3 *)
          temp := fetchNumFromFmtStr (fmtStr, fCount);
          IF temp # 0 THEN fsr.numInGroup := temp END;
          thisChar :=  fmtStr [fCount];
          IF fsr.numInGroup > MaxInGroup THEN
            Error (NumRequiredIntegralDigitsOverflow, fCount); RETURN
          END; (*IF fsr.numInGroup *)
          IF (fsr.numeralFormat = Bin) THEN
            fsr.prefixStr := "0b "
          ELSIF (fsr.numeralFormat = Hex)   THEN
            fsr.prefixStr := "0x "
          END; (* IF fsr. *)
          fCount := fCount + 1;
          IF (fCount = fStrLen) THEN RETURN END;
          thisChar :=  fmtStr [fCount];
        END (* if validFmt Done looking for group size*)
      END(*IF thisChar finished with any grouping char *)
    END(*IF groups Done this whole section *);

              (* step 8 check for align chars *)
    IF Align IN steps THEN
      IF thisChar IN ValidAlignChars THEN
        CASE thisChar OF
           "<" : fsr.just := Left
        |  "^" : fsr.just := Full;
        |  ">"  : fsr.just := Right
        END;
        fCount := fCount + 1;
        IF (fCount = fStrLen) THEN RETURN END;
        fsr.fieldLen := fetchNumFromFmtStr (fmtStr, fCount);
         IF fsr.fieldLen > MaxFieldWidth THEN
          Error (NumRequiredIntegralDigitsOverflow, fCount); RETURN
        END; (*IF fsr.numInGroup *)
        fCount := fCount + 1;
        IF (fCount = fStrLen) THEN RETURN END;
        thisChar := fmtStr [fCount];
      END(*IF thisChar*)
    END(*IF align*);


               (* step 9 check for concat -- always valid *)
    IF thisChar IN ValidConcatChars THEN
      fsr.isConcat := TRUE;
      fCount := fCount + 1;
      IF (fCount = fStrLen) THEN RETURN END;
      thisChar := fmtStr [fCount]
    END (* thisChar IN ValidConcatChars *);

              (* step 9a check for column options always valid *)
    IF thisChar IN validColChars THEN
      fsr.isColumnFormat := TRUE;
      fsr.arrayItemSeperator := ""; (* no separators when in columns *)
      fCount := fCount + 1;
      IF (fCount = fStrLen) THEN RETURN END;
      temp := fetchNumFromFmtStr (fmtStr, fCount);
      IF temp # 0 THEN fsr.numCols := temp END; (* if zero leave as default *)
      IF fsr.numCols > MaxCols THEN
        Error (NumRequiredIntegralDigitsOverflow, fCount); RETURN
      END(* IF fsr *);
      fCount := fCount + 1;
      IF (fCount = fStrLen) THEN RETURN END;
      thisChar := fmtStr [fCount];
  END(*IF thisChar IN*);

END; (* WHILE thisChar IN ValidOptionTriggerChars *)
    (* finished section III *)
    (* start section IV -- suffix *)


            (* step 10 check for suffix string *)
  IF Suffix IN steps THEN
    IF ISO646.isQuote (thisChar) THEN
      fCount := fCount + 1;
      IF (fCount = fStrLen) THEN
        Error (NoClosingStrQuote, fCount);
        RETURN
      END;
      Concat (fsr.suffixStr, fsr.suffixStr, fetchStrFromFmtStr (fmtStr, fCount));
      thisChar := fmtStr [fCount];
      IF NOT ISO646.isQuote (thisChar) THEN
        Error (NoClosingStrQuote, fCount);
        RETURN
      END; (* IF NOT ISO646.isQuote *)
      fCount := fCount + 1; (* skip over quote *)
      IF (fCount = fStrLen) THEN RETURN END;
      thisChar := fmtStr [fCount]
    END (* IF ISO646.isQuote *)
  END(*IF suffix*);

    (* finished section IV *)
    (* start section V -- array formatting options *)

              (* step 11 check for array formatting options--always valid except for strings and chars *)
  IF Array IN steps THEN
    IF thisChar = ArrayOptionStart THEN
      fsr.isArrayFormat := TRUE;
      fCount := fCount + 1;
      IF (fCount = fStrLen) THEN RETURN END;
      thisChar := fmtStr [fCount]; (*fetch first char after the array trigger *)

                  (* step 11a check for open char / set enclosure option *)
      IF thisChar IN ValidArrayOpenChars THEN
        openArrayChar := thisChar;
        CASE openArrayChar OF
          ISO646.TILDE :
            fsr.arrayOpenStr := NullStr;
            fsr.arrayCloseStr := NullStr;
         |  "(" :
            fsr.arrayOpenStr := "( ";
            fsr.arrayCloseStr := " )";
         |  "[" :
            fsr.arrayOpenStr := "[ ";
            fsr.arrayCloseStr := " ]";
         |  "{" :
            fsr.arrayOpenStr := "{ ";
            fsr.arrayCloseStr := " }"
        END (* case *);
        fCount := fCount + 1;
        IF (fCount = fStrLen) THEN
          IF (openArrayChar = ISO646.TILDE) THEN RETURN (* ok to have nothing more *)
            ELSE (* should be more *)
              Error (ArrayEnclosureMismatch, fCount); RETURN
            END (* IF (openArrayChar  *)
        END; (* IF (fCount = fStrLen) *)
        thisChar := fmtStr [fCount]
      END; (*IF thisChar IN ValidArrayOpenChars*)

                   (* step 11b check for separator options  *)
      IF thisChar IN ValidArraySeparatorChars THEN
        CASE thisChar OF
           ISO646.COMMA :
             fsr.arrayItemSeperator := CommaSeparator
        |  ISO646.SEMICOLON :
             fsr.arrayItemSeperator := SemicolonSeparator;
        |  ISO646.SPACE :
             fsr.arrayItemSeperator := SpcSeparator
        END; (* case *)
        fCount := fCount + 1;
        IF (fCount = fStrLen) THEN Error (ArrayEnclosureMismatch, fCount); RETURN END;
        thisChar := fmtStr [fCount];
      END; (*IF thisChar*)

                  (* step 11c check for correct close enclosure option  *)
      IF (thisChar IN ValidArrayCloseChars) THEN (* check mismatches *)
        IF ( (thisChar = ")" ) AND NOT (openArrayChar = "(" ) )
           OR ( (thisChar = "]" ) AND NOT (openArrayChar = "[" ) )
           OR ( (thisChar = "}" ) AND NOT (openArrayChar = "{" ) )
          THEN Error (ArrayEnclosureMismatch, fCount); RETURN
        END (* IF (thisChar IN ValidArrayCloseChars) *)
      END(* 12.4 first IF thisChar*);

          (* step 11d now, if we got here, we got all correct array options, so check on state of string scan *)
      IF (openArrayChar # ISO646.TILDE) AND NOT (thisChar = EscTrigger) THEN (* there could be more *)
        fCount := fCount + 1;
        IF (fCount = fStrLen) THEN RETURN END;
        thisChar := fmtStr [fCount]
      END (*  IF (openArrayChar *)
    END (* IF thisChar = ArrayOptionStart *)
  END; (* if array in steps section 11 *)

    (* finished section V *)
    (* start section VI -- terminal escape options *)

                 (* step 12 end of string check for esc sequence and extra chars *)
  IF thisChar = EscTrigger THEN
    fCount := fCount + 1;
    IF (fCount = fStrLen) THEN Error (missingEscChar, fCount); RETURN END;
    thisChar := fmtStr [fCount];
    IF thisChar IN ValidEscChars THEN
     fsr.terminalEscStr [0] := EscTrigger;
     fsr.terminalEscStr [1] := thisChar;
     fsr.terminalEscStr [2] := ISO646.NUL;
    ELSE (* it cannot be anything else *)
      Error (invalidEscChar, fCount); RETURN
    END; (* IF thisChar IN ValidEscChars *)
  ELSE
    Error (extraChars, fCount); RETURN
  END (* IF thisChar = EscTrigger *);

    (* finished section VI *)
    (* start section VII -- More chars? *)

    (* Step 13 and finally one more check to see whether after that there are still more chars *)
  fCount := fCount + 1;
  IF (fCount # fStrLen) THEN Error (extraChars, fCount); RETURN END
END ScanStrForFormat;


PROCEDURE Error (theErr : FmtRes; pos : CARDINAL);

BEGIN (* Error *)
  Perror.prerr ( "Format String error: ");
  CASE theErr OF
    StrOK:
      Perror.prerr ( " string OK.");
    | InvalidChar :
      Perror.prerr ( " invalid Character ");
    | NoFormatChar :
      Perror.prerr ( " non-empty string has no format character ");
    | FLenOverflow :
      Perror.prerr ( " field Length too large ");
    | NumRequiredIntegralDigitsOverflow :
      Perror.prerr ( " too many required digits specified ");
    | NumIntDigitsForFloatOverflow :
      Perror.prerr ( " too many integral digits specified for float format ");
    | NumColsOverflow :
      Perror.prerr ( " too many columns specified ");
    | NoClosingStrQuote :
      Perror.prerr ( " no closing single quote ");
    | ArrayEnclosureMismatch:
      Perror.prerr ( " array enclosure pair does not match ");
    | extraChars :
      Perror.prerr ( " extra characters left at end ");
    | missingEscChar :
      Perror.prerr ( " missing escaped character after escape trigger ");
    | invalidEscChar :
      Perror.prerr ( " invalid escaped character after escape trigger ")
  END(*CASE*);

  Perror.prerr ("error at string position ");
  Perror.prerrnum ( VAL (INTEGER, pos) );
  HALT;
END Error;


(* ===================================== *)

    (* PART THREE--The formatters: feed items in, get formatted versions back *)
     (* Section ONE Utility Procs *)


PROCEDURE CountDigitsInString (str : ARRAY OF CHAR) : CARDINAL;
(* pre: the string ends in a string of digits
        any sign is skipped
   post : returns how many consecutive hex digits
         stops when it reaches a non-digit, so stops at a decimal too
*)
(* this proc is not exported *)
VAR
  len, count, numIntegralDigits : CARDINAL;

BEGIN
  numIntegralDigits := 0;
  count := 0;
  len := LENGTH (str);
  IF len = 0 THEN RETURN 0 END;
  IF str[0] IN CharSet { "-" , "+", "*", " " } THEN count := count + 1 END; (* skip sign and fill *)
  REPEAT
    IF ISO646.isHexDigit ( str [count] ) THEN
      numIntegralDigits := numIntegralDigits + 1
    ELSE
      RETURN numIntegralDigits
    END; (*IF*)
  count := count + 1;
  UNTIL count = len;
  RETURN numIntegralDigits
END CountDigitsInString;


PROCEDURE RemoveLeadingSpaces (VAR str : ARRAY OF CHAR);
(* this proc is exported *)
VAR
  count0, countS, last : CARDINAL;

BEGIN
  count0 := 0;
  countS := 0;
  last := LENGTH ( str) - 1;
  IF str [count0] # " " THEN
    RETURN (* nothing to do *)
  ELSE
    REPEAT
      count0 := count0 + 1
    UNTIL str [count0] # " ";
    WHILE (count0 <= last) DO
      str [countS] := str [count0];
      count0 := count0 + 1;
      countS := countS + 1
    END; (*WHILE*)
    str [countS] := ISO646.NUL
  END (*IF*)
END RemoveLeadingSpaces;


PROCEDURE RemoveLeadingZeros (VAR str : ARRAY OF CHAR; minWidth : CARDINAL);
(* remove all leading zeros, but leave the last minWidth digits *)

VAR
  count0, countS, length, last : CARDINAL;

BEGIN
  count0 := 0; (* use to count to first non "0" position *)
  countS := 0;
  length := LENGTH ( str);
  IF length <= 1 THEN RETURN END; (* nothing to do *)
  last := length - 1;
  IF minWidth = 0 THEN
    minWidth := 1
  ELSIF minWidth > length THEN
    minWidth := length
  END; (* IF minWidth = 0 *)
  IF (str [count0] # "0") OR (length = 1) THEN
    RETURN (* nothing to do *)
  ELSE
    REPEAT
      count0 := count0 + 1
    UNTIL (str [count0] # "0") OR (count0 = last - minWidth );
    IF str[count0] = "." THEN DEC (count0) END; (* leave one leading zero *)
    WHILE (count0 <= last) DO
     str [countS] := str [count0];
     count0 := count0 + 1;
     countS := countS + 1
    END(* WHILE *);
    str [countS] := ISO646.NUL
  END (* str [count0] # "0" *)
END RemoveLeadingZeros;


PROCEDURE RemoveLeadingExtras (VAR str : ARRAY OF CHAR);
(* this proc is exported *)
VAR
  countO, countS, last : CARDINAL;

BEGIN
  countO := 0;
  countS := 0;
  last := LENGTH ( str) - 1;
  IF NOT (str [0] IN RemoveChars) 
        OR ( (str [0] = "0" ) AND (last = 0) ) THEN
    RETURN (* nothing to do and no nul to worry about *)
  ELSE
    REPEAT
      countO := countO + 1
    UNTIL NOT(str [countO] IN RemoveChars);
    WHILE (countO <= last) DO
      str [countS] := str [countO];
      countO := countO + 1;
      countS := countS + 1
    END (*WHILE*);
    str [countS] := ISO646.NUL (* there is room if we remved something *)
  END(* IF NOT (str [0] *)
END RemoveLeadingExtras;


PROCEDURE RemoveGroupingChars (VAR str : ARRAY OF CHAR);
(* this proc is exported *)
VAR
  counti, counto, last : CARDINAL;
  dirty : BOOLEAN;
BEGIN
  counti := 0;
  counto := 0;
  dirty := FALSE;
  last := LENGTH ( str ) - 1;
  WHILE (counti <= last) DO
    IF ~(str[counti] IN NumStrGroupSepChars) THEN
      str [counto] := str[counti];
      counto := counto + 1
    ELSE
      dirty := TRUE; (* skipped a char so take note *)
    END; (* if ~(str[counti] *)
    counti := counti + 1;
  END (*WHILE*);
    (* now in some versions there does not have to be a nul char, and if the string was full
    as it would be for a literal, then we can't add a nul *)
    IF dirty THEN str [counto] := ISO646.NUL END 
END RemoveGroupingChars;


PROCEDURE GroupTextStr (inputStr : ARRAY OF CHAR; sep : CHAR; gLen : CARDINAL; VAR outputStr : ARRAY OF CHAR);
(* this proc is exported *)
VAR
  inCount, outCount, gCount, inLen, digitCount, skipCount : CARDINAL;
  thisChar, nextChar : CHAR;
  skipSet1, skipSet2 : CharSet;
  skip : BOOLEAN;

BEGIN
  skipSet1 := CharSet{"+", "-", "$", "*", " "};
  skipSet2 := CharSet{".", "," ,"e"};
  inCount := 0;
  outCount := 0;
  skipCount := 0;
  inLen := LENGTH (inputStr);
  digitCount := CountDigitsInString (inputStr);
  IF inLen <= gLen THEN RETURN END; (* no grouping to do *)
  REPEAT
    (* we skip over certain lead chars, just copying them and ignoring them for grouping purposes *)
    thisChar := inputStr [inCount];
    skip := thisChar IN skipSet1;
    IF skip THEN
      outputStr [outCount] := thisChar;
      inCount := inCount + 1;
      outCount := outCount + 1;
      skipCount := skipCount + 1
    END; (*IF skip*)
  UNTIL NOT skip;
  (* inCount now points at the first integral digit *)

  gCount :=  gLen - (digitCount MOD gLen);
  REPEAT
    (* fetch current char *)
    thisChar := inputStr [inCount];
    (* peek at next *)
    nextChar := inputStr [inCount+1];
    skip := nextChar IN skipSet2;
    IF skip THEN (* on last integer char *)
      outputStr [outCount] := thisChar; (* copy the current char *)
      inCount := inCount + 1; (* and go to the next *)
      outCount := outCount + 1;
      gCount := 0;
      IF ( nextChar = "e" ) OR ( nextChar = "E" ) THEN
        sep := ISO646.NUL; (* no grouping after exponent *)
      ELSE
        sep := ISO646.SPACE (* commas and periods in decimal part become spaces *)
      END; (* IF nextChar *)
    ELSE (* not a skip so we have another digit coming *)
      gCount := gCount + 1; (* so increment the group counter *)
    END(* IF skip *);

    outputStr [outCount] := inputStr [inCount];
    inCount := inCount + 1;

    IF gLen # 0 THEN gCount := gCount MOD gLen END;

    IF (gCount = 0) AND (inCount # inLen) AND (sep # ISO646.NUL) AND (gLen # 0) AND NOT skip THEN
      outCount := outCount + 1;
      outputStr [outCount] := sep
    END; (* IF (gCount = 0) *)
    outCount := outCount + 1;
  UNTIL inCount = inLen;
  outputStr [outCount] := CHR (0);
END GroupTextStr;


PROCEDURE  Justify (inputStr : ARRAY OF CHAR; fieldLen : StrRange; just : Justification) : Str;
(* exported proc *)
(* pads out the string on the left, right, or both to achieve the required justification. For full, spaces on left <= spaces on right. *)

VAR
  temp, lPad, rPad :Str;
  inputLen, lPadLen : CARDINAL;

BEGIN
  inputLen := LENGTH (inputStr);
  IF fieldLen <=  inputLen THEN
    ARRAYOFCHAR.Assign ( temp, inputStr );
    RETURN  temp (* no padding needed *)
  ELSE
    CASE  just OF
    Left:
        rPad := GetPadStr (fieldLen-inputLen);
        Concat (temp, inputStr, rPad );
    | Right:
        lPad := GetPadStr (fieldLen-inputLen);
        Concat (temp, lPad, inputStr);
    | Full:
        lPadLen := (fieldLen-inputLen) DIV 2; (* shorter if different*)
        lPad := GetPadStr (lPadLen);
        Concat (lPad, inputStr, temp);
        rPad := GetPadStr (fieldLen - lPadLen);
        Concat (temp, rPad, temp)
     END; (* case *)
     RETURN temp
  END(*IF*)
END Justify;

PROCEDURE JustifyC (inputStr : ARRAY OF CHAR; padChar : CHAR; fieldLen : StrRange; just : Justification) : Str;
(* pads out the string on the left, right, or both to achieve the required justification.
 For full, spaces on left <= spaces on right. Uses the provided pad Char *)

VAR 
  temp : CHAR;
  tStr : Str;
  
BEGIN
  temp := GetPadChar ();
  SetPadChar (padChar);
  tStr := Justify (inputStr, fieldLen, just);
  SetPadChar (temp);
  RETURN tStr
END JustifyC;
  
PROCEDURE JustifyA (VAR str : ARRAY OF CHAR; inputStr : ARRAY OF CHAR; padChar : CHAR; fieldLen : StrRange; just : Justification);
(* pads out the string on the left, right, or both to achieve the required justification.
 For full, spaces on left <= spaces on right. Uses the provided pad Char *)
VAR 
  temp : CHAR;
  tStr : Str;
  
BEGIN
  temp := GetPadChar ();
  SetPadChar (padChar);
  tStr := Justify (inputStr, fieldLen, just);
  SetPadChar (temp);
  Assign ( str, tStr )
END JustifyA;


               (* Section TWO Format a string that represents a number entity *)


PROCEDURE FormatNumStr (inputStr : ARRAY OF CHAR; VAR fsr : FormatScanRec) : Str;

VAR
  temp , pad, modInputStr : Str;
  count, tcount, lPad, rPad,
  numFillChars, numActualDigits, totalPad, spaceNeeded, lenPrefix, lenSuffix : CARDINAL;
  sgnChar, thisChar : CHAR;
  isAlreadySigned : BOOLEAN;

BEGIN
  count := 0;
  tcount := 0;
  numActualDigits := 0;
  numFillChars := 0;
  totalPad := 0;
  lPad := 0;
  rPad := 0;
  sgnChar := ISO646.NUL;
  modInputStr := strNulls;
  temp := modInputStr;
  lenPrefix  := LENGTH (fsr.prefixStr);
  lenSuffix := LENGTH (fsr.suffixStr);
 (* now we put the string together according to the flags and rules collected into the fsr we were passed *)

      (* 1. check on the sign situation *)
 isAlreadySigned := (inputStr[0]= "-")  OR (inputStr[0]= "+");
  IF isAlreadySigned THEN
    sgnChar := inputStr[0]
  (* not already signed, so certainly not negative so  *)
  ELSIF fsr.reqSign THEN
    sgnChar := "+";
  ELSIF fsr.noSign  THEN
    sgnChar := " "
  END; (*IF isAlreadySigned*)

      (* 2. remove non-digit chars from the string start including the sign and zeros if appropriate . Put back later. *)
  RemoveLeadingExtras (inputStr);
  IF (fsr.fillChar = ISO646.NUL) AND (fsr.prefixStr[0] # "0" ) (* not removed if format is raw *) THEN
    RemoveLeadingZeros (inputStr, 0)
  END;

      (*3. Adjust the decimal char *)
  count := 0;
  WHILE count < LENGTH (inputStr) DO
   IF inputStr [count] IN ValidDecimalChars THEN
     inputStr [count] := fsr.decChar
   END; (* IF inputStr *)
   count := count + 1
  END; (* while *)
  count := 0;

     (* 4. catinate the fill chars to this string of digits and put into a second variable *)
  numActualDigits  := CountDigitsInString (inputStr);
  IF numActualDigits < fsr.numRequiredIntegralDigits THEN
   numFillChars := fsr.numRequiredIntegralDigits - numActualDigits
  END; (* IF numActualDigits *)
       (* If the length of a group is zero, all the following will do is copy the string, which we want  anyway *)
   Concat (modInputStr, GetPadStrOfChar (numFillChars, fsr.fillChar), inputStr);
  IF NOT (fsr.numInGroup = 0) AND NOT (fsr.groupSepChar = ISO646.NUL) THEN
    GroupTextStr (modInputStr, fsr.groupSepChar, fsr.numInGroup, modInputStr)
  END; (* IF NOT (fsr.numInGroup = 0) *)

     (* 5. now work out the padding needed from the field given and the space needed *)
  spaceNeeded := LENGTH (modInputStr); (* need to know length so far, including separator chars *)
  IF isAlreadySigned OR ((fsr.reqSign OR fsr.noSign) AND ( fsr.fieldLen > 0 )) THEN
     spaceNeeded := spaceNeeded + 1; (* one more space needed for the sign *)
  END; (* IF isAlreadySigned *)
  spaceNeeded := spaceNeeded + lenPrefix + lenSuffix;

  IF spaceNeeded < fsr.fieldLen THEN totalPad := fsr.fieldLen - spaceNeeded END;
  IF fsr.just =  Left THEN
    rPad := totalPad
  ELSIF fsr.just = Full THEN
    rPad := totalPad DIV 2
  END; (* IF fsr.just *)
  lPad := totalPad - rPad;

                (* put the rest of the  string together  in temp *)
       (* 6. put in the left pad *)
  IF lPad > 0 THEN Concat (temp, GetPadStr (lPad), temp) END;
  tcount := LENGTH (temp); (* continue where we just left off *)

       (* 7. put in the sign if needed *)
  IF sgnChar # ISO646.NUL THEN
    temp [tcount] := sgnChar;
    tcount := tcount + 1
  END; (* IF sgnChar # *) (* set start for passed in string *)

       (* 8. put in any prefix such as 0b  or 0x  or a user supplied one *)
  IF lenPrefix # 0 THEN Concat (temp, temp, fsr. prefixStr ) END;
  tcount := LENGTH (temp);

       (* 9. now put in the possibly modified passed string *)
  WHILE (count < spaceNeeded) DO
    thisChar := modInputStr [count];
    temp [tcount] := thisChar;
    count := count + 1;
    tcount := tcount + 1;
  END; (*WHILE*)

      (* 10. put in any suffix *)
  IF lenSuffix # 0 THEN Concat (temp, temp, fsr.suffixStr ) END;

    (* 11. put in any padding and terminate *)
  Concat (temp, temp, GetPadStr (rPad));
  Concat (temp, temp, NullStr);
  RETURN temp
END FormatNumStr;


PROCEDURE FormatTextStr ( inputStr : ARRAY OF CHAR; VAR fsr : FormatScanRec) : Str;
(* takes an input string and a format record , and returns the string
formatted according to the rules indicated by the format string *)
VAR
  temp : Str;
  pad : Str;
  count, tcount, lPad, rPad, len,
  totalPad, spaceNeeded, lenPrefix, lenSuffix : CARDINAL;

BEGIN
  temp := strNulls;
  len := LENGTH ( inputStr );
  count := 0;
  totalPad := 0;
  lPad := 0;
  rPad := 0;
  lenPrefix  := LENGTH (fsr.prefixStr);
  lenSuffix := LENGTH (fsr.suffixStr);
  spaceNeeded := LENGTH (inputStr); (* need to know length so far, including separator chars *)
  spaceNeeded := spaceNeeded + lenPrefix + lenSuffix;
  IF spaceNeeded < fsr.fieldLen THEN totalPad := fsr.fieldLen - spaceNeeded END;
  IF fsr.just =  Left THEN
    rPad := totalPad
  ELSIF fsr.just = Full THEN
    rPad := totalPad DIV 2
  END; (* IF fsr.just *)
  lPad := totalPad - rPad;
                (* put the rest of the  string together  in temp *)
  (* put in the left pad *)
  IF lPad > 0 THEN Concat (temp, GetPadStr (lPad), temp) END;
  tcount := LENGTH (temp); (* continue where we just left off *)

  (*  put in any prefix such as 0u or quotes  or a user supplied one *)
  IF lenPrefix # 0 THEN Concat (temp, temp, fsr. prefixStr ) END;
  tcount := LENGTH (temp);

  (* now put in the passed string *)
  WHILE (count < len) DO
    temp [tcount] := inputStr [count];
    count := count + 1;
    tcount := tcount + 1
  END; (*WHILE*)

    (*  put in any suffix such as quotes *)
  IF lenSuffix # 0 THEN Concat (temp, temp, fsr. suffixStr ) END;

  (* put in any padding and terminate *)
  Concat (temp, temp, GetPadStr (rPad));
  Concat (temp, temp, NullStr);
  RETURN temp
END FormatTextStr;


VAR
  padChar : CHAR;

BEGIN (* main module *)
  SetPadChar ( DefPadChar );
  CharArray.ZeroOutCharArray (strNulls);
  SetCustomScanSet (DefaultCustomScanSet);
  SetValidCustomFmtChars (DefaultValidCustomFmtChars);
  
END FormatIO.
(* sCardinalIO.Write (fCount); sCharIO.WriteStr(fsr.customFormatStr);sCharIO.WriteStr ("\n"); *)