IMPLEMENTATION MODULE LongWholeStrs;

  (* =========================================
          Implementation  1993
            by R. Sutcliffe
       (Portions coded by G. Tischer)
        Trinity Western University
7600 Glover Rd., Langley, BC Canada V3A 6H4
         e-mail: rsutc@twu.ca
    Last modification date 2017 11 02
 ===========================================*)

(* 1993 10 20 first cut as LongWholeStrs
1994 02 14 RS fixed bugs in indexing of strings pointed out by Norm Black; must check high first
cleaned up Gord's code closer to our standards
1994 05 18 first cut at a p1 version
  str := "0" changed to str [0] := "0"
1994 06 20 add to this str [1] := CHR (0) to correct new bug introduced
1994 09 03 WARNING: a bug in p1 prevents MIN(INTEGER) from being used in expressions and will in that case crash IntToDecStr
        Here it has been fixed to detect this and return the correct string.
    IntToDecStr and IntToCard improved in the process.
1996 08 08 some extranneous fossil local vars removed from IntToDecStr and CardToDecStr
1996 12 03 tracked down a nasty bug in p1 translation that prevented conversions to strings from working 
2017 many changes to get it working with R10 library 
     Now depends on a LONGCARD type of the same bitwidth as LONGINT
     2017 11 02 Name changed to WholeStrss to avoid conflict with p1 debug library*)

  (* Whole-number/string conversions *)

(* the string form of a signed whole number is
     ["+" | "-"], decimal digit, {decimal digit}
*)
IMPORT SYSTEM, UNSAFE;
FROM Transition IMPORT CAPACITYch;

IMPORT BaseConv, ISO646, FormatIO;
IMPORT Card8Strs, LongCardinalMath, LongIntegerMath, ARRAYOFCHAR, Bitwidth, CharArray;
IMPORT CARD8; FROM CARD8 IMPORT CARD8T;
IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;

TYPE
  CardType = (* ALIAS OF *) LONGCARD;
  IntType = (* ALIAS OF *) LONGINT;
  
CONST
  zeroAsc = ORD ("0");
  lastMinIntDigitChar = CHR ((ABS (MIN (LONGINT) + 10)) REM 10 + VAL (LONGINT, zeroAsc));

PROCEDURE DecStrToCard (str: ARRAY OF CHAR; VAR card: LONGCARD; VAR res : IOStatusT);
  (* Ignores any leading spaces in str. If the subsequent characters in str are
   in decimal format of an unsigned whole number, assigns a corresponding value to the single CARD8.
   Assigns a value indicating the format of str to res.  *)

VAR
  accum, room, addend, multiplier : LONGCARD;
  numDigits, index, base, stop : LONGCARD;

BEGIN (* DecStrToCardinal *)
  FormatIO.RemoveLeadingExtras (str);
  FormatIO.RemoveGroupingChars(str);

IF (str[0] IN FormatIO.NumStrSignChars) THEN stop := 1 ELSE stop := 0 END;
  base := 10;
  numDigits := LENGTH (str);
  multiplier := 1;
  index := numDigits;
  accum := 0;
  room := MAX (LONGCARD);
  REPEAT 
    index := index - 1;
    IF ( ORD (str [index]) >= ORD ("0") ) AND ( ORD (str [index]) <= ORD ("9") ) THEN
      addend := ( ORD (str [index]) - ORD ("0") ) * multiplier; 
      IF addend <= room THEN
        room := room - addend;
        accum := accum + addend;
        multiplier := multiplier * base;
      ELSE
        card := 0;
        res := IOStatusT{ TRUE, IOStatus.OutOfRange};
        RETURN
      END (* if addend *)
    ELSE
      res := IOStatusT{ TRUE, IOStatus.WrongFormat};
      RETURN
    END; (* IF ORD *) 
  UNTIL index = stop;
  card := accum;
  res := IOStatusT{ FALSE, IOStatus.Success};;
  RETURN;
END DecStrToCard;

PROCEDURE HexStrToCard (str: ARRAY OF CHAR; VAR card: LONGCARD; VAR res : IOStatusT);

VAR
  accum, room, addend, multiplier : LONGCARD;
  numDigits, index, base : LONGCARD;

BEGIN (* HexStrToCard8 *)
  FormatIO.RemoveLeadingSpaces (str);
  IF (str[0] # "0") OR (str [1] # "x") THEN
    card := 0;
    res := IOStatusT{ TRUE, IOStatus.WrongFormat};
    RETURN
  END;
  
  FormatIO.RemoveLeadingExtras (str);
  FormatIO.RemoveGroupingChars(str);
  base := 16;
  numDigits := LENGTH (str);
  IF numDigits > 2*(Bitwidth.CardOctetwidth) THEN
    res := IOStatusT{ TRUE, IOStatus.OutOfRange};;
    RETURN
  END;
  multiplier := 1;
  index := numDigits;
  accum := 0;
  room := MAX (LONGCARD);
  REPEAT 
    index := index - 1;
    IF ( ORD (str [index]) >= ORD ("0") ) AND ( ORD (str [index]) <= ORD ("9") ) THEN
      addend := ( ORD (str [index]) - ORD ("0") ) * multiplier;
    ELSIF ( ORD (str [index]) >= ORD ("A") ) AND ( ORD (str [index]) <= ORD ("F") ) THEN
      addend := ( ORD (str [index]) - ORD ("A") + 10) * multiplier;
    ELSE
      res := IOStatusT{ TRUE, IOStatus.WrongFormat};
      RETURN
    END; (* IF ORD *) 
    IF addend <= room THEN
      room := room - addend;
      accum := accum + addend;
      multiplier := multiplier * base;
    ELSE
      card := 0;
      res := IOStatusT{ TRUE, IOStatus.OutOfRange};
      RETURN
    END (* if addend *)
  UNTIL index = 0;
  card := accum;
  res := IOStatusT{ FALSE, IOStatus.Success};
  RETURN;
END HexStrToCard;

PROCEDURE BinStrToCard (str: ARRAY OF CHAR; VAR card: LONGCARD; VAR res : IOStatusT);
  (* Ignores any leading spaces in str. If the subsequent characters in str are
   in binary format of an unsigned whole number, assigns a corresponding value to the single CARD8.
   Assigns a value indicating the format of str to res.  *)

VAR
  accum, room, addend, multiplier : LONGCARD;
  numDigits, index, base : LONGCARD;

BEGIN (* BinStrToCard8 *)
  FormatIO.RemoveLeadingSpaces (str);
  IF (str[0] # "0") OR (str [1] # "b") THEN
    card := 0;
    res := IOStatusT{ TRUE, IOStatus.WrongFormat};
    RETURN
  END;
  FormatIO.RemoveLeadingExtras (str);
  FormatIO.RemoveGroupingChars(str);
  base := 2;
  numDigits := LENGTH (str);
  multiplier := 1;
  index := numDigits;
  accum := 0;
  room := MAX (LONGCARD);
  REPEAT 
    index := index - 1;
    IF ( ORD (str [index]) >= ORD ("0") ) AND ( ORD (str [index]) <= ORD ("1") ) THEN
      addend := ( ORD (str [index]) - ORD ("0") ) * multiplier;
      IF addend <= room THEN
        room := room - addend;
        accum := accum + addend;
        multiplier := multiplier * base;
      ELSE
        card := 0;
        res := IOStatusT{ TRUE, IOStatus.OutOfRange};
        RETURN
      END (* if addend *)
    ELSE
      res := IOStatusT{ TRUE, IOStatus.WrongFormat};
      RETURN
    END; (* IF ORD *) 
  UNTIL index = 0;
  card := accum;
  res := IOStatusT{ FALSE, IOStatus.Success};;
  RETURN;
END BinStrToCard;

PROCEDURE StrToCard (str: ARRAY OF CHAR; VAR card: LONGCARD; VAR res: IOStatusT);
  (* Ignores any leading spaces in str. If the subsequent characters in str are in the format of an unsigned whole number, assigns a corresponding value to card. Assigns a value indicating the format of str to res.  *)

BEGIN
  FormatIO.RemoveLeadingSpaces (str);
  IF (str[0] = "0")
    THEN
      IF (str [1] = "x")
        THEN
          HexStrToCard (str, card, res);
          RETURN
        ELSIF str [1] = "b" THEN
          BinStrToCard (str, card, res);
          RETURN
       END;
     ELSE (* is decimal string *)
        DecStrToCard (str, card, res);
     END(*IF*)
END StrToCard;

PROCEDURE DecStrToInt (str: ARRAY OF CHAR; VAR int: LONGINT; VAR res : IOStatusT);

VAR
  isNeg : BOOLEAN;
  tempCard : LONGCARD;
  
BEGIN
  FormatIO.RemoveLeadingSpaces (str);
  int := 0;
  isNeg := (str [0] = "-"); (* there should be nothing before a sign *)
  DecStrToCard (str, tempCard, res); (* sign will be ignored there *)
  IF ~isNeg THEN
    IF (tempCard <= ABS (MAX(LONGINT)) ) THEN 
      int := tempCard;
    ELSE
      int := 0;
      res := IOStatusT{ TRUE, IOStatus.OutOfRange};
      RETURN
    END
  ELSIF (* negative *) (tempCard < ABS (MIN(LONGINT)) ) THEN
    int := tempCard;
    int := -int
  ELSIF (tempCard = ABS (MIN(LONGINT)) ) THEN(* is MIN (LONGINT) *)
    int := MIN (LONGINT)
  ELSE
    int := 0;
    res := IOStatusT{ TRUE, IOStatus.OutOfRange};
    RETURN
  END (* IF ~isNeg *)

END DecStrToInt;


PROCEDURE HexStrToInt (str: ARRAY OF CHAR; VAR int: LONGINT; VAR res : IOStatusT);
VAR
  tempCard : LONGCARD;

BEGIN  
  HexStrToCard (str, tempCard, res);
  IF NOT res.failed THEN int := SYSTEM.CAST (LONGCARD, tempCard) END;
END HexStrToInt;


PROCEDURE BinStrToInt (str: ARRAY OF CHAR; VAR int: LONGINT; VAR res : IOStatusT);

VAR
  tempCard : LONGCARD;

BEGIN
  BinStrToCard (str, tempCard, res);
  IF NOT res.failed THEN int := SYSTEM.CAST (LONGCARD, tempCard) END;
END BinStrToInt;


PROCEDURE StrToInt (str: ARRAY OF CHAR; VAR int: LONGINT; VAR res : IOStatusT);
  (* Ignores any leading spaces in str. If the subsequent characters in str are in the format
   of a signed whole number, assigns a corresponding value to int.
    Assigns a value indicating the format of str to res.  *)

VAR
  isNeg : BOOLEAN;

BEGIN
  int := 0;
  FormatIO.RemoveLeadingSpaces (str);
  IF (str[0] = "0") THEN
    IF (str [1] = "x") THEN
      HexStrToInt (str, int, res);
      RETURN
    ELSIF str [1] = "b" THEN
      BinStrToInt (str, int, res);
      RETURN
    END; (* IF (str [1] *)
  ELSE (* is in decimal form *)
    DecStrToInt (str, int, res);
  END(* IF str[0]*)
END StrToInt;


TYPE
  CardinalConvArray = ARRAY [0..Bitwidth.CardOctetwidth-1] OF CARD8T;

PROCEDURE CardToDecStr (card: LONGCARD; VAR str: ARRAY OF CHAR);
  (* Converts the value of card to string form and copies the possibly truncated result to str. *)
VAR
  maxs, maxi, count, digit : LONGCARD;

BEGIN
  maxs := CAPACITYch (str) - 1;
  count := 0;
  WHILE count <= maxs - 1 DO
    str [count] := CHR(0);
    count := count + 1
  END (* while *);
  maxi := LongCardinalMath.nearestLargerPowerOf10 (card);
  IF maxs > maxi
    THEN
      count := maxi;
      str [count] := "";
    ELSE
      count := maxs;
      (* if string is too short, throw away some digits *)
      WHILE maxi > count
        DO
          card := card / 10;
          DEC (maxi)
        END
    END (* if *);
  str [count] := CHR(0);
  WHILE count > 0
    DO
      digit := card REM 10;
      str [count - 1] := CHR (digit + zeroAsc);
      card := card / 10;
      DEC (count)
    END
END CardToDecStr;

PROCEDURE CardToHexStr (card: LONGCARD; VAR str: ARRAY OF CHAR);
  (* Converts the value of card to string form and copies the possibly truncated
  result to str. *)

VAR
  cArray : BaseConv.ConvArray;
  count : LONGCARD;
  card8HexStr : Card8Strs.Card8HexStr;

BEGIN
  BaseConv.LongCardToConvArray (card, 256, cArray );
  count := BaseConv.COUNTca (cArray);
  CharArray.ZeroOutCharArray (str);
  WHILE count # 0 DO 
    count := count - 1;
    Card8Strs.Card8ToHexStr ( cArray.card8s[count], card8HexStr);
    ARRAYOFCHAR.Concat (str, str, card8HexStr)
  END;
END CardToHexStr;


PROCEDURE CardToBinStr (card: LONGCARD; VAR str: ARRAY OF CHAR);
  (* Converts the value of card to string form and copies the possibly truncated
  result to str. *)
VAR 
  bitCount, strCount : CARDINAL;
  
BEGIN
  bitCount := Bitwidth.CardBitwidth;
  strCount := 0;
  WHILE bitCount # 0 DO 
    bitCount := bitCount - 1;
    IF UNSAFE.BIT (card, bitCount) THEN
      str [strCount] := "1"
    ELSE
      str [strCount] := "0"
    END;
   strCount := strCount + 1;
  END;
  str [strCount] := ISO646.NUL;
END CardToBinStr;


PROCEDURE IntToDecStr (int: LONGINT; VAR str: ARRAY OF CHAR);
  (* Converts the value of int to string form and copies the possibly truncated result to str. *)

(* the string form of an unsigned whole number is
     decimal digit, {decimal digit} *)
VAR
  maxs, maxi, count, digit, stop : LONGCARD;
  isNeg : BOOLEAN;

BEGIN
  isNeg := (int < 0);
  maxs := CAPACITYch (str) - 1;
  maxi := LongIntegerMath.LengthInt (int); (* includes any neg *)
  IF maxs > maxi
     THEN
      count := maxi;
      str [count] := "";
    ELSE
      count := maxs;
      (* if string is too short, throw away some digits *)
      WHILE maxi > count
         DO
          int := int / 10;
          DEC (maxi)
        END
    END;
  IF isNeg
    THEN
      IF int = MIN (LONGINT) (* special case this one, cant do abs else *)
        THEN
          str [count - 1] := lastMinIntDigitChar;
          DEC (count);
          int := int / 10
        END;
      stop := 1;
      str [0] := "-";
      int := ABS (int);
    ELSE
      stop := 0
    END;
  WHILE count > stop
    DO
      digit := int REM 10;
      str [count - 1] := CHR (digit + zeroAsc);
      int := int / 10;
      DEC (count)
    END
END IntToDecStr;


PROCEDURE IntToHexStr (int: LONGINT; VAR str: ARRAY OF CHAR);
  (* Converts the value of int to string form and copies the possibly truncated
  result to str. We're just putting out raw hex here *)
VAR
  card : LONGCARD;

BEGIN
  card := SYSTEM.CAST (LONGCARD, int);
  CardToHexStr (card, str)
  
END IntToHexStr;

PROCEDURE IntToBinStr (int: LONGINT; VAR str: ARRAY OF CHAR);
  (* Converts the value of int to string form and copies the possibly truncated
  result to str. We're just putting out raw Binary here.*)

VAR
  card : LONGCARD;

BEGIN
  card := SYSTEM.CAST (LONGCARD, int);
  CardToBinStr (card,str)
  
END IntToBinStr;

END LongWholeStrs.
