(* (C) 2016 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE FileDescIO;

(* Driver for File IO on File Descriptors *)

(* This module interprets the contents of files as a stream of octets  without
   any particular structure.  This is a low level API.  For common use,  it is
   recommended to use the channel based API of module FileIO instead. *)

IMPORT LongCardBaseFDIO; (* Diagnostic purposes only *)

IMPORT SYSTEM, UNSAFE;
FROM Transition IMPORT OCTET, CAPACITYo, IncAdr;
FROM INTEGER32 IMPORT INTEGER32;

IMPORT Perror;
IMPORT ISO646, IOParams;
IMPORT PtrTable;
IMPORT UnixLib, errno;

IMPORT IOSIZE; FROM IOSIZE IMPORT IOSIZET;
IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;
IMPORT FileMode; FROM FileMode IMPORT FileModeT;
IMPORT Buffer; FROM Buffer IMPORT BufferT;

FROM Storage IMPORT ALLOCATE, DEALLOCATE;
IMPORT ARRAYOFCHAR, CharArray;

(* setup material *)
VAR
  descPtrTable : PtrTable.PtrTable;

CONST
  InsertBuffMax = IOParams.InsertBufferSize - 1;
  NulOctet = SYSTEM.CAST ( OCTET, ISO646.NUL );

TYPE
  InsertBuffPos = [0..InsertBuffMax];
  DString = ARRAY [0 .. 255] OF CHAR;
  CurrentOp = ( reading, writing );  (* internal use only *)
  FilePtr = POINTER TO FileData;
  FileData = RECORD
    desc : File;
    mode : FileModeT;
    name : DString;
    status : IOStatusT;

    insertBuffer : ARRAY [0.. InsertBuffMax] OF OCTET;
    iBuffIn, iBuffOut, iBuffCount : InsertBuffPos;
   
    buffer : BufferT;
    buffAdr : UNSAFE.ADDRESS;
    buffSize : Pos;
    curOp : CurrentOp
  END(* FileData *);


VAR
  defaultFileDataTable : FileData;
  stdIn, stdOut, stdErr : File;
  stdInPtr, stdOutPtr, StdErrPtr : FilePtr;
(*
(* local debugging procedure *)

CONST
 maxBDCalls = 200;
VAR
  callCount : CARDINAL;
  diag : BOOLEAN;
  

PROCEDURE BufferDiagnostic ( fd : File ; msg : ARRAY OF CHAR; num : IOSIZET);

VAR
  count : IOSIZET;
  c : CARDINAL;
  adr : UNSAFE.ADDRESS;
  ptr : FilePtr;
 
BEGIN 
  IF NOT diag THEN RETURN END;

  IF isSpecialFile (fd ) THEN RETURN END;
  ptr := PtrTable.GetPtr ( descPtrTable, fd );
  count := 0;
  adr :=  ptr^.buffAdr;
  WriteOctet (stdErr, CHR(10));
  WriteOctet (stdErr, "%");
  c := 0;
  WHILE (c <= HIGH (msg) ) AND (msg [c] <> CHR (0) ) DO
     WriteOctet (stdErr, msg [c]);
     c := c + 1;
  END;
  WriteOctet (stdErr, "%");
  WriteOctet (stdErr, " ");
  
  WriteOctet (stdErr, "B");
  WriteOctet (stdErr, "-");

  LongCardBaseFDIO.Write ( stdErr, Buffer.currentPos (ptr^.buffer) );
  WriteOctet (stdErr, "/");
  LongCardBaseFDIO.Write ( stdErr, (Buffer.offset (ptr^.buffer)) );
  WriteOctet (stdErr, "/");
  LongCardBaseFDIO.Write ( stdErr, Buffer.count (ptr^.buffer) );
  WriteOctet (stdErr, "/");
  LongCardBaseFDIO.Write ( stdErr, Buffer.lastValidPos (ptr^.buffer) );
  WriteOctet (stdErr, "/");
  WriteBlock (stdErr,"hasUnreadData= ", c);
  IF Buffer.hasUnreadData (ptr^.buffer) THEN
    WriteOctet (stdErr, "T");
  ELSE
    WriteOctet (stdErr, "F");
  END;
  WriteOctet (stdErr, " ");

  WriteBlock (stdErr,"isDirty= ", c);
  IF Buffer.isDirty (ptr^.buffer) THEN
    WriteOctet (stdErr, "T");
  ELSE
    WriteOctet (stdErr, "F");
  END;
  WriteOctet (stdErr, " ");

  WriteOctet (stdErr, "F");
  WriteOctet (stdErr, "-");
  LongCardBaseFDIO.Write ( stdErr, currentPos ( fd ) );
  WriteOctet (stdErr, "/");
  WriteOctet (stdErr, "P");
  LongCardBaseFDIO.Write ( stdErr,  UnixLib.lseek ( fd, 0, UnixLib.fromCurPos ) );
  WriteOctet (stdErr, "/");
  LongCardBaseFDIO.Write ( stdErr, lastValidSetPos (fd ) );

  WriteOctet (stdErr, " ");

  IF Buffer.count (ptr^.buffer) = 0 THEN WriteOctet (stdErr, "~") END;
  WHILE count < num DO
    WriteOctet (stdErr, adr^);
    count := count + 1;
    IncAdr ( adr )
  END;
  WriteOctet (stdErr, "%");
  WriteOctet (stdErr, CHR(10));
  
  callCount := callCount + 1;
  IF callCount >= maxBDCalls THEN HALT END
END BufferDiagnostic;

*)

PROCEDURE InitModule;

BEGIN
  defaultFileDataTable :=
  FileData {
    InvalidFile,
    FileModeT{ }, (* empty mode is invalid *)
    "", (* empty string for name *)
    IOStatusT { TRUE, IOStatus.Failure },
    {NulOctet, NulOctet}, (* zeros in insertBuffer *)
    0,0,0, (* insert buff pos initial values *)
    Buffer.InvalidBuffer(), (* initial buffer *)
    NIL, (* initial file buffer address *)
    0,    (* file buffer initial value for size *)
    reading
  };

  PtrTable.New (descPtrTable);

  stdIn := UnixLib.uStdIn;
  NEW (stdInPtr); (* get a pointer *)
  (* and entable it *)
  PtrTable.EnTable (descPtrTable, stdIn, stdInPtr);
  stdInPtr^ := defaultFileDataTable;
  stdInPtr^.desc := stdIn;
  stdInPtr^.mode := FileMode.RO;
  stdInPtr^.status := IOStatusT{ FALSE, IOStatus.Success};
 
  stdOut := UnixLib.uStdOut;
  NEW (stdOutPtr);
  PtrTable.EnTable (descPtrTable, stdOut, stdOutPtr);
  stdOutPtr^ := defaultFileDataTable;
  stdOutPtr^.desc := stdOut;
  stdOutPtr^.mode := FileMode.WO + FileModeT{ FileMode.Sync };
  stdOutPtr^.status := IOStatusT{ FALSE, IOStatus.Success};
 
  stdErr := UnixLib.uStdErr;
  NEW (StdErrPtr);
  PtrTable.EnTable (descPtrTable, stdErr, StdErrPtr);
  StdErrPtr^ := defaultFileDataTable;
  StdErrPtr^.desc := stdErr;
  StdErrPtr^.mode := FileMode.WO + FileModeT{ FileMode.Sync };
  StdErrPtr^.status := IOStatusT{ FALSE, IOStatus.Success};
END InitModule;

(* end setup material *)

(* utility procs *)


PROCEDURE CStatusToIOStatus ( ) :IOStatusT;
(* Pre : An error in the C file system access is known to have taken place
   and no other access has been undertaken since.
   Post : Caller decides what to do with the returned status *)
VAR
  errPtr : errno.INT32Ptr;
  errnum : errno.SysErrNum;

BEGIN
    errPtr := errno.__error ();
    errnum := errPtr^;
    CASE errnum OF
      errno.EOF :
        RETURN IOStatusT{ TRUE, IOStatus.AccessBeyondEOF};
    | errno.ENOERR :
        RETURN IOStatusT{ FALSE, IOStatus.Success};
    | errno.ENOENT :
        RETURN IOStatusT{ TRUE, IOStatus.FileNotFound};
    | errno.EIO :
        RETURN IOStatusT{ TRUE, IOStatus.Failure};
    | errno.ENXIO :
        RETURN IOStatusT{ TRUE, IOStatus.DeviceError};
    | errno.EBADF :
        RETURN IOStatusT{ TRUE, IOStatus.InvalidFile};
    | errno.ENOMEM :
        RETURN IOStatusT{ TRUE, IOStatus.AllocationFailed};
    | errno.EPERM, errno.EACCES:
        RETURN IOStatusT{ TRUE, IOStatus.AccessDenied};
    | errno.EROFS:
        RETURN IOStatusT{ TRUE, IOStatus.NotOpenForReading};
    | errno.EISDIR:
        RETURN IOStatusT{ TRUE, IOStatus.MayNotOpenDirectory};
    | errno.EEXIST:
        RETURN IOStatusT{ TRUE, IOStatus.AlreadyExists};
    | errno.ENFILE, errno.EMFILE :
        RETURN IOStatusT{ TRUE, IOStatus.OpenFileLimitExceeded};
    | errno.ENOSPC:
        RETURN IOStatusT{ TRUE, IOStatus.DeviceFull};
    | errno.ESPIPE:
        RETURN IOStatusT{ TRUE, IOStatus.InvalidFilePos};
    | errno.ENODEV :
        RETURN IOStatusT{ TRUE, IOStatus.OperationNotSupported};
    ELSE
        RETURN IOStatusT{ TRUE, IOStatus.Failure};
    END (*CASE*);
END CStatusToIOStatus;


PROCEDURE LookupFilePtr (fd : File) : FilePtr;

BEGIN
  IF (isValidAccessor ( fd ) ) THEN
  RETURN PtrTable.GetPtr (descPtrTable, fd);
  ELSE
    RETURN NIL
  END (* IF isValidAccessor *)
END LookupFilePtr;


PROCEDURE SetUpMainBuffer (ptr : FilePtr; buffSize : IOSIZET );

BEGIN
  Buffer.New (ptr^.buffer, buffSize);
  ptr^.buffAdr := Buffer.BufferAdr (ptr^.buffer);
  ptr^.buffSize :=  buffSize;
  Buffer.SetOffset (ptr^.buffer, currentPos (ptr^.desc) );
END SetUpMainBuffer;

(* end utility procs *)

(* ================================= *)

(* Resource Allocation *)


PROCEDURE CreateAndOpen ( (*NEW*)VAR fd       : File;
                 (*. CONST .*) filename : ARRAY OF CHAR;
                 mode           : FileModeT;
                 VAR status     : IOStatusT );
(* Creates and opens the file <filename> in file mode <mode>. If the file already
   exists, nothing happens and status is set to AlreadyExists. If sucessful passes
   a file accessor back in <file>,  otherwise passes NIL back in <file>.  Sets the
   file position depending on <mode>.  The status of the operation is passed
   back in <status>. *)
VAR
  filePtr : FilePtr;
  err : INTEGER32;
  errPtr: errno.INT32Ptr;
  oflag: UnixLib.OFlagSet;

BEGIN
  IF UnixLib.access ( filename, UnixLib.F_OK) = 0 THEN (* file is already there *)
    status := IOStatusT{ TRUE, IOStatus.AlreadyExists};
    RETURN;
  END; (* IF UnixLib.access *)

(* now check to ensure the mode is valid *)
  IF ( NOT FileMode.isValid (mode) ) OR (NOT (FileMode.Write IN mode) )  THEN
   status := IOStatusT{ TRUE, IOStatus.InvalidMode};
   fd := InvalidFile;
   RETURN
  END; (* NOT isValid (mode) *) 
  
  fd := UnixLib.creat (filename, UnixLib.s_irwugo );
  IF fd # InvalidFile THEN
    err := UnixLib.close (fd);
    Open (fd, filename, mode, status);
    RETURN
  END;
    
(*  fd := UnixLib.open (filename, UnixLib.oWrite + UnixLib.oCreate);*)
  status := CStatusToIOStatus ( );
  IF fd # InvalidFile THEN status.code := IOStatus.Success; 
     
      ReOpen (fd, mode, status);
    END (* IF status.code *)
  (* otherwise the status returned is the one from the attempt to re open *)

END CreateAndOpen;


PROCEDURE Open
  ( VAR fd         : File;
    (*CONST*) filename : ARRAY OF CHAR;
    mode           : FileModeT;
    VAR status     : IOStatusT );
(* Opens file <filename> in access mode <mode>.  On success, a file descriptor
   is passed back  in <fd>.  On failure, InvalidFile  is passed back  in <fd>.
   Sets the file position depending on <mode>.  The status of the operation is
   passed back in <status>. *)

BEGIN
  OpenWithBufferSize ( fd, filename, mode, IOParams.DefaultBufferSize, status )
END Open;


PROCEDURE OpenWithBufferSize
  ( (*NEW*) VAR fd         : File;
    (*CONST*) filename   : ARRAY OF CHAR;
    mode             : FileModeT;
    buffSize         : IOSIZET;
    VAR status       : IOStatusT );

VAR
  filePtr : FilePtr;
  err : INTEGER32;
  errPtr: errno.INT32Ptr;
  oflag: UnixLib.OFlagSet;

BEGIN
  (* check for too many files already open *)
  IF PtrTable.isFull (descPtrTable)  THEN
    status := IOStatusT{ TRUE, IOStatus.OpenFileLimitExceeded};
    RETURN
  END; (* IF *)

  (* check the mode and set up the UNIX mode *)

  IF mode = FileMode.RO
   THEN
     oflag := UnixLib.oRead
   ELSIF mode = FileMode.WO THEN
     oflag := UnixLib.oWrite
   ELSIF mode = FileMode.RW THEN
     oflag := UnixLib.oReadWrite;
   ELSIF FileMode.Append IN mode THEN
     oflag := UnixLib.oAppend + UnixLib.oWrite
   ELSE
     status := IOStatusT{ TRUE, IOStatus.InvalidMode};
     fd := InvalidFile;
   RETURN
  END; (* IF mode *) 
  
  (* now try to open the unix file *)
  fd := UnixLib.open (filename, oflag);
  IF fd = InvalidFile THEN
    status := CStatusToIOStatus ( );
    RETURN;
  END; (* IF fd *)

      (* got this far, so it is now open as a Unix file *)
  status := IOStatusT{ FALSE, IOStatus.Success};
  NEW (filePtr); (* get a pointer *)
  (* and entable it *)
  PtrTable.EnTable (descPtrTable, fd, filePtr);

    (* next set all the local data up properly *)
      (* set up the data table *)
  filePtr^ := defaultFileDataTable;
  filePtr^.desc := fd;
  filePtr^.mode := mode;
  ARRAYOFCHAR.Concat ( filePtr^.name, filePtr ^.name, filename); (* starts as "" *)
  filePtr^.status := status;
  (* insert buffer is reset by default *)
  (* next set up buffer *)
  IF buffSize < 6 THEN
    buffSize := IOParams.MinBufferSize
  END; (* IF buffSize *) 
  SetUpMainBuffer ( filePtr, buffSize );
  IF FileMode.Append IN mode THEN
    Buffer.SetOffset (filePtr^.buffer, lastValidSetPos ( fd ) );
  END; (* IF FileMode *)
END OpenWithBufferSize;


PROCEDURE ReOpen ( fd : File; mode : FileModeT; VAR status: IOStatusT );

VAR
  ptr : FilePtr;
  filename : DString;

BEGIN
  IF isSpecialFile ( fd ) THEN
    status := IOStatusT{ TRUE, IOStatus.OperationNotSupported};
    RETURN;
  END; (* IF isSpecialFile *)
  IF NOT FileMode.isValid (mode) THEN
    status := IOStatusT{ TRUE, IOStatus.InvalidMode};
    RETURN
  END; (* IF NOT FileMode *)
  ptr := PtrTable.GetPtr (descPtrTable, fd );
  IF ptr = NIL THEN
    status := IOStatusT{ TRUE, IOStatus.OperationNotSupported};
    RETURN
  END; (* IF ptr *)
 
  (* All checks out so get the filename *)
  filename := ptr^.name;
  Close (fd, status);
  Open (fd, filename, mode, status);
END ReOpen;


(* Introspection *)


PROCEDURE modeOf ( fd : File ) : FileModeT;
(* Returns the access mode of <fd>. *)
VAR
  ptr : FilePtr;

BEGIN
    ptr := PtrTable.GetPtr ( descPtrTable, fd );
    RETURN ptr^.mode
END modeOf;


PROCEDURE nameLen ( fd : File ) : CARDINAL;
(* Returns the length of the filename associated with <file>. *)
VAR
  ptr : FilePtr;

BEGIN
  ptr := PtrTable.GetPtr ( descPtrTable, fd );
  IF ( NOT isValidAccessor ( fd ) ) OR ( isSpecialFile ( fd ) ) THEN
    RETURN 0
  ELSE
    RETURN LENGTH (ptr^.name)
  END; (* IF NOT isValidAccessor *)
END nameLen;


PROCEDURE GetName ( fd : File; VAR filename : ARRAY OF CHAR );
(* Passes the name  of the file  associated with  file accessor <file> back in
   <filename>. If the name exceeds the capacity of <filename>, an empty string
   is passed back in <filename>.  Does not alter the status of <file>. *)
VAR
  ptr : FilePtr;

BEGIN
  ptr := PtrTable.GetPtr ( descPtrTable, fd );
  filename[0] := CHR(0);
  IF NOT isValidAccessor( fd ) OR (nameLen( fd ) > CAPACITYo (filename) - 2 ) OR ( isSpecialFile ( fd ) ) THEN RETURN
    ELSE
    ARRAYOFCHAR.Concat ( filename, filename, ptr ^.name );
  END; (* IF NOT isValidAccessor *)
END GetName;


PROCEDURE statusOf ( fd : File ) : IOStatusT;
(* Returns the status of the most recent operation on <fd>. *)
VAR
  ptr : FilePtr;

BEGIN
  IF NOT isValidAccessor (fd) THEN
    RETURN IOStatusT {TRUE, IOStatus.InvalidFile }
  END; (* if *)
  IF isSpecialFile ( fd ) THEN
    RETURN IOStatusT {FALSE, IOStatus.Success }
  ELSE
    ptr := PtrTable.GetPtr ( descPtrTable, fd );
    RETURN ptr^.status
  END; (* IF isSpecialFile *)
END statusOf;


PROCEDURE SetStatus
  ( fd : File; status : IOStatusT; VAR valid : BOOLEAN );
(* Sets the status of <file>  to <status> if it meets the integrity condition.
   Where <status.code> is set to Success <status.failed> must be FALSE, other-
   it must be TRUE.  Passes TRUE in <valid> if successful, otherwise FALSE. *)

VAR
  ptr : FilePtr;

BEGIN
  IF NOT isValidAccessor (fd) THEN
    valid := FALSE;
    RETURN
  END; (* IF NOT isValidAccessor (fd) *)
  valid := (status.code = IOStatus.Success) AND NOT (status.failed)
       OR (status.code # IOStatus.Success) AND (status.failed);
  IF valid THEN  
    ptr := PtrTable.GetPtr ( descPtrTable, fd );
    ptr^.status := status;
  END; (* IF valid *)

END SetStatus;


PROCEDURE StatusMsg ( fd : File; status : IOStatusT );
(* Writes a status message for <status> to <fd>. *)
VAR
  msg : IOStatus.statusMsg;
  count : IOSIZET;

BEGIN
  IOStatus.fetchMsg (status, msg);
  count := 0;
  WHILE (count < CAPACITYo (msg) ) AND ( msg [count] #  ISO646.NUL ) DO
    WriteOctet (fd, msg [count] );
    count := count + 1
  END; (* WHILE count *)
  WriteOctet ( fd, ISO646.LF )
END StatusMsg;


PROCEDURE isValidAccessor ( fd : File ) : BOOLEAN;
(* Returns TRUE if <fd> is a valid file descriptor, otherwise FALSE. *)

BEGIN
  IF isSpecialFile ( fd ) THEN
    RETURN TRUE;
  ELSIF (fd = InvalidFile) OR (PtrTable.GetPtr (descPtrTable, fd) = NIL) THEN
    RETURN FALSE
  ELSE
    RETURN TRUE
  END; (* IF isSpecialFile *)
END isValidAccessor;


PROCEDURE isSpecialFile ( fd : File ) : BOOLEAN;
(* Returns TRUE if <fd> is associated with a special file, otherwise FALSE. *)

BEGIN
  RETURN (fd  = stdIn) OR (fd = stdOut) OR (fd = stdErr)
END isSpecialFile;

(* I/O *)

(* local procedure so validity of file already checked, and can be called by ptr *)
PROCEDURE ResetInsertBuffer ( fd : File );

VAR
  ptr : FilePtr;

BEGIN
  ptr := PtrTable.GetPtr ( descPtrTable, fd );
  ptr^.iBuffIn := 0;
  ptr^.iBuffOut := 0;
  ptr^.iBuffCount := 0;
END ResetInsertBuffer;

(* Any attempt to perform the below I/O operations on any file or stream whose
   read flag is not set shall fail with status NotOpenForReading. *)

PROCEDURE insertReady ( fd : File ) : BOOLEAN;
(* Returns FALSE if the insert buffer of <file> is full, otherwise TRUE. *)
VAR
  ptr : FilePtr;

BEGIN
  ptr := PtrTable.GetPtr ( descPtrTable, fd );
  IF ptr = NIL THEN RETURN FALSE END;
  IF NOT (FileMode.Read IN ptr^.mode ) THEN (* also covers specials *)
    ptr^.status := IOStatusT{ TRUE, IOStatus.NotOpenForReading };
    RETURN FALSE
  ELSE
    RETURN (ptr^.iBuffCount < InsertBuffMax);
  END (* IF NOT FileMode *)
END insertReady;


PROCEDURE Insert ( fd : File; data : OCTET );
(* Inserts octet <data> into input stream <stream> to be read by the next read
   operation on <stream>.  Inserted octets are stored in a FIFO buffer of size
   InsertBufferSize.  The minimum size of the insert buffer is two octets. *)

VAR
  ptr : FilePtr;

BEGIN
  ptr := PtrTable.GetPtr ( descPtrTable, fd );
  IF ptr = NIL THEN RETURN END;
  IF NOT (FileMode.Read IN ptr^.mode ) THEN
    ptr^.status := IOStatusT{ TRUE, IOStatus.NotOpenForReading };
  RETURN
  ELSIF NOT (insertReady ( fd ) ) THEN
    ptr^.status := IOStatusT{ TRUE, IOStatus.InsertBufferFull };
    RETURN;
  ELSE
    ptr^.insertBuffer [ptr^.iBuffIn] := data;
    IF ptr^.iBuffIn = InsertBuffMax THEN
      ptr^.iBuffIn := 0;
    ELSE
      ptr^.iBuffIn := ptr^.iBuffIn + 1
    END; (* IF ptr *)
    ptr^.iBuffCount := ptr^.iBuffCount + 1;
    ptr^.status := IOStatusT{ FALSE, IOStatus.Success }
  END; (* IF NOT FileMode *)
END Insert;


PROCEDURE Retrieve ( ptr : FilePtr; VAR data : OCTET );
(* Pre: Since this is internal, and we only get here from a read operaton,
   read mode and the validity of the accessor have already been checked.
   The caller shall also check whether
   the insert buffert has anything in it before calling this procedure.
   Post: removes the next available octet from the insert buffer of file (if any)
  and returns it in <data>, adjusting the buffer parameters as needed *)
(* local; called by above or by read after checking to see if needed *)

BEGIN
  data := ptr^.insertBuffer [ptr^.iBuffOut];
  IF ptr^.iBuffOut = InsertBuffMax THEN
    ptr^.iBuffOut := 0;
  ELSE
    ptr^.iBuffOut := ptr^.iBuffOut + 1
  END; (* IF ptr *)
  ptr^.iBuffCount := ptr^.iBuffCount - 1;
END Retrieve;


PROCEDURE FileToBuffer ( fd : File; VAR numRead : IOSIZET );
(* if Dirty, flushes the file. Reads as much as possible from the
  current position of the physical file and writes to the specified buffer *)
VAR
  ptr : FilePtr;

BEGIN
  ptr := PtrTable.GetPtr ( descPtrTable, fd );
  IF ptr = NIL THEN RETURN END;
  IF Buffer.isDirty ( ptr^.buffer ) THEN
    Flush ( fd );
  END; (* IF ptr *)
(* the way this is we always read to the beginning of the buffer. *)
  Buffer.Reset ( ptr^.buffer );
  (* with an offset of the file pos when the read starts *)
  Buffer.SetOffset (ptr^.buffer,  UnixLib.lseek ( fd, 0, UnixLib.fromCurPos ) );
  numRead := UnixLib.read (fd, ptr^.buffAdr , ptr^.buffSize );
  IF numRead > 0 THEN
    Buffer.RecordDirectEntry ( ptr^.buffer, 0, numRead );
  END (* IF numRead *)
END FileToBuffer;


PROCEDURE ReadOctet ( fd : File; VAR data : OCTET );
(* Reads an octet from <fd> and passes it back in <data>. *)
VAR
  numRead : IOSIZET;
  filePtr : FilePtr;
  fileBuff : Buffer.BufferT;

BEGIN
  IF NOT isValidAccessor ( fd ) THEN
    Perror.perror ( 'IO error on ReadOctet; invalid file'); RETURN;
  ELSIF NOT (FileMode.Read IN modeOf ( fd ) ) THEN
    Perror.perror ( 'IO error on ReadOctet; not in read mode');
    filePtr^.status := IOStatusT {TRUE, IOStatus.NotOpenForReading };
    RETURN
  END; (* IF NOT isValidAccessor *)
  (* all OK to read so: *)
  filePtr := LookupFilePtr ( fd );

(* now first check to see if data is in the insert buffer *)
  IF filePtr^.iBuffCount > 0 THEN
    Retrieve (filePtr, data);
    RETURN
  ELSIF isSpecialFile( fd ) THEN
    numRead := UnixLib.read (fd, SYSTEM.ADR(data), 1);
  ELSE (* have a regular buffered file *)
    filePtr^.curOp := reading;
       (* so see if it is in the main buffer *)
    fileBuff := filePtr^.buffer;
    IF NOT ( Buffer.hasUnreadData ( fileBuff ) ) THEN (* if not, get some *)
      FileToBuffer ( fd, numRead )  
    END; (* IF NOT ( Buffer.hasUnreadData ( fileBuff ) *)
    IF ( Buffer.hasUnreadData ( fileBuff ) ) AND ( Buffer.isValidReadPos ( fileBuff, Buffer.currentPos (fileBuff ) ) ) THEN
      Buffer.ReadOctet ( fileBuff, data );
      RETURN
    ELSE (* whether reading to buffer was successful or not, we couldn't read our octet *)
      numRead := 0;
    END; (* IF ( Buffer.hasUnreadData ( fileBuff ) ... *)

    IF numRead < 0 THEN
      IF NOT isSpecialFile( fd ) THEN
        filePtr^.status := IOStatusT {TRUE, IOStatus.DeviceError };
        Perror.perror ( 'IO error on ReadOctet, nothing read'); RETURN
      END; (* IF NOT isSpecialFile( fd ) *)
    ELSIF numRead = 0 THEN
      IF NOT isSpecialFile( fd ) THEN
        filePtr^.status := IOStatusT {TRUE, IOStatus.AccessBeyondEOF };
        Perror.perror ( 'IO error on ReadOctet; already at EOF'); RETURN
     END (* IF NOT isSpecialFile( fd ) *)
    END (* IF numRead < 0 *)
  END (* IF filePtr^.iBuffCount > 0 *)
  
END ReadOctet;


PROCEDURE ReadBlock
  ( fd : File; VAR data : ARRAY OF OCTET; VAR octetsRead : IOSIZET );
(* Reads as many octets from <fd> as are available and fit into buffer <data>.
   Reading starts at the current position. The number of octets read is passed
   back in <octetsRead>.  The current position is advanced by <octetsRead>. *)

VAR
  numToRead, numRead, c, dataOffset, totalRead : IOSIZET;
  filePtr : FilePtr;
  fileBuff : Buffer.BufferT;
  startFilePos, startBuffPos : Pos;

BEGIN 
  IF NOT isValidAccessor ( fd ) THEN
    Perror.perror ( 'IO error on Read; invalid file'); RETURN;
  ELSIF NOT (FileMode.Read IN modeOf ( fd ) ) THEN
    Perror.perror ( 'IO error on ReadBlock; not in read mode');
    filePtr^.status := IOStatusT {TRUE, IOStatus.NotOpenForReading };
    RETURN;
  END; (* IF NOT isValidAccessor ( fd ) *)
 
  (* all OK to read so: *)
  numToRead := SIZE(data);
  numRead := 0;
  totalRead := 0;
  IF isSpecialFile( fd ) THEN
    octetsRead := UnixLib.read (fd, SYSTEM.ADR(data), numToRead )
  ELSE
  (*  first dump the insert buffer *)
    ResetInsertBuffer ( fd );
  (* got here so have a regular buffered file *)
       (* so see if any written data is in the main buffer *)
    filePtr := LookupFilePtr ( fd );
    filePtr^.curOp := reading;
    fileBuff := filePtr^.buffer;
    IF Buffer.isDirty ( fileBuff ) THEN Flush ( fd ) END;

    dataOffset := 0;
    totalRead := 0;
    startFilePos := Buffer.offset (fileBuff);
    startBuffPos := Buffer.currentPos (fileBuff );
    IF NOT ( Buffer.hasUnreadData ( fileBuff ) ) THEN (* if not, get some *);
      FileToBuffer ( fd, numRead )
    END; 

     REPEAT 
      Buffer.SetPos ( fileBuff, startBuffPos);
      IF ( Buffer.hasUnreadData ( fileBuff ) ) AND ( Buffer.isValidReadPos ( fileBuff, Buffer.currentPos (fileBuff ) ) ) THEN
        Buffer.ReadBlockToOffset ( filePtr^.buffer, data, dataOffset, octetsRead );
        dataOffset := dataOffset + octetsRead;
        totalRead := totalRead + octetsRead;
        IF NOT ( Buffer.hasUnreadData ( fileBuff ) ) AND ( octetsRead <= numToRead ) THEN
          FileToBuffer ( fd, numRead ); (* read as much as possible *)
        END; 
      END; (* IF Buffer.hasUnreadData *)
      startBuffPos := 0;
    UNTIL ( totalRead = numToRead ) OR eof ( fd )
  END; (* IF isSpecialFile(fd) *)

(* error section *)
    IF octetsRead < 0   THEN
      IF NOT isSpecialFile( fd ) THEN
        filePtr^.status := IOStatusT {TRUE, IOStatus.DeviceError };
        Perror.perror ( 'IO error on ReadOctet, nothing read'); RETURN
      END; (* IF NOT isSpecialFile( fd ) *)
    ELSIF octetsRead = 0 THEN
      IF NOT isSpecialFile( fd ) THEN
        filePtr^.status := IOStatusT {TRUE, IOStatus.AccessBeyondEOF };
        Perror.perror ( 'IO error on ReadBlock; already at EOF'); RETURN
      END (* IF NOT isSpecialFile( fd ) *)
    END (* IF octetsRead < 0 *)
END ReadBlock;


PROCEDURE ReadBlockAtOffset
  ( fd             : File;
    VAR data       : ARRAY OF OCTET;
    VAR octetsRead : IOSIZET;
    offset         : Pos );
(* Reads as many octets from <fd> as are available and fit into buffer <data>.
   Reading starts at position <offset>.  The number of  octets read  is passed
   back in <octetsRead>.  The current position is NOT advanced. *)
VAR
  numRead, oldPos : Pos;
  filePtr : FilePtr;

BEGIN
  oldPos := currentPos (fd);
  filePtr^.curOp := reading;
  SetPos ( fd, offset );
  ReadBlock ( fd, data, octetsRead);
  SetPos ( fd, oldPos );
END ReadBlockAtOffset;


(* Any attempt to perform the below I/O operations on any file or stream whose
   write flag is not set shall fail with status NotOpenForWriting. *)

PROCEDURE WriteOctet ( fd : File; data : OCTET );
(* Writes the octet passed in <data> to <fd>. *)

VAR
  numWritten : INTEGER32;
  filePtr : FilePtr;
  fileBuff : Buffer.BufferT;
  sync : BOOLEAN;
  numRead : Pos;

BEGIN
  IF NOT isValidAccessor ( fd ) THEN
    Perror.perror ( 'IO error on WriteOctet; invalid file'); RETURN;
  ELSIF NOT (FileMode.Write IN modeOf ( fd ) ) THEN
    Perror.perror ( 'IO error on WriteOctet; not in write mode');
    filePtr^.status := IOStatusT {TRUE, IOStatus.NotOpenForWriting };
    RETURN;
  END; (* IF NOT isValidAccessor ( fd ) *)
  (* all OK to write so: check and clear insetBuffer if it is being used *)
  filePtr := LookupFilePtr ( fd );
  IF filePtr^.iBuffCount > 0 THEN
    ResetInsertBuffer ( fd )
  END; (* IF filePtr *)

  sync := FileMode.Sync IN modeOf ( fd );
  IF sync THEN
    numWritten := UnixLib.write (fd, SYSTEM.ADR(data), 1);
  ELSE (* have a regular buffered file *)
    filePtr^.curOp := writing;
       (* so see if we can write to the main buffer *)
    fileBuff := filePtr^.buffer; 
     (* Now, we need to ensure that we are writing into a buffer that is indeed a window into 
       the correct portion of the file. *)    
    IF NOT ( Buffer.hasUnreadData ( fileBuff ) ) THEN (* if not, get some *)
      FileToBuffer ( fd, numRead )  
    END; (* IF NOT ( Buffer.hasUnreadData ( fileBuff ) *)

(* note the logic here. At this point the buffer must have unread data, hence
 the position cannot be at eob, hence there should be a writable pos. 
 If not, there is an error *)
    IF Buffer.isValidWritePos (fileBuff, Buffer.currentPos ( fileBuff ) ) THEN
      Buffer.WriteOctet ( fileBuff, data ); 
      RETURN
    ELSE
      numWritten := -1
    END (* IF Buffer.isValidWritePos *)
  END;(* IF sync *)
  IF numWritten < 0 THEN
    Perror.perror ( 'IO - error on Write'); RETURN;
    IF NOT isSpecialFile( fd ) THEN
      filePtr^.status := IOStatusT {TRUE, IOStatus.AccessBeyondEOF }
    END (* IF numWritten *)
  END; (* IF isSpecialFile *)
END WriteOctet;


PROCEDURE WriteBlock
  ( fd : File; (*CONST*) data : ARRAY OF OCTET; VAR octetsWritten : IOSIZET );
(* Writes buffer <data> to <fd>.  Writing starts at the current position.  The
   number of  octets written  is passed back in <octetsWritten>.  The  current
   position is advanced by <octetsRead>. *)
VAR
  numToWrite, dataOffset : IOSIZET;
  filePtr : FilePtr;
  fileBuff : Buffer.BufferT;
  sync : BOOLEAN;

BEGIN 
  IF NOT isValidAccessor ( fd ) THEN
    Perror.perror ( 'IO error on WriteAtOffset; invalid file'); RETURN;
  ELSIF NOT (FileMode.Write IN modeOf ( fd ) ) THEN
    Perror.perror ( 'IO error on WriteAtOffset; not in write mode');
    filePtr^.status := IOStatusT {TRUE, IOStatus.NotOpenForWriting };
    RETURN;
  END; (* IF NOT isValidAccessor ( fd ) *)
 
    (* all OK to write so: *)
  numToWrite := SIZE(data);
  sync := FileMode.Sync IN modeOf ( fd );

  IF sync THEN
    octetsWritten := UnixLib.write (fd, SYSTEM.ADR(data), numToWrite );
  ELSE 
    (* now first dump the insert buffer *)
    IF FileMode.Read IN modeOf ( fd ) THEN
      ResetInsertBuffer ( fd )
    END; (* IF FileMode.Read *)
    (* have a regular buffered file *)
    
       (* so see if any data is in the main buffer *)
    filePtr := LookupFilePtr ( fd );
    fileBuff := filePtr^.buffer;
    filePtr^.curOp := writing; 
    IF numToWrite > filePtr^.buffSize - Buffer.currentPos (fileBuff) THEN
      IF Buffer.isDirty ( fileBuff ) THEN
        Flush ( fd ) (* resets buffer too *)
      END; (* IF Buffer *)
    END; (* IF numToWrite *)
    dataOffset := 0;
    
    REPEAT 
      Buffer.WriteBlockFromOffset ( filePtr^.buffer, data, dataOffset, octetsWritten );
      (* note that octetsWritten is cumulated by repeated calls *) 
      IF octetsWritten <= numToWrite THEN 
        Flush ( fd );
        dataOffset := octetsWritten
      END;

    UNTIL octetsWritten = numToWrite;
  END; (* IF sync *)

  (* error section *)
  IF octetsWritten < 0 THEN
    Perror.perror ( 'IO - error on Write'); RETURN;
    IF NOT isSpecialFile( fd ) THEN
      filePtr^.status := IOStatusT {TRUE, IOStatus.DeviceError };
      Perror.perror ( 'IO - error on WriteBlock '); RETURN;
    END (* IF NOT isSpecialFile( fd ) *)
    (* ELSE User responsibility if not all octets were written *)
  END; (* IF octetsWritten *)
END WriteBlock;


PROCEDURE WriteBlockAtOffset
  ( fd             : File;
    data           : ARRAY OF OCTET;
    VAR octetsWritten : IOSIZET;
    offset         : Pos );
(* Writes buffer <data>  to <fd>.  Writing  starts  at position <offset>.  The
   number of  octets written  is passed back in <octetsWritten>.  The  current
   position is NOT advanced. On a special file sets OperationNotSupported *)
VAR
  numToWrite, dataOffset : IOSIZET;
  filePtr : FilePtr;
  fileBuff : Buffer.BufferT;
  oldPos, newPos : Pos;
  sync : BOOLEAN;

BEGIN

  IF NOT isValidAccessor ( fd ) THEN
    Perror.perror ( 'IO error on WriteBlockAtOffset; invalid file'); RETURN;
  ELSIF isSpecialFile ( fd ) AND (offset > 0) THEN
    Perror.perror ( 'IO error on WriteBlockAtOffset; not supported');
    filePtr^.status := IOStatusT {TRUE, IOStatus.OperationNotSupported };
    RETURN;
  END;
  oldPos := currentPos (fd);
  filePtr^.curOp := writing;
  SetPos ( fd, offset );
  WriteBlock ( fd, data, octetsWritten);
  SetPos ( fd, oldPos );
  
END WriteBlockAtOffset;


PROCEDURE eof ( fd : File ) : BOOLEAN;
(* Returns TRUE if <fd> has reached the end of the file, otherwise FALSE. *)

BEGIN
  RETURN ( maxPos ( fd ) <= currentPos ( fd ) )
END eof;


PROCEDURE currentPos ( fd : File ) : Pos;

VAR
  filePtr : FilePtr;

BEGIN
  IF NOT isValidAccessor ( fd ) OR isSpecialFile ( fd ) THEN
    RETURN 0
  END; (* IF NOT isValidAccessor ( fd ) *)
  IF FileMode.Sync IN modeOf ( fd ) THEN
    RETURN UnixLib.lseek ( fd, 0, UnixLib.fromCurPos );
  ELSE
    filePtr := LookupFilePtr ( fd );
    RETURN Buffer.offset (filePtr^.buffer) + Buffer.currentPos ( filePtr^.buffer )
  END (* IF FileMode.Sync *)
END currentPos;


PROCEDURE isValidPos ( fd : File; pos : Pos ) : BOOLEAN;
(* Returns TRUE if <pos> is a valid file position for <file>, otherwise FALSE. *)

BEGIN
  IF NOT isValidAccessor ( fd ) OR isSpecialFile( fd ) THEN
    RETURN FALSE
  ELSE
    RETURN pos <= lastValidSetPos ( fd )
  END (* IF NOT isValidAccessor *)
END isValidPos;


PROCEDURE lastValidSetPos ( fd : File ) : Pos;
(* Returns the most advanced position that may be passed to SetPos for file
   accessor <file>.  The return value depends on length and mode of <file>

   (1) if <file> is in append mode, the function returns zero and sets the
       file's status to OperationNotSupported.
   (2) if <file> is empty, then TMIN(Pos) is returned.
   (3) if <file> is not empty and has its write flag set, then the position
       of the last octet in <file> plus one is returned.
   (4) if <file> is not empty and has its read but not its write flag set,
       then the position of the last octet in <file> is returned. *)

VAR
  filePtr  : FilePtr;
  max : Pos;

BEGIN
  IF NOT isValidAccessor ( fd ) OR isSpecialFile ( fd ) THEN RETURN MIN(Pos) END;
  filePtr := LookupFilePtr ( fd ); 
  IF FileMode.Append IN filePtr^.mode THEN
    filePtr^.status := IOStatusT{ TRUE, IOStatus.OperationNotSupported };
    RETURN 0
  END; (* IF isValidAccessor *)
  max := maxPos ( fd );
  IF FileMode.Write IN filePtr^.mode THEN
    filePtr^.status := IOStatusT{ FALSE, IOStatus.Success };
    RETURN max + 1
  ELSIF FileMode.Read IN filePtr^.mode THEN
    filePtr^.status := IOStatusT{ FALSE, IOStatus.Success };
    RETURN max
  ELSE
    (* if we get here, the file's mode is corrupt *)
    RETURN 0
  END (* IF FileMode *)
END lastValidSetPos;


(* Positioning *)

PROCEDURE isPosInBuffer ( ptr : FilePtr; pos : Pos ) : BOOLEAN;

BEGIN
  RETURN ( pos >= Buffer.offset (ptr^.buffer) ) AND ( pos <= Buffer.lastValidPos ( ptr^.buffer) )
END isPosInBuffer;

PROCEDURE SetPos ( fd : File; pos : Pos );
(* Sets the current position of <fd> to <pos>. *)
VAR
  newPos : Pos;

BEGIN
  SetRelPos ( fd, Start, pos, newPos )
END SetPos;


PROCEDURE Advance ( fd : File; offset : Pos );
(* Advances the current position of <fd> by <offset>. *)
VAR
  newPos : Pos;

BEGIN
  SetRelPos ( fd, Current, offset, newPos );
END Advance;


PROCEDURE Rewind ( fd : File );
(* Sets the current position of <fd> to zero and resets its eof status. *)
VAR
  newPos : Pos;

BEGIN
  (* set the position; if the buffer needs flushing, SetRelPos will do that *)
  SetRelPos ( fd, Start, 0, newPos );
  IF newPos # 0 THEN Perror.perror ( 'IO - error on rewind ') END;
END Rewind;


PROCEDURE SetAtmaxPos ( fd : File );
(* Sets the current position of <fd> to the last position in the file *)
VAR
  newPos : Pos;

BEGIN
  SetRelPos ( fd, End, 0, newPos )
END SetAtmaxPos;


PROCEDURE SetRelPos
  ( fd : File; from : FromPosition; offset : RelPos; VAR newPos : Pos );
  (* Sets the current position of <fd> relative to <from> by <offset> and passes
   the resulting absolute position back in <newPos>. *)
VAR
  filePtr  : FilePtr;
  absPos : Pos;
 
BEGIN
  IF NOT isValidAccessor ( fd ) OR isSpecialFile( fd ) THEN RETURN END;
  filePtr := LookupFilePtr ( fd );

  (* calculate the new position *)
  IF filePtr = NIL THEN RETURN END;
 
  (* can't set pos in append mode *)
  IF FileMode.Append IN filePtr^.mode THEN
    filePtr^.status := IOStatusT{ TRUE, IOStatus.OperationNotSupported };
    newPos := currentPos ( fd )
  END; (* IF FileMode.Append *)
  absPos := ABS ( offset );
  
  (* now run through the three possibilities for From Position *)
  IF (from = Start) THEN
    IF (offset >= 0 )THEN
      newPos := absPos;
    ELSE
      newPos := 0
    END; (* IF (offset >= 0 ) *)

  ELSIF (from = End) THEN
    IF (newPos <= 0) THEN
      newPos := lastValidSetPos ( fd ) - absPos;
    ELSE
      newPos := lastValidSetPos ( fd )
    END (*  IF (newPos <= 0) *)

  ELSIF (from = Current ) THEN
    IF offset >= 0 THEN
      newPos :=  currentPos ( fd ) + absPos;
      IF newPos > lastValidSetPos ( fd ) THEN
        newPos := lastValidSetPos ( fd )
      END; (* IF offset >= 0 *)
    ELSE (* offset is negative *)
      IF absPos <= currentPos ( fd ) THEN
        newPos :=  currentPos ( fd ) - absPos
      ELSE
        newPos := 0;
      END (* IF absPos <= currentPos ( fd ) *)
    END (* IF offset >= 0 *)
  END; (* if from *)
  (* now, if valid, set it either in the buffer or in the file *)
  IF isValidPos ( fd, newPos ) THEN
    IF isPosInBuffer ( filePtr, newPos ) THEN
       Buffer.SetPos ( filePtr^.buffer, newPos - Buffer.offset (filePtr^.buffer) )
    ELSE  (* not in buffer so *)
      IF Buffer.isDirty ( filePtr^.buffer ) THEN
        Flush ( fd ) (* also resets the buffer *)
      END; (* IF Buffer *)
      newPos := UnixLib.lseek (fd, newPos, UnixLib.fromStart);
      Buffer.Reset ( filePtr^.buffer );
      Buffer.SetOffset (filePtr^.buffer, newPos )
    END; (* IF isPosInBuffer *)
  ELSE (* not a valid pos *)
    filePtr^.status := IOStatusT{ TRUE, IOStatus.InvalidFilePos };
    Perror.perror ( 'IO error on SetPos; invalid pos'); RETURN;
  END; (* IF isValidPos *)

(* error section; check what happened with that supposedly valid pos *)
  IF newPos < 0 THEN
    Perror.perror( 'error on SetPos'); RETURN;
    filePtr^.status := IOStatusT {TRUE, IOStatus.DeviceError }
  ELSE
    filePtr^.status := IOStatusT {FALSE, IOStatus.Success };
  END(* IF newPos *);

END SetRelPos;

(* a local utility *)


PROCEDURE maxPos ( fd : File ) : Pos;
  (* we do it this way so as not to upset any of our local information
  The physical file's info is temporarily set to the maxPos
  but once it is set back again, nothing here in this module has changed
  so there is no need, for instance, to flush the buffer *)
VAR
  curPos, endPos, posPerBuffer : Pos;
  filePtr  : FilePtr;

BEGIN
  IF NOT isValidAccessor ( fd ) OR isSpecialFile( fd ) THEN
    RETURN 0
  END; (* IF NOT isValidAccessor ( fd ) *)
  curPos := UnixLib.lseek (fd, 0, UnixLib.fromCurPos);
  endPos := UnixLib.lseek (fd, 0, UnixLib.fromEnd);
  curPos := UnixLib.lseek (fd, curPos, UnixLib.fromStart);
  filePtr := LookupFilePtr ( fd );
  IF Buffer.isDirty (filePtr^.buffer) THEN
    posPerBuffer :=  Buffer.offset (filePtr^.buffer) + Buffer.currentPos ( filePtr^.buffer );
    IF posPerBuffer > endPos THEN
      endPos := posPerBuffer;
    END (* IF posPerBuffer > *)
  END; (* IF Buffer.isDirty *)
  RETURN endPos;
END maxPos;

(* Special File Descriptors *)


PROCEDURE defaultInFile (): File;
(* Returns the file descriptor associated with the current process' stdin *)

BEGIN
  RETURN stdIn
END defaultInFile;


PROCEDURE defaultOutFile (): File;
(* Returns the file descriptor associated with the current process' stdout *)

BEGIN
  RETURN stdOut
END defaultOutFile;


PROCEDURE defaultErrFile (): File;
(* Returns the file descriptor associated with the current process' stderr *)

BEGIN
  RETURN stdErr
END defaultErrFile;


(* Resource Deallocation *)


PROCEDURE Close ( VAR fd : File; VAR status : IOStatusT );
(* Closes the file associated with <fd>, passes status back in <status>. *)
VAR
  filePtr  : FilePtr;

BEGIN
  IF isSpecialFile ( fd ) THEN RETURN END; (* do nothing when its a special *)
  filePtr := LookupFilePtr ( fd );
  IF filePtr # NIL THEN
    IF ( isFlushable ( fd ) ) AND ( Buffer.isDirty ( filePtr^.buffer ) ) THEN Flush ( fd ) END;
    IF filePtr^.buffer # Buffer.InvalidBuffer() THEN
      Buffer.Release ( filePtr^.buffer );
    END; (* IF filePtr *)
      (* got this far, so have it out of the table *)
    IF ( PtrTable.DeTableByItemNum (descPtrTable, fd) ) THEN
      IF UnixLib.close ( fd ) # InvalidFile THEN
        DISPOSE (filePtr);
        status.failed := FALSE;
        status.code := IOStatus.Success;
        RETURN (* only successful exit *)
      END (* IF UnixLib *)
    END  (* if PtrTable *)
  END; (* if FilePtr *)
  (* if we get here something went wrong *)
  status.failed := TRUE;
  status.code := IOStatus.FileNotFound;
END Close;


PROCEDURE CloseAll ( VAR status : IOStatusT );
(* Closes all open file descriptors of the current process, except for those
   associated with stdin, stdout and stderr, passes status in <status>. *)

VAR
  fd : File;

BEGIN
  status.failed := FALSE;
  WHILE ( PtrTable.Count(descPtrTable) > 3) (* leave the first three as not closable *)
        AND (NOT status.failed) DO
    fd := PtrTable.GetHighestIndexedItemNum (descPtrTable);
    Close (fd, status);
  END; (* WHILE NOT PtrTable.isEmpty *)
  IF status.failed THEN
    status.code := IOStatus.Failure
  ELSE
    status.code := IOStatus.Success
  END (* IF status.failed *)
END CloseAll;


PROCEDURE isFlushable ( fd : File ) : BOOLEAN;
(* Returns TRUE if <file> is flushable, otherwise FALSE. *)

BEGIN
  RETURN NOT isSpecialFile ( fd )
END isFlushable;


PROCEDURE Flush ( fd : File );
(* Writes any unwritten buffer data to <file> if flushable. *)

(* Any attempt  to perform a flush operation on any file or stream that is not
   flushable will fail with status OperationNotSupported. *)(*

 Notes:
(1) read operations are permitted only if the read flag is set
(2) write operations are permitted only if the write or append flag is set
(3) all write (not flush) clear and reset the Insert buffer
(4) all block operations clear the Insert buffer and flush the main buffer if dirty
(5) Block operations are limited to the buffer size. If you anticipate a block operation
  a larger sized block, use OpenWithBuffer and set a larger buffer size.
(6) Flush leaves the file position where it found it. *)


VAR
  ptr : FilePtr;
  numToWrite, numWritten, c : IOSIZET;
  err : UnixLib.UErrorNum;
  oldPos, newPos : Pos;

BEGIN
  ptr := PtrTable.GetPtr ( descPtrTable, fd );
  IF ptr = NIL THEN RETURN END;
  IF NOT isFlushable( fd ) THEN
    ptr^.status := IOStatusT{ TRUE, IOStatus.OperationNotSupported };
    RETURN;
  END; (* IF NOT isFlushable( fd ) *)
  
(* Store the current file position on entry *)
  oldPos := currentPos (fd);
(* now synchronize the starting positions of the buffer and the physical file *)
  Buffer.SetPos (ptr^.buffer, 0);
  numToWrite := Buffer.count(ptr^.buffer);
  newPos := UnixLib.lseek (fd, Buffer.offset (ptr^.buffer), UnixLib.fromStart);
  (* and dump the buffer data into the UNIX file *)
  numWritten := UnixLib.write (fd, ptr^.buffAdr, numToWrite );
  (* then flush the physical file as it is likely buffered too *)

  (* and check out the error situation *)
  err := UnixLib.fsync ( fd );
  IF ( numToWrite # numWritten ) OR ( err # 0 ) THEN
    ptr^.status := IOStatusT{ TRUE, IOStatus.DeviceError };
  ELSE
    ptr^.status := IOStatusT{ FALSE, IOStatus.Success }
  END; (* IF ( numToWrite # *)
  (* reset the buffer parameters as it has been flushed out *)
  Buffer.Reset ( ptr^.buffer );
    (* and put the file pos back to what it was before we came here. *)
  SetPos (fd, oldPos);
  (* and put the physical file pos back to that point too. *)
  newPos := UnixLib.lseek (fd, Buffer.offset (ptr^.buffer), UnixLib.fromStart);
END Flush;


VAR
  gstatus : IOStatusT;
  n : IOSIZET;
  ch : CHAR;
  num:IOSIZET;

BEGIN
  InitModule;
END FileDescIO.