(* (C) 2016 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE FileIO;

FROM Storage IMPORT ALLOCATE, DEALLOCATE;
IMPORT ChanIO, FileDescIO;

IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;
IMPORT FileMode; FROM FileMode IMPORT FileModeT;
IMPORT IOParams, PtrSet;
FROM Transition IMPORT OCTET;

(* File Specific Channel Based IO *)

(* crib until R10 sntax ready *)
VAR
  opVector : ChanIO.Operations;
  num : Size;
  fileTable : PtrSet.PtrSet;

(* initialization *)

PROCEDURE InitModule;

BEGIN
  PtrSet.New (fileTable);
  NEW (opVector);
  opVector^.statusOf := statusOf;
  opVector^.StatusMsg := StatusMsg;
  opVector^.SetStatus := SetStatus;
  opVector^.modeOf := modeOf;
  opVector^.nameLen := nameLen;
  opVector^.GetName := GetName;
  opVector^.dataReady := dataReady;
  opVector^.ReadOctet := ReadOctet;
  opVector^.ReadBlock := ReadBlock;
  opVector^.insertReady := insertReady;
  opVector^.Insert := Insert;
  opVector^.WriteOctet := WriteOctet;
  opVector^.WriteBlock := WriteBlock;
  opVector^.isFlushable := isFlushable;
  opVector^.Flush := Flush
END InitModule;

PROCEDURE InitFile ( VAR file : File );

BEGIN
  NEW (file);
  IF file # NIL THEN
    file^.op := opVector;
  END (*IF file # NIL *)
END InitFile;

(* Introspection *)

PROCEDURE isFileChannel ( file : File ) : BOOLEAN;
(* Returns TRUE if <chan> is a valid file channel, else FALSE. *)

BEGIN
  IF file = NIL THEN
    RETURN FALSE
  ELSE
    RETURN PtrSet.isPresent( fileTable, file )
  END (* IF file = NIL *)
END isFileChannel;


PROCEDURE isValidAccessor ( file : File ) : BOOLEAN;
(* Returns TRUE if the file accessor of <file> is valid, else FALSE. *)

BEGIN
  RETURN isFileChannel (file) AND (FileDescIO.isValidAccessor(file^.handle) )
END isValidAccessor;


(* Opening and Re-opening a File *)

(* Opening or reopening a file in a mode  whose append flag is not set  causes
   the file position  to be set  to the beginning of the file.  Opening or re-
   opening a file in a mode whose append flag is set  causes the file position
   to be set  to the end of the file.  An attempt to open a file that is a di-
   rectory fails with status code MayNotOpenDirectory. *)


PROCEDURE CreateAndOpen ( (*NEW*)VAR file       : File;
                 (*. CONST .*) filename : ARRAY OF CHAR;
                 mode           : FileModeT;
                 VAR status     : IOStatusT );
(* Creates and opens the file <filename> in file mode <mode>. If the file already
   exists, nothing happens and status is set to AlreadyExists. If sucessful passes
   a file accessor back in <file>,  otherwise passes NIL back in <file>.  Sets the
   file position depending on <mode>.  The status of the operation is passed
   back in <status>. Note that the file handling module does the creation. *)

BEGIN
  IF isFileChannel (file) THEN
   (* that channel already exists *)
    status := IOStatusT{ TRUE, IOStatus.AlreadyExists };
    RETURN
  END; (* IF isFileChannel (file)s *)

  (* OK to go ahead, so *)
  InitFile (file);
  FileDescIO.CreateAndOpen(file^.handle, filename, mode, status);
  IF status.failed THEN
    DISPOSE (file)
  END (* IF status.failed *)
  (* and simply return with that error *)
END CreateAndOpen;


PROCEDURE Open
  ( (*NEW*) VAR file       : File;
    (*CONST*) filename : ARRAY OF CHAR;
    mode           : FileModeT;
    VAR status     : IOStatusT );
(* Opens file <filename> in file mode <mode>.  On success, a  newly allocated
   and initialised file channel is passed back in <file>.  On failure, NIL is
   passed back in <file>.  A status code is passed back in <status>. *)

BEGIN
  OpenWithBufferSize (file, filename, mode, IOParams.DefaultBufferSize, status)
END Open;

PROCEDURE OpenWithBufferSize
  ( (*NEW*) VAR file         : File;
    (*CONST*) filename   : ARRAY OF CHAR;
    mode             : FileModeT;
    buffSize         : Size;
    VAR status       : IOStatusT );
(* Opens file <filename>  in file mode <mode>  using <buffer> as a custom file
   buffer.   On success,  a newly allocated  and initialised  file channel  is
   passed back in <file>.  On failure, NIL is passed back in <file>.  A status
   code is passed back in <status>. *)

BEGIN
  IF isFileChannel (file) THEN
   (* that channel already exists *)
    status := IOStatusT{ TRUE, IOStatus.AlreadyExists };
    RETURN
  END; (* IF isFileChannel (file)s *)
  (* OK to go ahead, so *)
  InitFile (file);
  FileDescIO.OpenWithBufferSize(file^.handle, filename, mode, buffSize, status);
  IF status.failed THEN 
    DISPOSE (file) (* and simply return with that error *)
  ELSE
    PtrSet.Insert(fileTable, file);
  END (* IF status.failed *)
END OpenWithBufferSize;


PROCEDURE ReOpen ( file : File; mode : FileModeT );
(* Flushes <file> and changes its access mode to <mode>.
   The file position is set depending on <mode>. *)
VAR
  status : IOStatusT;

BEGIN
  FileDescIO.ReOpen(file^.handle, mode, status )
  (* we don't need to do anything with this status.
   It is set on the handle (file) and can be queried by StatusOf below *)
END ReOpen;


(* Closing a File *)

PROCEDURE Close ( VAR file : File; VAR status : IOStatusT );
(* Performs Flush on <file>,  closes  the associated file,  deallocates <file>
   and passes NIL back in <file>. A status code is passed back in <status>. *)

BEGIN
  FileDescIO.Close(file^.handle, status);
  PtrSet.Remove(fileTable, file);
  DISPOSE (file)
END Close;

PROCEDURE CloseAll ( VAR status : IOStatusT );
(* Closes all open file channels opened by this module, passes status in <status>. 
Note that channels created elsewhere, such as the standard streams are unaffected *)

VAR
  file : File;

BEGIN
  status.failed := FALSE;
  WHILE ( NOT PtrSet.isEmpty(fileTable) ) AND (NOT status.failed)  DO
    PtrSet.GetMostRecent (fileTable, file);
    Close (file, status);
  END; (* WHILE NOT PtrTable.isEmpty *)
  IF status.failed THEN
    status.code := IOStatus.Failure
  ELSE
    status.code := IOStatus.Success
  END (* IF status.failed *)

END CloseAll;

(* Status operations *)


PROCEDURE statusOf ( file : File ) : IOStatusT;
(* Returns the status of the most recent operation on <file>. *)

BEGIN
  RETURN FileDescIO.statusOf(file^.handle)
END statusOf;


PROCEDURE StatusMsg ( file : File; status : IOStatusT );
(* Writes a status message for <status> to <file>. *)

BEGIN
  FileDescIO.StatusMsg(file^.handle, status)
END StatusMsg;


PROCEDURE SetStatus
  ( file : File; status : IOStatusT; VAR valid : BOOLEAN );
(* Sets  the status of <file> to <status> if it meets the following condition:
   Where <status.code> is set to Success <status.failed> must be FALSE, other-
   wise TRUE.  Passes TRUE in <valid> if successful, otherwise FALSE. *)

BEGIN
  FileDescIO.SetStatus(file^.handle, status, valid )
END SetStatus;

(* Querying File Parameters *)


PROCEDURE modeOf ( file : File ) : FileModeT;
(* Returns the access mode of <file>. *)

BEGIN
  RETURN FileDescIO.modeOf(file^.handle)
END modeOf;


PROCEDURE nameLen ( file : File ) : CARDINAL;
(* Returns the length of the filename associated with <file>. *)

BEGIN
  RETURN FileDescIO.nameLen(file^.handle)
END nameLen;


PROCEDURE GetName ( file : File; VAR filename : ARRAY OF CHAR );
(* Passes the name of the file  associated with <file> back in <filename>.  If
   the name exceeds the capacity of <filename> the operation fails with status
   code NameTooLong and an empty string is passed back in <filename>. *)

BEGIN
  FileDescIO.GetName(file^.handle, filename)
END GetName;


(* Fie Positioning *)

PROCEDURE eof ( file : File ) : BOOLEAN;
(* Returns TRUE if the end of <file> has been reached, otherwise FALSE. *)

BEGIN
  RETURN (FileDescIO.eof(file^.handle) )
END eof;


PROCEDURE currentPos ( file : File ) : Pos;
(* Returns the current read/write position of <file>. *)

BEGIN
  RETURN FileDescIO.currentPos(file^.handle)
END currentPos;


PROCEDURE isValidPos ( file : File; pos : Pos ) : BOOLEAN;
(* Returns TRUE if <pos> is a valid position for <file>, else FALSE. *)

BEGIN
  RETURN FileDescIO.isValidPos(file^.handle, pos)
END isValidPos;


PROCEDURE lastValidSetPos ( file : File ) : Pos;
(* Returns the most advanced position that may be passed to SetPos for <file>.
   The return value depends on the size and access mode of <file> as follows:

   (1) if <file> is in append mode, the function returns zero and sets the
       status of <file> to OperationNotSupported.
   (2) if <file> is empty, then TMIN(FilePos) is returned.
   (3) if <file> is not empty and has its write flag set, then the position
       of the last octet in <file> plus one is returned.
   (4) if <file> is not empty and has its read but not its write flag set,
       then the position of the last octet in <file> is returned. *)

BEGIN
  RETURN FileDescIO.lastValidSetPos(file^.handle)
END lastValidSetPos;


PROCEDURE SetPos ( file : File; pos : Pos );
(* Sets the  read/write position of <chan> to <pos>. For any empty file, <pos>
   must be zero.  For any non-empty file in read mode, <pos> must be less than
   LastValidWritePos. For any non-empty file in write mode, <pos> must be less
   than or equal to LastValidWritePos. *)

BEGIN
  FileDescIO.SetPos(file^.handle, pos)
END SetPos;


PROCEDURE Advance ( file : File; offset : Pos );
(* Advances the read/write position of <file> by <offset>. *)

BEGIN
  FileDescIO.Advance(file^.handle, offset)
END Advance;


PROCEDURE Rewind ( file : File );
(* Sets the read/write position of <file> to the beginning of the file and re-
   sets its end-of-file status. *)

BEGIN
  FileDescIO.Rewind(file^.handle)
END Rewind;

(* Any attempt to call  LastValidWritePos, SetPos, Advance or Rewind on a file
   channel whose access mode  has the append flag set  shall fail  with status
   OperationNotSupported.  Any attempt  to set  a file channel's read position
   past  the  end  of the file  shall fail  with  status AccessBeyondEOF.  Any
   attempt to set a file channel's write position past LastValidWritePos shall
   fail with status AccessBeyondEOF. *)


(* I/O operations *)


PROCEDURE dataReady ( file : File; octets : Size ) : BOOLEAN;
(* Returns TRUE if <chan> has at least <octets> number of octets available for
   reading, otherwise FALSE. *)
   (* this function isn't in the current spec for FilePtrIO or FileDescIO,
    so we synthesize it here including the read and write checks *)

VAR
  lastValidReadPos : Pos;
  mode : FileModeT;
  
BEGIN
  mode := modeOf (file);
  IF NOT (FileMode.Read IN mode) THEN
    RETURN FALSE
  ELSE
    lastValidReadPos := lastValidSetPos ( file );
    IF FileMode.Write IN mode THEN
      lastValidReadPos := lastValidReadPos - 1;
    END; (* IF FileMode.Writ IN *)
    RETURN ( lastValidReadPos - currentPos ( file ) + 1 ) >= octets
  END; (* IF FileMode.Read IN *)
END dataReady;


PROCEDURE ReadOctet ( file : File; VAR data : OCTET );
(* Reads one octet of data  at the current position of <file>,  passes it back
   in <data> and advances the current read/write position of <file> by one. *)

BEGIN
  FileDescIO.ReadOctet(file^.handle, data)
END ReadOctet;


PROCEDURE ReadBlock
  ( file : File; VAR data : ARRAY OF OCTET; VAR octetsRead : Size );
(* Reads a block of data  at the current position of <file>.  Passes the block
   of data  back in <data>.  The read/write position of <file>  is advanced by
   the number of octets read. *)
   
VAR
  n : CARDINAL;
 
BEGIN
  FileDescIO.ReadBlock(file^.handle, data, n)
END ReadBlock;


PROCEDURE insertReady ( file : File ) : BOOLEAN;
(* Returns FALSE if the insert buffer of <chan> is full, otherwise TRUE. *)

BEGIN
  RETURN FileDescIO.insertReady(file^.handle);
END insertReady;


PROCEDURE Insert ( file : File; data : OCTET );
(* Inserts octet <data> into input channel <chan>  to be read by the next read
   operation on <chan>.  Inserted octets are stored in a  FIFO buffer  of size
   InsertBufferSize.  The minimum size of the insert buffer is two octets. *)

BEGIN
  FileDescIO.Insert(file^.handle, data)
END Insert;

(* Any attempt  to perform the above I/O operations  on any file channel whose
   Read flag is not set will fail with status NotOpenForReading. *)


PROCEDURE WriteOctet ( file : File; data : OCTET );
(* Performs a write or append operation depending on the access mode of <file>
   and advances the current read/write position by one after the data has been
   written. In write mode <data> is written to the current position of <file>.
   In append mode the procedure  atomically  sets the current read/write posi-
   tion to the end of the file and appends <data> to the end of <file>. *)

BEGIN
  FileDescIO.WriteOctet(file^.handle, data)
END WriteOctet;


PROCEDURE WriteBlock
  ( file : File; data : ARRAY OF OCTET; VAR octetsWritten : Size );
(* Performs a write or append operation depending on the access mode of <file>
   and passes the actual number of octets written in <octetsWritten>. The cur-
   rent read/write position is advanced accordingly. In write mode, the proce-
   dure writes <data> at the current read/write position of <chan>.  In append
   mode,  it atomically sets the current read/write position to the end of the
   file and appends <data> to the end of <file>. *)

BEGIN
  FileDescIO.WriteBlock(file^.handle, data, octetsWritten)
END WriteBlock;


PROCEDURE isFlushable ( file : File ) : BOOLEAN;
(* Returns TRUE if <chan> is flushable, otherwise FALSE. *)

BEGIN
  RETURN FileDescIO.isFlushable(file^.handle)
END isFlushable;


PROCEDURE Flush ( file : File );
(* Writes unwritten data in any buffer of <chan> to its associated file. *)

BEGIN
  FileDescIO.Flush(file^.handle)
END Flush;

(* Any attempt  to write to  or flush a file whose write flag is not set shall
   fail with status OperationNotSupported. *)

BEGIN
  InitModule;
END FileIO.