(* (C) 2009-2017 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE FilePtrIO;

(* Driver for File IO on File Pointers *)

IMPORT SYSTEM, UNSAFE;
FROM Transition IMPORT OCTET, CAPACITYo;
FROM INTEGER32 IMPORT INTEGER32;

IMPORT ISO646, IOParams, ARRAYOFCHAR, PtrSet;
IMPORT IOSIZE; FROM IOSIZE IMPORT IOSIZET;
IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;
IMPORT FileMode; FROM FileMode IMPORT FileModeT;

FROM Storage IMPORT ALLOCATE, DEALLOCATE;

IMPORT CStdIOLib, errno;
(*IMPORT Perror;*)

VAR
  num : IOSIZET;

(* File accessor and file status *)

CONST
  InsertBuffMax = IOParams.InsertBufferSize - 1;

TYPE
  DString = ARRAY [0 .. 255] OF CHAR;
  CmodeStr = ARRAY [0 .. 4] OF CHAR;
  InsertBuffPos = [0..InsertBuffMax];

  File = POINTER TO FileDataTable; (* implementation defined file accessor *)
  FileDataTable = RECORD
    fd : CStdIOLib.FILE;
    name : DString;
    mode : FileModeT;
    status : IOStatusT;
    buffSize,
    index : IOSIZET;
    buffer : UNSAFE.ADDRESS;
    insertBuffer : ARRAY [0.. InsertBuffMax] OF OCTET;
    iBuffIn, iBuffOut, iBuffCount : InsertBuffPos;
END;

VAR
  stdIn, stdOut, stdErr : File;
  
(* Utility Procs *)


PROCEDURE RFileModeToCFileMode ( rMode : FileModeT; VAR cMode : CmodeStr );
(* Pre: Mode has been checked for validity *)

BEGIN
IF FileMode.Append IN rMode THEN cMode := CStdIOLib.append
  ELSIF rMode = FileMode.RO THEN cMode := CStdIOLib.read
  ELSIF rMode = FileMode.WO THEN cMode := CStdIOLib.write
  ELSIF rMode = FileMode.RW THEN cMode := CStdIOLib.update
  END; (* IF FileMode.Append*)
END RFileModeToCFileMode;


PROCEDURE CStatusToIOStatus ( ) :IOStatusT;
(* Pre : An error in the C file system access is known to have taken place
   and no other access has been undertaken since.
   Post : Caller decides what to do with the returned status *)
VAR
  errPtr : errno.INT32Ptr;
  errnum : errno.SysErrNum;

BEGIN
    errPtr := errno.__error ();
    errnum := errPtr^;
    CASE errnum OF
      errno.EOF :
        RETURN IOStatusT{ TRUE, IOStatus.AccessBeyondEOF};
    | errno.ENOERR :
        RETURN IOStatusT{ FALSE, IOStatus.Success};
    | errno.ENOENT :
        RETURN IOStatusT{ TRUE, IOStatus.FileNotFound};
    | errno.EIO :
        RETURN IOStatusT{ TRUE, IOStatus.Failure};
    | errno.ENXIO :
        RETURN IOStatusT{ TRUE, IOStatus.DeviceError};
    | errno.EBADF :
        RETURN IOStatusT{ TRUE, IOStatus.InvalidFile};
    | errno.ENOMEM :
        RETURN IOStatusT{ TRUE, IOStatus.AllocationFailed};
    | errno.EPERM, errno.EACCES:
        RETURN IOStatusT{ TRUE, IOStatus.AccessDenied};
    | errno.EROFS:
        RETURN IOStatusT{ TRUE, IOStatus.NotOpenForReading};
    | errno.EISDIR:
        RETURN IOStatusT{ TRUE, IOStatus.MayNotOpenDirectory};
    | errno.EEXIST:
        RETURN IOStatusT{ TRUE, IOStatus.AlreadyExists};
    | errno.ENFILE, errno.EMFILE :
        RETURN IOStatusT{ TRUE, IOStatus.OpenFileLimitExceeded};
    | errno.ENOSPC:
        RETURN IOStatusT{ TRUE, IOStatus.DeviceFull};
    | errno.ESPIPE:
        RETURN IOStatusT{ TRUE, IOStatus.InvalidFilePos};
    | errno.ENODEV :
        RETURN IOStatusT{ TRUE, IOStatus.OperationNotSupported};
    ELSE
        RETURN IOStatusT{ TRUE, IOStatus.Failure};
    END (*CASE*);
END CStatusToIOStatus;

(* File Table *)

VAR
  fileTable : PtrSet.PtrSet;

(* initialization *)


PROCEDURE Init;

BEGIN
  PtrSet.New (fileTable);
  OpenStdFiles
END Init;


(* Inspection *)


PROCEDURE isValidAccessor ( file : File ) : BOOLEAN;
(* Returns TRUE if <file> is a valid file accessor, otherwise FALSE. *)

BEGIN
  IF file = NIL THEN
    RETURN FALSE
  ELSE
    RETURN PtrSet.isPresent( fileTable, file )
  END (* IF file = NIL *)
END isValidAccessor;


(* Opening and Closing Files *)

PROCEDURE OpenStdFiles;
VAR
 status : IOStatusT;
BEGIN
    (* stdIn *)
  Open ( stdIn, "/dev/stdin", FileMode.RO, status );
  IF (status.code # IOStatus.Success) OR (stdIn = NIL) THEN
   (* Perror.prerr ("Open StdIn Failed\n") *)
  END;
  
    (* stdOut *)
  Open ( stdOut, "/dev/tty", FileMode.WO, status );
  IF status.code # IOStatus.Success THEN
    (*Perror.prerr ("Open StdOut Failed\n")*)
  END;

    (* std err *)
  Open ( stdErr, "/dev/tty", FileMode.WO, status );
  IF status.code # IOStatus.Success THEN
    (*Perror.prerr ("Open StdOut Failed\n")*)
  END; (* IF status.code *)
END OpenStdFiles;

PROCEDURE CreateAndOpen ( (*NEW*)VAR file       : File;
                 (*. CONST .*) filename : ARRAY OF CHAR;
                 mode           : FileModeT;
                 VAR status     : IOStatusT );
(* Creates and opens the file <filename> in file mode <mode>. If the file already
   exists, nothing happens and status is set to AlreadyExists. If sucessful passes
   a file accessor back in <file>,  otherwise passes NIL back in <file>.  Sets the
   file position depending on <mode>.  The status of the operation is passed
   back in <status>. *)

BEGIN
  IF CStdIOLib.access (filename, CStdIOLib.F_OK) = 0 THEN
   (* file already exists *)
    file := NIL;
    status := IOStatusT{ TRUE, IOStatus.AlreadyExists };
    RETURN
  END; (* IF CStdIOLib.access *)

(* now check to ensure the mode is valid *)
  IF ( NOT FileMode.isValid (mode) ) OR (NOT (FileMode.Write IN mode) )  THEN
   status := IOStatusT{ TRUE, IOStatus.InvalidMode};
   file := NIL;
   RETURN
  END; (* NOT isValid (mode) *) 
  
  (* OK so try to get it open with the clean mode initially *)
  Open (file, filename, FileMode.WO, status); (* will create *)
  IF status.code = IOStatus.Success
    THEN
      ReOpen (file, mode, status);
    END (* IF status.code *)
  (* otherwise the status returned is the one from the attempt to re open *)
END CreateAndOpen;

(* Opening or reopening a file in a mode whose append flag is not set causes
   the file position to be set to the beginning of the file.  Opening or re-
   opening a file in a mode whose append flag is set causes the file position
   to be set to the end of the file.  An attempt to open a file that is a di-
   rectory fails with status code mayNotOpenDirectory. *)


PROCEDURE Open ( (*NEW*)VAR file       : File;
                 (*. CONST .*) filename : ARRAY OF CHAR;
                 mode           : FileModeT;
                 VAR status     : IOStatusT );

BEGIN
  OpenWithBufferSize (file, filename, mode, IOParams.DefaultBufferSize, status)
END Open;


PROCEDURE OpenWithBufferSize ( (*NEW*)VAR file         : File;
                           (*. CONST .*) filename   : ARRAY OF CHAR;
                           mode             : FileModeT;
                           buffSize         : IOSIZET;
                           VAR status       : IOStatusT );

VAR
  fc : CStdIOLib.FILE;
  cMode : CmodeStr;
  res : INTEGER32;
  resB : INTEGER;

BEGIN 
(* do some checking *)
  IF FileMode.isValid (mode) THEN
    RFileModeToCFileMode (mode, cMode )
  ELSE
    status := IOStatusT{ TRUE, IOStatus.InvalidMode };
    RETURN
  END; (* IF FileMode.isValid *)
(* request a C FILE handle *)
  fc := CStdIOLib.fopen (filename, cMode);
  (* on failure bail out *)
  IF fc = NIL THEN 
    status := IOStatusT{ TRUE, IOStatus.FileNotFound };
    RETURN
  END; (* IF fc = NIL *)

 (* set up a FilePtr file *)
  NEW (file);
 (* on failure bail out *)
  IF file = NIL THEN
    res := CStdIOLib.fclose(fc);
    status := IOStatusT{ TRUE, IOStatus.AllocationFailed };
    RETURN
  END; (* IF file = NIL *)
  file ^.fd := fc;
  file ^.name := "";
  ARRAYOFCHAR.Concat ( file ^.name, file ^.name, filename);
 (* get a buffer *)
  IF buffSize < IOParams.MinBufferSize THEN
    buffSize := IOParams.MinBufferSize;
  END; (* IF buffSize *)
  ALLOCATE (file^.buffer, buffSize);
  file^.buffSize := buffSize;
  file^.mode := mode;
  ResetInsertBuffer ( file );
(* next attach that buffer to the lower level file *)
  res := CStdIOLib.setvbuf ( fc, file^.buffer, CStdIOLib._IOFBF, file^.buffSize );
  IF res # 0 THEN
     Close (file, status);
     status := IOStatusT{ TRUE, IOStatus.DeviceError };
     RETURN
  END; (* IF res # 0 *)

 (* record that pointer *)
  PtrSet.Insert(fileTable, file);
  (* finally, set the file position according to the mode *)
  status := IOStatusT{ FALSE, IOStatus.Success };
  file^.status := status;
END OpenWithBufferSize;


PROCEDURE ReOpen ( file : File; mode : FileModeT; VAR status: IOStatusT );
(* Flushes the file associated with file accessor <file> and changes its mode
   to <mode>. Sets the file position depending on <mode>. *)
VAR
  cMode : CmodeStr;
  res : INTEGER32;

(* do some checking *)

BEGIN
  IF file = NIL THEN
    status := IOStatusT{ TRUE, IOStatus.InvalidFile };
    RETURN
  END; (* IF file = NIL*)
  IF FileMode.isValid (mode) THEN
    RFileModeToCFileMode (mode, cMode )
  ELSE
    file^.status := IOStatusT{ TRUE, IOStatus.InvalidMode };
    RETURN
  END; (* IF FileMode.isValid *)
  IF FileMode.Write IN file^.mode THEN
    Flush (file)
  END; (* IF FileMode.Write IN *)
  CStdIOLib.freopen ( file^.name, cMode, file^.fd );
  ResetInsertBuffer ( file )
END ReOpen;


PROCEDURE Close ( VAR file : File; VAR status : IOStatusT );
(* Performs Flush on <file>, closes the associated file and passes NIL back
   in <file>.  The status of the operation is passed back  in <status>. *)
VAR
  res : INTEGER32;

BEGIN
  IF NOT isValidAccessor (file) THEN
    status := IOStatusT{ TRUE, IOStatus.InvalidFile };
    RETURN
  END; (* IF NOT isValidAccessor (file) *)
  IF FileMode.Write IN file^.mode THEN
    Flush (file);
  END; (* IF FileMode.Write *)
  res := CStdIOLib.fclose (file^.fd);
  IF res # 0 THEN
    status := IOStatusT{ TRUE, IOStatus.DeviceError };
    RETURN
  END; (* IF res # 0 *)
  PtrSet.Remove (fileTable, file);
  DEALLOCATE (file^.buffer, file^.buffSize);
  DISPOSE (file);
  status := IOStatusT{ FALSE, IOStatus.Success}
END Close;


PROCEDURE CloseAll;
(* Closes all open files. *)
VAR
  file : File;
  status : IOStatusT;

BEGIN
  status.failed := FALSE;
  PtrSet.GetMostRecent (fileTable, file);
  WHILE (file # NIL) AND (NOT status.failed) DO
    Close (file, status);
    PtrSet.GetMostRecent (fileTable, file)
  END; (* WHILE file # NIL *)
  IF status.failed THEN
    status.code := IOStatus.Failure
  ELSE
    status.code := IOStatus.Success
  END (* IF status.failed *)
END CloseAll;


(* Querying File Parameters *)


(* The mode of a file consists of four flags: Read, write, append and sync.
   These flags may be combined arbitrarily, subject to the following rules:

     1) a file mode may not be the empty set
     2) append may not be set if Read is set
     3) append may only be set if write is set
     4) sync may only be set if write is set

    *)


PROCEDURE modeOf ( file : File ) : FileModeT;
(* Returns the file mode of file accessor <file>. *)

BEGIN
  IF NOT isValidAccessor (file) THEN
    RETURN FileModeT{} (* return invalid mode *)
  ELSE
    RETURN file^.mode
  END; (* IF NOT isValidAccessor (file) *)
END modeOf;


PROCEDURE nameLen ( file : File ) : CARDINAL;
(* Returns the length of the filename associated with <file>. *)

BEGIN
   IF NOT isValidAccessor (file) THEN
    RETURN 0
  ELSE
    RETURN LENGTH (file^.name)
  END; (* IF NOT isValidAccessor (file) *)
END nameLen;


PROCEDURE GetName ( file : File; VAR filename : ARRAY OF CHAR );
(* Passes the name  of the file  associated with  file accessor <file> back in
   <filename>. If the name exceeds the capacity of <filename>, an empty string
   is passed back in <filename>.  Does not alter the status of <file>. *)

BEGIN
  filename[0] := CHR(0);
  IF NOT isValidAccessor (file) OR (nameLen(file) > CAPACITYo (filename) -1 ) THEN RETURN
  ELSE
    ARRAYOFCHAR.Concat ( filename, filename, file ^.name );
  END; (* IF NOT isValidAccessor (file) *)
END GetName;


PROCEDURE statusOf ( file : File ) : IOStatusT;
(* Returns the status of the most recent operation for <file>. *)

BEGIN
  IF NOT isValidAccessor (file) THEN
    RETURN IOStatusT{ TRUE, IOStatus.InvalidFile };
  ELSE
    RETURN file^.status
  END (* IF NOT isValidAccessor (file) *)
END statusOf;

PROCEDURE isFileOK ( file : File ) : BOOLEAN;
(* returns TRUE provided the accessor is valid, status.failed is FALSE and 
the status code is Success. Otherwise returns FALSE *)
VAR
  status : IOStatusT;
  
BEGIN
  IF NOT isValidAccessor (file) THEN
    RETURN FALSE;
  ELSE
    status := statusOf ( file );
    RETURN ( NOT (status.failed) ) AND ( status.code = IOStatus.Success )
  END (* IF NOT isValidAccessor *)
END isFileOK;

PROCEDURE StatusMsg ( fd : File; status : IOStatusT );
(* Writes a status message for <status> to <fd>. *)
VAR
  msg : IOStatus.statusMsg;
  count : IOSIZET;
  last : CARDINAL;

BEGIN
  IOStatus.fetchMsg (status, msg);
  count := 0;
  last := CAPACITYo ( msg );
  WHILE (count < last ) AND ( msg [count] #  ISO646.NUL ) DO
    WriteOctet (fd, msg [count] );
    count := count + 1
  END; (* WHILE (count < *)
  WriteOctet ( fd, ISO646.LF )
END StatusMsg;


PROCEDURE SetStatus ( file : File; status : IOStatusT; VAR valid : BOOLEAN );
(* Sets the status of <file>  to <status> if it meets the integrity condition.
   Where <status.code> is set to Success <status.failed> must be FALSE, other-
   it must be TRUE.  Passes TRUE in <valid> if successful, otherwise FALSE. *)

BEGIN
  IF NOT isValidAccessor (file) THEN
    valid := FALSE;
    RETURN
  END; (* IF NOT isValidAccessor (file) *)
  valid := (status.code = IOStatus.Success) AND NOT (status.failed)
       OR (status.code # IOStatus.Success) AND (status.failed);
  IF valid THEN
    file^.status := status;
  END; (* IF valid *)
END SetStatus;


(* File-Only Positioning Operations *)


PROCEDURE eof ( file : File ) : BOOLEAN;
(* Returns TRUE if the end of the file associated with file accessor <file>
   has been reached, otherwise FALSE. *)

BEGIN
  IF NOT isValidAccessor (file) THEN
    RETURN FALSE;
  ELSE
    file^.status := IOStatusT{ FALSE, IOStatus.Success };
    RETURN ( CStdIOLib.feof (file^.fd) # 0 )
  END; (* IF NOT isValidAccessor (file) *)
END eof;


PROCEDURE currentPos ( file : File ) : Pos;
(* Returns the current read/write position for file accessor <file> where a
   value of TMIN(Pos) represents the beginning of a file. *)

BEGIN
  IF NOT isValidAccessor (file) THEN
    file^.status := IOStatusT{ TRUE, IOStatus.InvalidFile };
    RETURN 0
  ELSE
    file^.status := IOStatusT{ FALSE, IOStatus.Success };
    RETURN VAL (Pos, CStdIOLib.ftell (file^.fd))
  END (* IF NOT isValidAccessor (file) *)
END currentPos;

(* local Proc *)
PROCEDURE maxPos ( file : File ) : Pos;
(* Pre: calling proc has already checked validity of file accessor *)
VAR
  oldPos, max : Pos;
  res : INTEGER;

BEGIN
  Flush (file);
  ResetInsertBuffer ( file );
  oldPos := currentPos (file);
  res := CStdIOLib.fseek (file^.fd, 0, CStdIOLib.SEEK_END );
  max := currentPos (file);
  res := CStdIOLib.fseek (file^.fd, oldPos, CStdIOLib.SEEK_SET );
  RETURN max;
END maxPos;

(* back to exported procs *)


PROCEDURE isValidPos ( file : File; pos : Pos ) : BOOLEAN;
(* Returns TRUE if <pos> is a valid file position for <file>, otherwise FALSE. *)

BEGIN
  IF NOT isValidAccessor (file) THEN
    RETURN FALSE
  ELSE
    RETURN pos <= maxPos ( file )
  END (* IF NOT isValidAccessor (file) *)
END isValidPos;


PROCEDURE lastValidSetPos ( file : File ) : Pos;
(* Returns the most advanced position that may be passed to SetPos for file
   accessor <file>.  The return value depends on length and mode of <file>

   (1) if <file> is in append mode, the function returns zero and sets the
       file's status to OperationNotSupported.
   (2) if <file> is empty, then TMIN(Pos) is returned.
   (3) if <file> is not empty and has its write flag set, then the position
       of the last octet in <file> plus one is returned.
   (4) if <file> is not empty and has its read but not its write flag set,
       then the position of the last octet in <file> is returned. *)
VAR
  max : Pos;

BEGIN
  IF NOT isValidAccessor (file) THEN
    RETURN 0
  ELSIF FileMode.Append IN file^.mode THEN
    file^.status := IOStatusT{ TRUE, IOStatus.OperationNotSupported };
    RETURN 0
  END; (* IF NOT isValidAccessor (file) *)
  max := maxPos ( file );
  IF FileMode.Write IN file^.mode THEN
    file^.status := IOStatusT{ FALSE, IOStatus.Success };
    RETURN max + 1
  ELSIF FileMode.Read IN file^.mode THEN
    file^.status := IOStatusT{ FALSE, IOStatus.Success };
    RETURN max
  ELSE
    (* if we get here, the file's mode is corrupt *)
    RETURN 0
  END (* IF FileMode.Write *)
END lastValidSetPos;


PROCEDURE SetPos ( file : File; pos : Pos );
(* Sets the  read/write position  for  file accessor <file>  to <pos>.  For
   any empty file, <pos> must be zero. For any non-empty file in read mode,
   <pos> must be  less than  LastValidWritePos.  For any  non-empty file in
   write mode, <pos> must be  less than or equal  to LastValidWritePos. *)
VAR
  res : Pos;

BEGIN
  IF NOT isValidAccessor (file) THEN
    RETURN
  ELSIF FileMode.Append IN file^.mode THEN
    file^.status := IOStatusT{ TRUE, IOStatus.OperationNotSupported };
    RETURN
  ELSIF pos > lastValidSetPos(file) THEN
    file^.status := IOStatusT{ TRUE, IOStatus.AccessBeyondEOF };
    RETURN
  ELSE (* first check if eof is set and clear if so *)
    IF eof (file) THEN
      CStdIOLib.clearerr(file^.fd)
    END; (* IF eof (file) *)
    (* go ahead and do it *)
    Flush (file);
    ResetInsertBuffer ( file );
    res := CStdIOLib.fseek (file^.fd, pos, CStdIOLib.SEEK_SET );
    IF res # 0 THEN
      file^.status := IOStatusT{ TRUE, IOStatus.OperationNotSupported };
    ELSE
      file^.status := IOStatusT{ FALSE, IOStatus.Success };
      ResetInsertBuffer ( file )
    END (* IF res # 0 *)
  END (* IF NOT isValidAccessor (file) *)
END SetPos;


PROCEDURE Advance ( file : File; offset : Pos );
(* Advances the read/write position for file accessor <file> by <offset>. *)
(* all we do is calculate the new position. SetPos will take care of any errors this causes *)

BEGIN
  SetPos ( file, currentPos (file ) + offset )
END Advance;


PROCEDURE Rewind ( file : File );
(* Sets the read/write position for file accessor <file>  to the  beginning
   of the file and resets its end-of-file status. *)

BEGIN
  SetPos ( file, 0 )
END Rewind;

(* Any attempt to call  LastValidWritePos, SetPos, Advance or Rewind  on  a
   file accessor  whose file mode  has the append flag set  shall fail with
   status IOStatus.OperationNotSupported.  Any attempt to set a file's read position
   past the end of the file  shall fail  with status  IOStatus.AccessBeyondEOF.  Any
   attempt to set a file's write position past LastValidWritePos shall fail
   with status IOStatus.AccessBeyondEOF. *)

(* Common I/O operations *)


PROCEDURE ReadOctet ( file : File; VAR data : OCTET );
(* Reads one octet of data at the current position of <file>, passes it back
   in <data> and advances the read/write position of <file> by one. *)
VAR
  numRead : Size;

BEGIN
  IF NOT isValidAccessor (file) THEN
    RETURN
  END; (* IF NOT isValidAccessor (file) *)
  IF NOT (FileMode.Read IN file^.mode ) THEN
    file^.status := IOStatusT{ TRUE, IOStatus.NotOpenForReading };
  RETURN
  END; (* IF NOT (FileMode.Read *)
  IF file^.iBuffCount > 0 THEN
    Retrieve (file, data);
    RETURN
  ELSE
    numRead := CStdIOLib.fread (SYSTEM.ADR(data), 1,1, file^.fd );
    IF numRead = 1 THEN
      file^.status := IOStatusT{ FALSE, IOStatus.Success };
      RETURN
    ELSIF eof (file) THEN
      file^.status := IOStatusT{ TRUE, IOStatus.AccessBeyondEOF };
    ELSE
      file^.status := CStatusToIOStatus ( );
    END (* IF numRead = 1 *)
  END (* IF file^.iBuffCount > 0 *)
END ReadOctet;


PROCEDURE ReadBlock ( file : File; VAR data : ARRAY OF OCTET; VAR octetsRead : IOSIZET );
(* Reads as many octets from <file> as are available and fit into buffer <data>.
   Reading starts at the current position. The number of octets read is passed
   back in <octetsRead>.  The current position is advanced by <octetsRead>. *)
VAR
  numToRead : Size;
  dCount : CARDINAL;
  buffAdr : UNSAFE. ADDRESS;

BEGIN
  IF NOT isValidAccessor (file) THEN
    RETURN
  END; (* IF NOT isValidAccessor (file) *)
   IF NOT (FileMode.Read IN file^.mode ) THEN
    file^.status := IOStatusT{ TRUE, IOStatus.NotOpenForReading };
  RETURN
  END; (* IF NOT (FileMode.Read *)
  dCount := 0;
  WHILE file^.iBuffCount > 0 DO
    Retrieve (file, data [dCount]);
    dCount := dCount + 1;
  END;
  numToRead := CAPACITYo ( data ) - dCount;
  buffAdr := SYSTEM.ADDADR ( SYSTEM.ADR (data), dCount);
  octetsRead := CStdIOLib.fread ( buffAdr, 1, numToRead, file^.fd );
  IF octetsRead = numToRead THEN
    file^.status := IOStatusT{ FALSE, IOStatus.Success };
    RETURN
  ELSIF eof (file) THEN
    file^.status := IOStatusT{ TRUE, IOStatus.AccessBeyondEOF };
  ELSE
    file^.status := CStatusToIOStatus ( );
  END (* IF numRead = numToRead *)

END ReadBlock;

(* local procedure so validity of file already checked *)
PROCEDURE ResetInsertBuffer ( file : File );

BEGIN
  file^.iBuffIn := 0;
  file^.iBuffOut := 0;
  file^.iBuffCount := 0;
END ResetInsertBuffer;


PROCEDURE insertReady ( file : File ) : BOOLEAN;
(* Returns FALSE if the insert buffer of <file> is full, otherwise TRUE. *)

BEGIN
  IF NOT isValidAccessor (file) THEN
    RETURN FALSE
  END; (* IF NOT isValidAccessor (file) *)
  IF NOT (FileMode.Read IN file^.mode ) THEN
    file^.status := IOStatusT{ TRUE, IOStatus.NotOpenForReading };
    RETURN FALSE
  ELSE
    RETURN (file^.iBuffCount < InsertBuffMax);
  END (* IF NOT (FileMode.Read *)
END insertReady;


PROCEDURE Insert ( file : File; data : OCTET );
(* Inserts octet <data> into input stream <stream> to be read by the next read
   operation on <stream>.  Inserted octets are stored in a FIFO buffer of size
   InsertBufferSize.  The minimum size of the insert buffer is two octets. *)

(* Any attempt to perform the above I/O operations on any file or stream whose
   read flag is not set shall fail with status OperationNotSupported. *)

BEGIN
  IF NOT isValidAccessor (file) THEN
    RETURN
  END; (* IF NOT isValidAccessor (file) *);
  IF NOT (FileMode.Read IN file^.mode ) THEN
    file^.status := IOStatusT{ TRUE, IOStatus.NotOpenForReading };
  RETURN
  ELSIF NOT (insertReady (file) ) THEN
    file^.status := IOStatusT{ TRUE, IOStatus.InsertBufferFull};
    RETURN;
  ELSE
    file^.insertBuffer [file^.iBuffIn] := data;
    IF file^.iBuffIn = InsertBuffMax THEN
      file^.iBuffIn := 0;
    ELSE
      file^.iBuffIn := file^.iBuffIn + 1
    END; (* IF file^.iBuffIn *)
    file^.iBuffCount := file^.iBuffCount + 1;
  END; (* IF NOT (FileMode.Read *)
END Insert;


PROCEDURE Retrieve ( file : File; VAR data : OCTET );
(* Pre: Since this is internal, and we only get here from a read operaton,
   read mode and the validity of the accessor have already been checked.
   The caller shall also check whether
   the insert buffet has anything in it before calling this procedure.
   Post: removes the next available octet from the insert buffer of file (if any)
  and returns it in <data>, adjusting the buffer parameters as needed *)

BEGIN
  data := file^.insertBuffer [file^.iBuffOut];
  IF file^.iBuffOut = InsertBuffMax THEN
    file^.iBuffOut := 0;
  ELSE
    file^.iBuffOut := file^.iBuffOut + 1
  END; (* IF file^.iBuffOut *)
  file^.iBuffCount := file^.iBuffCount - 1
END Retrieve;


PROCEDURE WriteOctet ( file : File; data : OCTET );
(* Performs a write or append operation depending on the mode of <file>.
   In write mode, the procedure writes one octet in <data> to the current
   position of <file>. In append mode, the procedure atomically sets the
   current read/write position of <file> to the end of the file and appends
   one octet in <data> to the end of the file. In either mode, the current
   read/write position of <file> is advanced by one after the data has been
   written. *)
VAR
  numWritten : Size;

BEGIN
  IF NOT isValidAccessor (file) THEN
    RETURN
  END; (* IF NOT isValidAccessor (file) *)
  IF NOT (FileMode.Write IN file^.mode ) THEN
    file^.status := IOStatusT{ TRUE, IOStatus.NotOpenForWriting };
    RETURN
  END; (* IF NOT (FileMode.Write *)
  numWritten := CStdIOLib.fwrite (SYSTEM.ADR ( data ), 1, 1, file^.fd );
  IF numWritten = 1 THEN
    file^.status := IOStatusT{ FALSE, IOStatus.Success };
    RETURN;
  ELSE
    file^.status := CStatusToIOStatus ( );
    RETURN
  END (* IF numWritten = 1 *)
END WriteOctet;


PROCEDURE WriteBlock ( file : File; data : ARRAY OF OCTET; VAR octetsWritten : Size );

(* Performs a write or append operation depending on the mode of <file>.
   In write mode, the procedure writes the block of data passed in <data>
   starting at the current read/write position of <file>. In append mode, the
   procedure atomically sets the current read/write position of <file> to the
   end of the file and appends the block of data passed in <data> to the end
   of the file. In either mode, the current read/write position of <file> is
   advanced by the number of octets written and the number of octets written
   is passed back in <octetsWritten> after the data has been written. *)

VAR
  numToWrite, numWritten : Size;

BEGIN
  IF NOT isValidAccessor (file) THEN
    RETURN
  END; (* IF NOT isValidAccessor (file) *)
  IF NOT (FileMode.Write IN file^.mode ) THEN
    file^.status := IOStatusT{ TRUE, IOStatus.NotOpenForWriting };
    RETURN
  END; (* IF NOT (FileMode.Write *)
  ResetInsertBuffer ( file );
  numToWrite := CAPACITYo (data );
  numWritten := CStdIOLib.fwrite (SYSTEM.ADR ( data ), 1, numToWrite, file^.fd );
  IF numWritten = numToWrite THEN
    file^.status := IOStatusT{ FALSE, IOStatus.Success };
    RETURN;
  ELSE
    file^.status := CStatusToIOStatus ( );
    RETURN;
  END; (* IF numWritten = numToWrite *)
END WriteBlock;

(* File-Only I/O operations *)


PROCEDURE isFlushable ( file : File ) : BOOLEAN;
(* Returns TRUE if <file> is flushable, otherwise FALSE. *)

BEGIN
  (*RETURN NOT isSpecialFile ( fd ) *)
END isFlushable;


PROCEDURE Flush ( file : File );

VAR
  res : INTEGER32;

BEGIN
  IF NOT isValidAccessor (file) THEN
    RETURN
  END; (* IF NOT isValidAccessor (file) *)
  IF NOT (FileMode.Write IN file^.mode ) THEN
    file^.status := IOStatusT{ TRUE, IOStatus.NotOpenForWriting };
    RETURN
  END; (* IF NOT (FileMode.Write *)
  res := CStdIOLib.fflush (file^.fd);
  IF res = 0 THEN
    file^.status := IOStatusT{ FALSE, IOStatus.Success };
  ELSE
    file^.status := IOStatusT{ TRUE, IOStatus.DeviceError };
  END; (* IF res = 0 *)
END Flush;


(* Writes unwritten data in any buffer of <file> to its associated file. *)

(* Any attempt to write to or flush a file whose write flag is not set shall
   fail with status IOStatus.OperationNotSupported. *)
(*

 Notes:
(1) open and close are not permitted on standard streams
(2) open with buffer size is only permitted on files, not streams/pipes
(3) positioning and flush is only permitted on files, not streams/pipes
(4) read operations are permitted only if the read flag is set
(5) write operations are permitted only if the write or append flag is set

*)


(* Standard Files *)

PROCEDURE stdInFile (): File;
(* Returns the file pointer associated with the current process' stdin *)

BEGIN
  RETURN stdIn
END stdInFile;


PROCEDURE stdOutFile (): File;
(* Returns the file pointer associated with the current process' stdout *)

BEGIN
  RETURN stdOut
END stdOutFile;


PROCEDURE stdErrFile (): File;
(* Returns the file pointer associated with the current process' stderr *)

BEGIN
  RETURN stdErr
END stdErrFile;

BEGIN
  Init ();
  
END FilePtrIO.
