(* (C) 2016 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE StdIO;

(* stream Based Standard IO Streams *)

FROM Transition IMPORT OCTET;
FROM Transition IMPORT CAPACITYo;
FROM Storage IMPORT ALLOCATE, DEALLOCATE;
IMPORT ChanIO, FileDescIO;

IMPORT IOSIZE; FROM IOSIZE IMPORT IOSIZET;
IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;
IMPORT FileMode; FROM FileMode IMPORT FileModeT;

IMPORT ISO646, CharIO;

CONST
  ENOERR = 0;  (* no error *)
  EINVST = 1; (* Invalid stream *)
  ENOTSUP = 2; (* not supported *)
  MaxStreamError = 2;
  InvalidFile = FileDescIO.InvalidFile;


TYPE
  StreamErrors = [0..MaxStreamError];

VAR
  StreamErrMsg : ARRAY StreamErrors OF IOStatus.statusMsg;
  LastErrNum : StreamErrors;

VAR
  opVector : ChanIO.Operations;

  defaultIn, defaultOut, defaultErr : Stream;
  currentIn, currentOut, currentErr : Stream;

  echoIn, echoOut, echoErr : BOOLEAN;

(* Inspection *)


PROCEDURE isStdIOChannel ( stream : Stream  ) : BOOLEAN;
(* Returns TRUE if <stream> is a valid file stream, else FALSE. *)

BEGIN
  RETURN (stream = currentIn) OR (stream = currentOut) OR (stream = currentErr)
       OR (stream = defaultIn) OR (stream = defaultOut) OR (stream = defaultErr)
END isStdIOChannel;


PROCEDURE isValidStream ( stream : Stream ) : BOOLEAN;
(* Returns TRUE if the stream accessor of <stream> is valid, else FALSE. *)

BEGIN
  RETURN isStdIOChannel (stream)
END isValidStream;


(* Note that in the following we put in the handles here because
  (a) doing it in the main part of the module did not work in P1 for unknown reasons
  (b) it is conceivable that FileDesciO could change these from time to time *)

PROCEDURE in() : Stream;
(* Returns a StdIO stream associated with stdin *)

BEGIN
  defaultIn^.handle := FileDescIO.defaultInFile();
  RETURN currentIn
END in;


PROCEDURE out() : Stream;
(* Returns a StdIO stream associated with stdout *)

BEGIN
  defaultOut^.handle := FileDescIO.defaultOutFile();
  RETURN currentOut
END out;


PROCEDURE err() : Stream;
(* Returns a StdIO stream associated with stderr *)

BEGIN
  defaultErr^.handle := FileDescIO.defaultErrFile();
  RETURN currentErr
END err;


PROCEDURE SetIn ( stream : Stream; echo : BOOLEAN );

BEGIN
  currentIn := stream;
  echoIn := echo
END SetIn;


PROCEDURE SetOut ( stream : Stream; echo : BOOLEAN );

BEGIN
  currentOut := stream;
  echoOut := echo
END SetOut;


PROCEDURE SetErr ( stream : Stream; echo : BOOLEAN );

BEGIN
  currentErr := stream;
  echoErr := echo
END SetErr;

PROCEDURE ResetDefaultIn;

BEGIN
  defaultIn^.handle := FileDescIO.defaultInFile();
  currentIn := defaultIn;
  echoIn := FALSE
END ResetDefaultIn;


PROCEDURE ResetDefaultOut;

BEGIN
  defaultOut^.handle := FileDescIO.defaultOutFile();
  currentOut := defaultOut;
  echoOut := FALSE
END ResetDefaultOut;


PROCEDURE ResetDefaultErr;

BEGIN
  defaultErr^.handle := FileDescIO.defaultErrFile();
  currentErr := defaultErr;
  echoErr := FALSE
END ResetDefaultErr;


PROCEDURE statusOf ( stream : Stream  ) : IOStatusT;
(* Returns the status of the most recent operation on <stream>. *)

BEGIN
  IF isValidStream ( stream ) THEN
    LastErrNum := ENOERR;
    RETURN IOStatusT {FALSE, IOStatus.Success }
  ELSE
    LastErrNum := EINVST;
    RETURN IOStatusT {TRUE, IOStatus.Failure }
  END
END statusOf;


PROCEDURE StatusMsg ( stream : Stream ; status : IOStatusT );
(* Writes a status message for <status> to <stream>. *)
(* comment out until we have clarity about this *)
VAR
  msg : IOStatus.statusMsg;
  count : IOSIZET;
  numWritten : IOSIZET;

BEGIN
  IF NOT (isValidStream ( stream ) ) THEN
     LastErrNum := EINVST;
     msg := StreamErrMsg [EINVST]
  ELSIF stream = in() THEN
    LastErrNum := ENOTSUP;
    msg := StreamErrMsg [ENOTSUP]
  ELSE;
    (* all ok so far, so continue with whatever was passed in *)
    IOStatus.fetchMsg (status, msg);
    count := 0;
    WHILE (count < CAPACITYo (msg) ) AND ( msg [count] #  ISO646.NUL ) DO
      WriteOctet (stream, msg [count] );
      count := count + 1
    END; (* while *)
    CharIO.WriteStr( stream, "\n" );
  END (* main if *)
END StatusMsg;


PROCEDURE SetStatus
  ( stream : Stream ; status : IOStatusT; VAR valid : BOOLEAN );
(* Sets  the status of <stream> to <status> if it meets the following condition:
   Where <status.code> is set to Success <status.failed> must be FALSE, other-
   wise TRUE.  Passes TRUE in <valid> if successful, otherwise FALSE. *)

BEGIN
  IF isValidStream ( stream ) THEN
    FileDescIO.SetStatus (stream^.handle, status, valid ); (* checks validity *)
    LastErrNum := ENOERR;
  ELSE
    valid := FALSE;
    LastErrNum := EINVST
  END
END SetStatus;

(* Querying Stream Parameters *)


PROCEDURE modeOf ( stream : Stream  ) : FileModeT;
(* Returns the access mode of <stream>. *)

BEGIN
  IF isValidStream ( stream ) THEN
    RETURN FileDescIO.modeOf(stream^.handle)
  ELSE
    LastErrNum := EINVST;
    RETURN FileModeT{}  (* return an invalid mode *)
  END
END modeOf;


PROCEDURE nameLen ( stream : Stream  ) : CARDINAL;
(* Returns the length of the name associated with <stream>. *)

BEGIN
  IF isValidStream ( stream ) THEN
    RETURN 0
  ELSE
    LastErrNum := EINVST;
  END
END nameLen;


PROCEDURE GetName ( stream : Stream ; VAR name : ARRAY OF CHAR );
(* Passes the name of the stream associated with <stream> back in <name>. If the
   name exceeds  the capacity of <name>,  an empty string is passed in <name>.
   Does not alter the status of <stream>. *)

BEGIN
  IF isValidStream ( stream ) THEN
    name [0] := ""
  ELSE
    LastErrNum := EINVST;
  END
END GetName;

(* I/O operations *)


PROCEDURE dataReady ( stream : Stream ; octets : IOSIZET ) : BOOLEAN;
(* Returns TRUE if <stream> has at least <octets> number of octets available for
   reading, otherwise FALSE. *)

BEGIN
  IF isValidStream ( stream ) THEN
    RETURN (stream = defaultIn)
  ELSE
    LastErrNum := EINVST;
  END
END dataReady;


PROCEDURE ReadOctet ( stream : Stream ; VAR data : OCTET );
(* Reads one octet of data from <stream> and passes it back in <data>. *)

BEGIN
  IF isValidStream ( stream ) THEN
    IF (stream = currentIn) THEN
      FileDescIO.ReadOctet(stream^.handle, data);
      IF echoIn THEN FileDescIO.WriteOctet ( defaultOut^.handle, data ) END
    ELSE
      LastErrNum := ENOTSUP
    END;
  ELSE
    LastErrNum := EINVST;
  END
END ReadOctet;


PROCEDURE ReadBlock
  ( stream : Stream ; VAR data : ARRAY OF OCTET; VAR octetsRead : IOSIZET );
(* Reads a block of data from <stream> and passes it back in <data>. *)

BEGIN
  IF isValidStream ( stream ) THEN
    IF (stream = currentIn) THEN
      FileDescIO.ReadBlock(stream^.handle, data, octetsRead);
      IF echoIn THEN
        IF stream^.handle = InvalidFile THEN ResetDefaultIn END;
        FileDescIO.WriteBlock ( defaultOut^.handle, data, octetsRead )
      END (* IF echoIn *)
    ELSE
      LastErrNum := ENOTSUP;
    END; (* IF (stream = currentIn) *)
  ELSE
    LastErrNum := EINVST
  END (* IF isValidStream ( stream ) *)
END ReadBlock;


PROCEDURE insertReady ( stream : Stream  ) : BOOLEAN;
(* Returns FALSE if the insert buffer of <stream> is full, otherwise TRUE. *)

BEGIN
  IF isValidStream ( stream ) THEN
    RETURN FALSE (* no inserting on streams *)
  ELSE
    LastErrNum := EINVST
  END (* IF isValidStream ( stream ) *)
END insertReady;


PROCEDURE Insert ( stream : Stream ; data : OCTET );
(* Inserts octet <data> into input stream <stream>  to be read by the next read
   operation on <stream>. *)

BEGIN
  IF isValidStream ( stream ) THEN
    LastErrNum := ENOTSUP;
    RETURN (* nothing else happens *)
  ELSE
    LastErrNum := EINVST
  END (* IF isValidStream ( stream ) *)
END Insert;

(* Any attempt to perform the above I/O operations on a stream whose Read flag
   is not set will fail with status NotOpenForReading. *)


PROCEDURE WriteOctet ( stream : Stream ; data : OCTET );
(* Writes <data> to <stream>. *)

BEGIN
  IF isValidStream ( stream ) THEN
    IF (stream = currentIn) THEN
      LastErrNum := ENOTSUP;
      RETURN
    ELSIF (stream = currentOut) THEN
      IF echoOut OR (currentOut = defaultOut) THEN
        FileDescIO.WriteOctet ( defaultOut^.handle, data )
      ELSE
        FileDescIO.WriteOctet ( currentOut^.handle, data );
      END; (* IF echoOut *)
    ELSIF (stream = currentErr)  THEN
      IF echoErr OR (currentErr = defaultErr) THEN
        FileDescIO.WriteOctet ( defaultErr^.handle, data )
      ELSE
        FileDescIO.WriteOctet ( currentErr^.handle, data );
      END (* IF echoErr *)
    END (* IF (stream = currentIn) *)
  ELSE
    LastErrNum := EINVST
  END (* IF isValidStream ( stream ) *)
END WriteOctet;


PROCEDURE WriteBlock
  ( stream : Stream ; data : ARRAY OF OCTET; VAR written : IOSIZET );
(* Writes <data> to <stream> and passes the actual number of octets written back
   in <written>. *)

BEGIN
  IF isValidStream ( stream ) THEN
    IF (stream # currentIn) THEN
      FileDescIO.WriteBlock(stream^.handle, data, written);
      IF (stream = currentOut) AND echoOut THEN
        IF stream^.handle = InvalidFile THEN ResetDefaultOut END;
        FileDescIO.WriteBlock ( defaultOut^.handle, data, written )
      ELSIF (stream = err() ) AND echoErr THEN
        IF stream^.handle = InvalidFile THEN ResetDefaultErr END;
        FileDescIO.WriteBlock ( defaultErr^.handle, data, written )
      END (* IF isValidStream ( stream ) *)
    ELSE
      LastErrNum := ENOTSUP
    END (* IF (stream # currentIn) *)
  ELSE
    LastErrNum := EINVST
  END (* IF isValidStream ( stream ) *)
END WriteBlock;

(* Any attempt to perform the above I/O operations on a stream whose Write flag
   is not set will fail with status NotOpenForWriting. *)


PROCEDURE isFlushable ( stream : Stream  ) : BOOLEAN;
(* Always returns FALSE. *)

BEGIN
   RETURN FALSE
END isFlushable;


PROCEDURE Flush ( stream : Stream  );
(* Always fails with status OperationNotSupported. *)

BEGIN
  LastErrNum := ENOTSUP
END Flush;

VAR num : CARDINAL; VAR outChan : Stream;

BEGIN (* main *)
  (* install methods *)
  NEW ( opVector );
  opVector^.statusOf := statusOf;
  opVector^.StatusMsg := StatusMsg;
  opVector^.SetStatus := SetStatus;
  opVector^.statusOf := statusOf;
  opVector^.modeOf := modeOf;
  opVector^.nameLen := nameLen;
  opVector^.GetName := GetName;
  opVector^.dataReady := dataReady;
  opVector^.ReadOctet := ReadOctet;
  opVector^.ReadBlock := ReadBlock;
  opVector^.insertReady := insertReady;
  opVector^.Insert := Insert;
  opVector^.WriteOctet := WriteOctet;
  opVector^.WriteBlock := WriteBlock;
  opVector^.isFlushable := isFlushable;
  opVector^.Flush := Flush;

  (* init stdin stream *)
  NEW ( defaultIn );
  defaultIn^.op := opVector;
  echoIn:= FALSE;
  defaultIn^.handle := FileDescIO.defaultInFile();

  (* init stdout stream *)
  NEW ( defaultOut );
  defaultOut^.op := opVector;
  echoOut := FALSE;
  defaultOut^.handle := FileDescIO.defaultOutFile();

  (* init stderr stream *)
  NEW ( defaultErr );
  defaultErr^.op := opVector;
  echoErr := FALSE;
  defaultErr^.handle := FileDescIO.defaultErrFile();
  
  (* init errors and messages *)
  LastErrNum := ENOERR;
  StreamErrMsg [ENOERR] := "No Stream error encountered.";
  StreamErrMsg [EINVST] := "Invalid Stream.";
  StreamErrMsg [ENOTSUP] := "Operation not supported on streams.";

  ResetDefaultIn;
  ResetDefaultOut;
  ResetDefaultErr
END StdIO.