IMPLEMENTATION MODULE RealMath;


(*          Implementation  1994, 2015
                by R. Sutcliffe
        Trinity Western University
7600 Glover Rd., Langley, BC Canada V3A 6H4
         e-mail: rsutc@twu.ca
    Last modification date 2017 11 14
=========================================== *)
 (* Mathematical functions for the type REAL *)

IMPORT Perror;
IMPORT Transition; FROM Transition IMPORT FLOATi, INT;

CONST
  minDelta = 1.0e-7;
  minDeltaL = 1.0e-15;
  maxIts = 300;
  ln2 = 0.6931471805599453094172321214581765680755001343602552;
  ln10 = 2.3025850929940456840179914546843642076011014886287729;
  e5 = 148.41315910257660342111558004055227962348766759387898;
  e20 = 4.85165195409790277969106830541540558684638988944E08;
  tau10 = 62.831853071795864769252867665590057683943387987502116;
  tau100 = 628.31853071795864769252867665590057683943387987502116;
  piOver2 = 1.5707963267948966192313216916397514420985846996875529;


PROCEDURE sgn ( r : REAL ) : REAL;

BEGIN
  IF r > 0.0 THEN
    RETURN 1.0
  ELSIF r < 0.0 THEN
    RETURN -1.0
  ELSE
    RETURN 0.0
  END (* IF r > 0.0 *)
END sgn;


PROCEDURE floor ( r : REAL ) : REAL;
(* Returns the floor value of r. *)
VAR
  temp : REAL;

BEGIN
  temp := FLOATi (INT (r));
  IF temp > r THEN temp := temp - 1.0 END;
  RETURN temp
END floor;


PROCEDURE ceil ( r : REAL ) : REAL;
(* Returns the ceiling value of r. *)

VAR
  temp : REAL;

BEGIN
  temp := FLOATi (INT (r));
  IF temp < r THEN temp := temp + 1.0 END;
  RETURN temp
END ceil;


PROCEDURE frac ( r : REAL ) : REAL;
(* Returns the fractional part of r. *)

BEGIN
  RETURN r - FLOATi (INT (r))
END frac;

(* local *)
PROCEDURE int ( r : REAL ) : REAL;
(* Returns the integer part of r. *)

BEGIN
  RETURN FLOATi (INT (r))
END int;


PROCEDURE inv ( r : REAL ) : REAL;
(* Returns the multiplicative inverse 1/r of r. *)

BEGIN
  RETURN 1.0/r
END inv;


PROCEDURE trunc ( r : REAL; n : CARDINAL ) : REAL;
(* Returns the value of r truncated to n decimal places. *)

BEGIN
  r := r * powerN(10.0, n);
  r := int (r);
  r := r / powerN(10.0, n);
  RETURN r
END trunc;


PROCEDURE round ( r : REAL; n : CARDINAL ) : REAL;
(* Returns the value of r rounded to n decimal places. *)

BEGIN
  r := r * powerN(10.0, n-1);
  IF r > 0.0 THEN
    r := int (r + 0.5);
  ELSE
    r := int (r - 0.5)
  END; (* IF r > 0.0 *)
  r := r / powerN(10.0, n-1);
  RETURN r
END round;


PROCEDURE power (base, exponent: REAL): REAL;
  (* Returns the value of the number base raised to the power exponent *)

BEGIN
  RETURN exp(exponent * ln (base))
END power;


PROCEDURE powerN (base: REAL; exponent: INTEGER): REAL;
  (* Returns the value of the number base raised to the power exponent *)

BEGIN
  IF exponent < 0 THEN
    RETURN powerN (1.0 / base, -exponent)
  ELSIF exponent = 0  THEN RETURN  1.0;
  ELSIF exponent = 1 THEN RETURN base;
  ELSIF ODD(exponent) THEN RETURN  base * powerN (base * base, (exponent - 1) DIV 2)
  ELSE RETURN powerN (base * base, exponent DIV 2)
  END (* IF exponent *)
END powerN;


PROCEDURE sqrt (x: REAL): REAL;
  (* Returns the positive square root of x *)
VAR
 rOldEst, rNewEst : REAL;
 fx, dfx : REAL; (* the function and its derivative *)
 numIts : CARDINAL;

BEGIN
  IF x < 0.0 THEN
    Perror.prerr ( "Can't take square root of negative value.")
  ELSE
    numIts := 0;
    rNewEst :=power (x, 0.5); (* try with exp first , then sharpen a bit to see if it improves. *)
    REPEAT
      rOldEst := rNewEst; (* old estimate gets newest one *)
      fx := rOldEst * rOldEst - x; (* function we want to equal zero *)
      dfx := 2.0 * rOldEst; (* derivative of function above *)
      rNewEst := rOldEst - (fx/dfx); (* compute newest estimate *)
      numIts := numIts + 1
    UNTIL (ABS (fx) < minDelta) OR (numIts = maxIts); (* when to quit computing *)
  END; (*if *)
  RETURN rNewEst
END sqrt;


PROCEDURE exp (x: REAL): REAL;
  (* Returns the exponential of x *)
  (* uses the Taylor series--slow but portable *)

VAR
  addend, ratio, partialSum : REAL;
  n : CARDINAL;

BEGIN
  addend := 1.0;
  partialSum := addend;
  n := 1;
  REPEAT
    ratio := x / FLOATi (n);
    addend := addend * ratio;
    partialSum := partialSum + addend;
    n := n + 1;
  UNTIL ( ABS(addend) <= minDelta) OR (n = maxIts);
  RETURN partialSum;

END exp;


PROCEDURE ln (x: REAL): REAL;
  (* Returns the natural logarithm of x *)

VAR
 rOldEst, rNewEst, expOld : REAL;
 delta : REAL; (* the function and its derivative *)
 numIts, m : CARDINAL;
 xLessOne : BOOLEAN;

BEGIN
  IF x = 1.0 THEN
    RETURN 0.0
  ELSIF x <= 0.0 THEN
    Perror.prerr ( "Can't take the logarithm of a negative or zero value.")
  ELSE
    xLessOne := ( x < 1.0);
    IF xLessOne THEN x := 1.0/x END;
    (* write as x = exp(m)*r so we deal only with x close to one *)
    m := 0;
    WHILE x > e20 DO (* bigger rocks first *)
      x := x / e20;
      m := m + 20
    END; (* while *)
    WHILE x > e5 DO (* bigger rocks first *)
      x := x / e5;
      m := m + 5
    END; (* while *)
    WHILE x > e DO
      x := x / e;
      m := m + 1
    END; (* while *)
    numIts := 0;
    rNewEst := 1.0; (* simple first estimate *)
    (* The Taylor series converges very slowly and would give poor results even in the sixth place, do we
    use the iterative formula y(n+1) = y(n) + 2 *((x-exp(y(n))/((x+exp(y(n))
    as the series for exp converges rapidly *)
    REPEAT
      rOldEst := rNewEst; (* old estimate gets newest one *)
      expOld := exp(rOldEst); (* only want to compute it once *)
      delta := (2.0 * ( x -  expOld) / (x + expOld));
      rNewEst := rOldEst + delta; (* compute newest estimate *)
      numIts := numIts + 1
    UNTIL (ABS (delta) < minDelta) OR (numIts = maxIts); (* when to quit computing *)
    rNewEst := rNewEst + FLOATi (m);
    IF xLessOne THEN rNewEst := -rNewEst END;
    RETURN rNewEst
  END (* if *)
END ln;


PROCEDURE exp10 (x: REAL): REAL;
  (* Returns 10 to the x *)
  (* do this by e to the ln10 * x) *)

BEGIN
  RETURN (exp (x * ln10))
END exp10;


PROCEDURE log ( r : REAL ) : REAL;
(* Returns the decimal logarithm of r. *)
VAR
  m : CARDINAL;

BEGIN
(* special case a few as logs of powers of ten can be problematic. *)

  IF r = 0.1 THEN
   RETURN -1.0
  ELSIF r = 0.01 THEN
   RETURN -2.0
  ELSIF r = 0.001 THEN
   RETURN -3.0
  ELSIF r = 0.0001 THEN
   RETURN -4.0
  ELSIF r = 0.00001 THEN
   RETURN -5.0
  ELSIF r = 0.000001 THEN
   RETURN -6.0
  END;

  m := 0;
  WHILE r > 1.0E10 DO
    r := r/ (1.0E10);
    m := m + 10
  END; (* while *)
  WHILE r > 1.0E3 DO
    r := r/ (1.0E3);
    m := m + 3
  END;  (* while *)
  RETURN (ln (r) / ln10) + FLOATi (m)
END log;


PROCEDURE exp2 (x: REAL): REAL;
  (* Returns 2 to the x *)

BEGIN
  RETURN (exp (x * ln2))
END exp2;


PROCEDURE log2 ( r : REAL ) : REAL;
(* Returns the base 2 logarithm of r. *)

BEGIN
  RETURN ln (r) / ln2
END log2;


  (* The angle in all trigonometric functions is measured in radians *)


PROCEDURE sin (x: REAL): REAL;
  (* Returns the sine of x *)
  (* Returns the exponential of x *)
  (* uses the Taylor series--slow but portable *)
VAR
  x2, addend, ratio, partialSum : REAL;
  n : CARDINAL;
  sinNeg : BOOLEAN;

BEGIN
  IF x < 0.0 THEN
    sinNeg := TRUE;
    x := ABS (x)
  ELSE
    sinNeg := FALSE
  END; (* IF x < 0.0 *)
  (* hammer it down to a reference angle
    surely they won't start with values in the thousands of tau *)
  WHILE x > tau100 DO
    x := x - tau100
  END;(* while *)
  WHILE x > tau10 DO
    x := x - tau10
  END; (* while *)
  WHILE x > tau DO
    x := x - tau
  END; (* while *)
  IF x > pi THEN
    sinNeg := NOT sinNeg;
    x := x - pi
  END;
  IF x > piOver2 THEN
    x := x - piOver2
  END; (* IF x > piOver2 *)

  addend := x;
  x2 := x * x;
  partialSum := addend;
  n := 1;
  REPEAT
    ratio := -x2 / FLOATi ((n+1) * (n + 2));
    addend := addend * ratio;
    partialSum := partialSum + addend;
    n := n + 2;
  UNTIL ( ABS(addend) <= minDelta) OR (n = maxIts);
  IF sinNeg THEN partialSum := -partialSum END;
  RETURN partialSum
END sin;


PROCEDURE cos (x: REAL): REAL;
  (* Returns the cosine of x *)

BEGIN
  RETURN sin (piOver2 - x)
END cos;


PROCEDURE tan (x: REAL): REAL;
  (* Returns the tangent of x *)
VAR
 cosine : REAL;

BEGIN
  cosine := cos(x);
  IF cosine = 0.0  THEN
    Perror.prerr ( "Can't take tan of (2n-1)*pi/2 multiple.")
  ELSE
    RETURN sin(x)/cosine
  END (* IF cosine *)
END tan;


PROCEDURE arcsin (x: REAL): REAL;
  (* Returns the arcsine of x
  We use the Taylor series *)
VAR
  x2, addend, ratio, partialSum, nRatio : REAL;
  n : CARDINAL;
  sinNeg : BOOLEAN;

BEGIN (* works best in the range 0 to .5, so *)
  sinNeg := x < 0.0;
  x := ABS (x);
  IF x > 1.0  THEN
    Perror.prerr ( "Can't take arcsin outside -1 <= x <= 1")
  ELSIF x > 0.71 THEN
    RETURN piOver2 - arcsin (sqrt(1.0 - x * x) )
  ELSE
    addend := x;
    x2 := x * x;
    partialSum := addend;
    n := 1;
    REPEAT
    ratio := x2 * FLOATi( (2*n - 1)*(2*n - 1)) / FLOATi(4*n*n + 2*n);
    addend := addend * ratio;
    partialSum := partialSum + addend;
    n := n + 1;
    UNTIL ( ABS(addend) <= minDelta) OR (n = maxIts);
    IF sinNeg THEN partialSum := -partialSum END; (* return in range -pi/2..pi/2 *)
    RETURN partialSum
  END; (* IF x > 1.0 *)
END arcsin;


PROCEDURE arccos (x: REAL): REAL;
  (* Returns the arccosine of x *)

BEGIN
  RETURN piOver2 - arcsin (sqrt(1.0 - x * x) )
END arccos;


PROCEDURE arctan (x: REAL): REAL;

VAR
 rOldEst, rNewEst : REAL;
 fx, dfx, c : REAL; (* the function and its derivative *)
 numIts : CARDINAL;

BEGIN
(* this works best up to tan (y) = 1, so for values over that, go by the cotangent and subtract *)
  IF x > 1.1 THEN
    RETURN piOver2 - arctan (1.0 / x)
  END; (* IF x > 1.1 *)
  (* after that, it's Newton-Raphson all the way *)
  numIts := 0;
  rNewEst := 0.0;
  REPEAT
    rOldEst := rNewEst; (* old estimate gets newest one *)
    fx := tan (rOldEst) - x; (* function we want to equal zero *)
    c := cos (rOldEst); (* since the derivative is sec squared, we multiply by cos squared *)
    dfx := (c*c); (* recip of derivative of function above *)
    rNewEst := rOldEst - (fx*dfx); (* compute newest estimate *)
    numIts := numIts + 1
  UNTIL (ABS (fx) < minDelta) OR (numIts = maxIts); (* when to quit computing *)
  RETURN rNewEst
END arctan;

(* hyperbolics *)


PROCEDURE sinh (x: REAL): REAL;
  (* Returns the hyperbolic sine of x *)

BEGIN
  RETURN 0.5 * (exp(x) - exp (-x))
END sinh;


PROCEDURE cosh (x: REAL): REAL;
  (* Returns the hyperbolic cosine of x *)

BEGIN
  RETURN 0.5 * (exp(x) + exp (-x))
END cosh;


PROCEDURE tanh (x: REAL): REAL;
  (* Returns the hyperbolic tangent of x *)
VAR
  num, denom : REAL;

BEGIN
  num := exp(x) - exp (-x);
  denom := exp(x) + exp (-x);
  RETURN num/denom
END tanh;

BEGIN (* main *)

END RealMath.
